/* 
!C*****************************************************************************

!File: ids.c

!Description: Routines for putting out the intermediate CMG product (IDS product)
              from MOD_PR09.exe.  These routines take care of SDstarting the files,
	      filling them with data, and SDending them.  They do NOT take care of 
	      the IDS file metadata -- routines for IDS file metadata are found in 
	      meta.c.

!Input Parameters: none
    
!Output Parameters: none

!Input/Output Parameters: none

!Returns: nothing

!Revision History:

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

    MODIS Science Team Members:
      Chris Justice
      MODIS Land Science Team           University of Maryland
      justice@kratmos.gsfc.nasa.gov     NASA/GSFC Code 923
      phone: 301-286-7372               Greenbelt, MD 20771

      Eric F Vermote
      Univerisity of MD.
      NASA's GSFC Code 923
      Greenbelt, MD , 20771
      eric@kratmos.gsfc.nasa.gov
      301-286-6232

      Jim Ray
      Science Systems and Applications Inc.
      NASA's Goddard Space Flight Center Code 923
      Greenbelt MD, 20771
      jim@kratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
     
     These routines update several global variables: 
     
*         int currentlyopen, current_n_CMGwindows, n_ret;  }
*         int make_this_cmg[3];                            } must be initialized to zero before start
*         input_metadata **CMGmetadata;    } one *CMGmetadata for each IDS file (three at most)
*         d_sds **sds_outputlist;          } one *sds_outputlist for each SDS in an IDS file
                                            (see documentation for get_set_sds(), below).

     and use several global definitions:
     
*         enum{NTHPL, MIDDLE, STHPL, NCMGFILES}; 
*         #define ULLAT 90.0000F
*         #define LRLAT -90.0000F
*         #define ULLON -180.0000F
*         #define LRLON 180.0000F 
*         #define POLELIM 60.0000F
*         #define NONPOLEWIDTH 90.0000F
     
     Usage: for each scan processed, do this:

*        if (commands.makegcm == 1) {
*	      if (AOTR_flag == 1) which_ids(local_input[GEO_FILE]->data[0]);
*	      else                update_ids(scan_day, local_input, local_output[0]);
*	      scan_day++;
*	      }

     During the first pass, which_ids() determines which IDS files are to be made.
     During the second pass, update_ids() initializes and makes them; update_ids()
     also initializes IDS metadata, and updates the inputpointer metadata.
     At the end, simply run this:

*        end_coarse_grid(sds_outputlist, currentlyopen, current_n_CMGwindows);

     Other routines must be run to finish off the IDS metadata (see meta.c).
     These must be run BEFORE end_coarse_grid() is run.
     
!END***************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include "mfhdf.h"
#include "math.h"

#include "PGS_PC.h"

#include "PGS_MODIS_37092.h"
#include "modis_srefl.h"
#include "protos.h"
#include "ids.h"
#include "smf.h"
#define AOTAQA_FILL 127

void which_ids(float *lat)
/*
!C*****************************************************************************
!Description: Wrapper for preview_coarse_grid() (below); determines which IDS files
              are to be generated.

!Input Parameters: 
    lat         latitude values for entire scan being processed
    
!Output Parameters: none

!Input/Output Parameters: none

!Returns: nothing

!Revision History:

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
     
        Updates global array make_this_cmg[].
	
	Called by routine atmospheric_correction() in atmcor.c.
	Calls routine preview_coarse_grid() in this file.
	
!END***************************************************************************
*/
{
int i, which_file;
/*char location[50]="ids.c, which_ids\0";
printf("in %s\n", location);*/

for(i=0;i<NBSCANS_1KM;i++) {
 preview_coarse_grid(i*NBSPOTS_1KM, lat, &which_file);
 if (which_file == -1) {  /* error condition: bad latitude values in lat */
   }
 else if (which_file == 3) {
    make_this_cmg[0] = 1;
    make_this_cmg[1] = 1;
   }
 else if (which_file == 4) {
    make_this_cmg[2] = 1;
    make_this_cmg[1] = 1;
   }
 else 
    make_this_cmg[which_file] = 1;
   }
	    
/*printf("out %s\n", location);*/
return;
}


PGSt_SMF_status update_ids(int scan_number, dims_struct **input, dims_struct *local_output)
/*
!C*****************************************************************************
!Description: Wrapper for update_coarse_grid() (below); initializes IDS files
               for scan_number==0, updates IDS files otherwise.

!Input Parameters: 
    scan_number            Set to 0 for the first scan of the first granule.
                           Scan number for whole swath, not granule.
			   
    input                  Structure holding all input info (L1B and geolocation)
    
    local_output           Structure holding all fine-resolution reflectance data.
    
!Output Parameters: none

!Input/Output Parameters: none

!Returns: nothing

!Revision History:

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
     
        Reads global array make_this_cmg[].
	Initializes global variables currentlyopen, current_n_CMGwindows, n_ret, etc. 
	Initializes and updates global variable sds_outputlist[], the IDS file(s) structure.
	
	Called by routine atmospheric_correction() in atmcor.c.
	Calls routines start_coarse_grid() and update_coarse_grid() in this file,
	as well as metadata-handling routines (e. g., grow_Mstringarray() in meta.c).
	
!END***************************************************************************
*/
{
int i, ii, j, which_file, CMGret;
char location[50]="ids.c, update_ids\0";
char sfilename[PGSd_PC_FILE_PATH_MAX];
char filename[PGSd_PC_FILE_PATH_MAX];
/*printf("in %s\n", location);*/

/*if (scan_number == 0) for (i=0;i<3;i++) printf("%d\n", make_this_cmg[i]);*/

if (scan_number == 0) {
     currentlyopen = 0;
     current_n_CMGwindows = 1;
     n_ret = 0;   /* Added 19-DEC-03 */
     
     CMGmetinit(input[RAD1_FILE]->xxxxx->rbdt);
	  
     strcpy(CMG_lgid_template, strstr(input[RAD1_FILE]->xxxxx->localgranuleid, "A"));      
	     	     
     /*strcpy(CMG_lgid_template, "whatever\0");  / * for now */
	  
     sds_outputlist = (d_sds **)AllocMem(CRSGRDSDSS*sizeof(d_sds *), location);
     for(i=0;i<CRSGRDSDSS;i++)
         sds_outputlist[i] = (d_sds *)AllocMem(sizeof(d_sds), location);
       	   
     if (start_coarse_grid(make_this_cmg, sds_outputlist) != 0) commands.makegcm = 0;

}
/*else {   /   * 19-DEC-03: Perhaps remove this, such that the below is executed for scan==0... */
  for(i=0;i<NBSCANS_1KM;i++) {
     CMGret = update_coarse_grid(i*NBSPOTS_1KM, input[GEO_FILE], local_output, &which_file);
     if (CMGret == 1)
          commands.makegcm = 0;
     else if (CMGret == 2)
	  n_ret++;
	  
     /*   This updates the inputpointers for the CMG's metadata  */
     for(ii=0;ii<ALL_L1B_FILES;ii++) {
        if (input[ii]->OK == 0) {
           strcpy(sfilename, input[ii]->filename);
           if (rmpath(sfilename)!=MODIS_S_SUCCESS){
               modsmf((PGSt_SMF_code)MODIS_E_FUNCTION_ERROR, "Failure from rmpath()", location);
               return((PGSt_SMF_status)MODIS_E_FUNCTION_ERROR);
               }
       
	   if (which_file == -1) {  /* error condition: bad latitude values in allgeoloc.slatP */
	       }
	   else if (which_file == 3) {
               grow_Mstringarray(CMGmetadata[0]->inputpointer, PGSd_PC_FILE_NAME_MAX, sfilename);
               grow_Mstringarray(CMGmetadata[1]->inputpointer, PGSd_PC_FILE_NAME_MAX, sfilename);
	       }
	   else if (which_file == 4) {
               grow_Mstringarray(CMGmetadata[2]->inputpointer, PGSd_PC_FILE_NAME_MAX, sfilename);
               grow_Mstringarray(CMGmetadata[1]->inputpointer, PGSd_PC_FILE_NAME_MAX, sfilename);
	       }
	   else {
               grow_Mstringarray(CMGmetadata[which_file]->inputpointer, PGSd_PC_FILE_NAME_MAX, sfilename);
               }
	   
	 }
 
      }

     }   /* for(i=0;i<NBSCANS_1KM;i++) */

  /* }*/
/*printf("out %s\n", location);*/
return 0;
}

int start_coarse_grid(int make_this_cmg[3], d_sds **sds_outputlist)
/*
!C*****************************************************************************
!Description: opens up the CMG file, starts and initializes all SDSs in the CMG
                file and the sds_outputlist buffers.

!Input Parameters: 
    make_this_CMG          tells routine which CMG to make 
    
!Output Parameters: none

!Input/Output Parameters: 
    sds_outputlist         updated

!Returns: success (0), fail (nonzero)

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes

      Called by update_ids(), above.
      
!END***************************************************************************
*/
{
#include "qa_map.h"   /* new, 29-SEP-03 */
int32 sdid[NCMGFILES];
int32 i, j, sds, rank, dimid, sdidout, outputdims[2];
int jumpstatus, size, currentlyopen;
jmp_buf jumpbuf;
char location[100]="ids.c, start_coarse_grid\0";
char tf_name[256];
char msg[256];
PGSt_PC_Logical lun;
PGSt_integer ver;
char message[1700] = {"\0"};
char long_message[2400] = {"\0"};
uint8 *ui8_array;
/*printf("in %s\n", location);*/

size = (int) DEGPERWIN*PIXPERDEG;  /* lines per window */
currentlyopen=0;

outputdims[0] = (int32)((ULLAT-LRLAT)*PIXPERDEG);
outputdims[1] = (int32)((LRLON-ULLON)*PIXPERDEG);

rank = 2;

if(!(jumpstatus=setjmp(jumpbuf))) {
  for (j=0;j<NCMGFILES;j++) {
    if (!make_this_cmg[j])continue;
    
    lun = NPOLEFNAMELUN + j;
    ver = (PGSt_integer)1;
    if (PGS_PC_GetReference(lun, &ver, tf_name) != PGS_S_SUCCESS){
         sprintf(msg, "PGS_PC_GetReference; Unable to open CMG output file: LUN %ld "
		      " Terminal error", (LINT)lun); 
         modsmf(MODIS_F_SDP_ERROR, msg, location);
      }
    if ((sdid[j] = SDstart(tf_name, DFACC_CREATE))==FAIL)
         longjmp(jumpbuf, 10);
    }
  }
if(jumpstatus==10){
    modsmf(MODIS_E_CMG_CREATE, "Error starting HDF SDS in CMG.", location);
    /*printf("out %s\n", location);*/
    return(1);
  }

for(i=0;i<CRSGRDSDSS;i++){   
  strcpy(sds_outputlist[i]->SDSname, CRSGRID_Name_Array[i]);
  sds_outputlist[i]->type = CRSGRID_Data_Type[i];
  sds_outputlist[i]->currentsize = 0;
  sds_outputlist[i]->data = NULL;
  sds_outputlist[i]->cmask = NULL;
  sds_outputlist[i]->fdata = NULL;
  sds_outputlist[i]->dims[0] = outputdims[0];
  sds_outputlist[i]->dims[1] = outputdims[1];
  for (j=0;j<NCMGFILES;j++){
      sds_outputlist[i]->sdsid[j] = -1;
      sds_outputlist[i]->sdid[j] = -1;
     }
     
  if(!(jumpstatus=setjmp(jumpbuf))){
    for (j=0;j<NCMGFILES;j++) {
      if (!make_this_cmg[j])continue;
    
      if (j!=MIDDLE) {
         sds_outputlist[i]->dims[0] = (int32)((ULLAT-POLELIM)*PIXPERDEG);
         sds_outputlist[i]->dims[1] = outputdims[1];
	 }
      else {
         sds_outputlist[i]->dims[0] = (int32)((2*POLELIM)*PIXPERDEG);
	 /* Set this here JUST FOR THE PURPOSES OF SDS SETUP; will be and needs to be
	    set BACK to outputdims[1] afterwards */
         sds_outputlist[i]->dims[1] = (int32)(NONPOLEWIDTH*PIXPERDEG);
	 }
	 
      sds_outputlist[i]->sdid[j] = sdid[j];
      if ((sds = SDcreate(sdid[j], CRSGRID_Name_Array[i], 
                        CRSGRID_Data_Type[i], rank, sds_outputlist[i]->dims))==FAIL)
			    longjmp(jumpbuf, 10);

      switch (i) {
         case(CRSGRDB1):
         case(CRSGRDB2):
         case(CRSGRDB3):
         case(CRSGRDB4):
         case(CRSGRDB5):
         case(CRSGRDB6):
         case(CRSGRDB7):
	    sprintf(msg, "CMG surface reflectance SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)DATAFILLVALUE, VRANGEHIGHREF, VRANGELOWREF, (float64)REFSCALE, 
	                         (float64)REFOFFSET, msg, 1, UNITSREFLECTANCE, CRSGRID_Name_Array[i]);
	    break;
         case(CRSGRDTAU1):
	    sprintf(msg, "CMG AOT SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)AOT1_FILL, (int16)3000, (int16)0, (float64)1000.0F, 
	                         (float64)REFOFFSET, msg, 1, "(unitless)\0", CRSGRID_Name_Array[i]);
	    break;
         case(CRSGRDTAU3):
	    sprintf(msg, "CMG AOT SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)AOT3_FILL, (int16)3000, (int16)0, (float64)1000.0F, 
	                         (float64)REFOFFSET, msg, 1, "(unitless)\0", CRSGRID_Name_Array[i]);
	    break;
         case(CRSGRDQA):
	    sprintf(msg, "CMG QA SDS (code internal band index %d)", i);
	    writeHDFSDSattributes_long(sds, (int32)0, (int32)(1<<28)*4, (int32)0, (float64)1000.0F, 
	                         (float64)REFOFFSET, msg, 0, "bit field\0", CRSGRID_Name_Array[i]);
            /* new, 03-JAN-07 (v5.3.15)	 */			 
            strcpy(long_message, QAMAP_1KM);
            SDsetattr(sds, "QA index\0", DFNT_CHAR, strlen(long_message), (void *)long_message);
	    break;
         case(CRSGRDICM):
	    sprintf(msg, "CMG Internal cloud mask SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)8191, (int16)0, (float64)1.0F, 
	                         (float64)REFOFFSET, msg, 0, "bit field\0", CRSGRID_Name_Array[i]);
            strcpy(message, QAMAP_INTCM);
            SDsetattr(sds, "QA index\0", DFNT_CHAR, strlen(message), (void *)message);
	    break;
         case(CRSGRDAQA):
	    sprintf(msg, "CMG AOT QA SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)AOTAQA_FILL, (int16)19, (int16)0, (float64)1.0F, 
	                         (float64)REFOFFSET, msg, 0, "(unitless)\0", CRSGRID_Name_Array[i]);
            strcpy(message, QAMAP_AOTQA);
	    strcat(message + strlen(message), "\t127     Fill value (no data)\n\0");  /* AOTAQA_FILL */
            SDsetattr(sds, "QA index\0", DFNT_CHAR, strlen(message), (void *)message);
	    break;
         case(CRSGRDSZ):
         case(CRSGRDVZ):
	    sprintf(msg, "CMG angle SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)18000, (int16)0, (float64)100.0F, 
	                         (float64)REFOFFSET, msg, 1, UNITSDEGREES, CRSGRID_Name_Array[i]);
	    break;
         case(CRSGRDRA):
	    sprintf(msg, "CMG angle SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)18000, (int16)-18000, (float64)100.0F, 
	                         (float64)REFOFFSET, msg, 1, UNITSDEGREES, CRSGRID_Name_Array[i]);
	    break;

         /* 18-MAY-04 */
         case(CRSGRDB20):
         case(CRSGRDB21):
         case(CRSGRDB31):
         case(CRSGRDB32):
	    sprintf(msg, "CMG angle SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)40000, (int16)0, (float64)100.0F, 
	                         (float64)REFOFFSET, msg, 1, "degrees K\0", CRSGRID_Name_Array[i]);
	    break;
         /* 04-AUG-05 */
         case(CRSGRDGTM):
	    sprintf(msg, "CMG time SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)2355, (int16)0, (float64)1.0F, 
	                         (float64)0.0, msg, 1, UNITSTIME, CRSGRID_Name_Array[i]);
	    break;
        case(CRSGRDOZ):
	    sprintf(msg, "CMG ozone SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)255, (int16)0, (float64)400.0F, 
	                         (float64)REFOFFSET, msg, 1, "cm atm\0", CRSGRID_Name_Array[i]);
	    break;
         case(CRSGRDN):
	    sprintf(msg, "CMG N averaged SDS (code internal band index %d)", i);
	    writeHDFSDSattributes(sds, (int16)0, (int16)40, (int16)0, (float64)1000.0F, 
	                         (float64)REFOFFSET, msg, 0, "(unitless)\0", CRSGRID_Name_Array[i]);
	    break;
           }
			    
      sds_outputlist[i]->sdsid[j] = sds;
    
      if((dimid=SDgetdimid(sds, 0))==FAIL)longjmp(jumpbuf, 10);
      if(SDsetdimname(dimid, DIM_ZERO)==FAIL)longjmp(jumpbuf, 10);

      if((dimid=SDgetdimid(sds, 1))==FAIL)longjmp(jumpbuf, 10);
      if(SDsetdimname(dimid, DIM_ONE)==FAIL)longjmp(jumpbuf, 10);
      }
    sds_outputlist[i]->dims[1] = outputdims[1];
/*
printf("%p %p %s %d\n", sds_outputlist[i]->sdid, sds_outputlist[i]->sdsid, sds_outputlist[i]->SDSname, 
   sds_outputlist[i]->type);
 */
  }  
  if(jumpstatus==10){
    modsmf(MODIS_E_CMG_CREATE, "Error creating HDF SDS in CMG.", location);
    /*printf("out %s\n", location);*/
    return(2);
  }
  
  /* Initialize */
  if(!(jumpstatus=setjmp(jumpbuf))){
     if ( (get_set_sds(0, sds_outputlist[i], currentlyopen, size))!=0)longjmp(jumpbuf, 10);
     ids_zero_out(sds_outputlist[i], size);
     if (i == CRSGRDAQA) {
        ui8_array = (uint8 *)sds_outputlist[i]->data;
  	memset(ui8_array, AOTAQA_FILL, size*sds_outputlist[i]->dims[1]*sizeof(uint8));
        }
     for (j=0;j<(outputdims[0]/size);j++) {
        if ( (get_set_sds(2, sds_outputlist[i], j*size, size))!=0)longjmp(jumpbuf, 10);
	}
  }  
  if(jumpstatus==10){
    modsmf(MODIS_E_CMG_IO, "Error reading/writing HDF SDS in CMG.", location);
    /*printf("out %s\n", location);*/
    return(3);
  }

  } /*  for(i=0;i<CRSGRDSDSS;i++)  */

if(!(jumpstatus=setjmp(jumpbuf))){
   if ( (transfer_sds(sds_outputlist))!=0)longjmp(jumpbuf, 10);
   }
if(jumpstatus==10){
    modsmf(MODIS_E_CMG_TRANSFER, "Error transferring CMG buffers.", location);
    /*printf("out %s\n", location);*/
    return(4);
  }

/*printf("out %s\n", location);*/
return(0);
}

int update_coarse_grid(int k, dims_struct *geo_input, dims_struct *local_output, int *which_file)
/*
!C*****************************************************************************
!Description: updates the contents of sds_outputlist buffer with stuff the pointers
              point at, for the current line.  Only works with single lines of 1km 
	      data.  Also writes sds_outputlist buffers to the CMG file when necessary.

!Input Parameters:
    k                      offset of current 1km line within 1km buffers
    geo_input              all geolocation data
    local_output           all reflectance output data
    
!Output Parameters: 
    which_file             which file is being written to (0=northpole, 1=nonpolar, 2=southpole,.
                            3=northpole AND nonpole, 4=nonpole AND southpole (borders)
			   -1=error (no latitude values are usable)).

!Input/Output Parameters:  none
    
!Returns: 0     success
          1     serious failure due to memory problem or HDF library read/write problem
	  2     less serious problem with mapping data to the IDN file
	 

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes

   Updates global variable leftlon, as well as currentlyopen and 
      current_n_CMGwindows and sds_outputlist.
   
   Called by routine update_ids(), above.
   
   New (24-JUL-02): The internal cloudmask SDS values in the CMG outputs are
   determined differently.  They used to be determined by simply taking 
   whichever value mapped to the CMG pixel last.  They are now determined by 
   compositing whichever values represent the majority of pixels that map to the 
   CMG pixel.  See below.
!END***************************************************************************
*/
{
int jumpstatus;
jmp_buf jumpbuf;
char location[50]="ids.c, update_coarse_grid\0";
int CMGwindowsize, nextopen, nextmax, nextmin;
int local_currentlyopen, local_current_n_CMGwindows;
float maxlat, minlat, avglat, rows_in_CMG, cols_in_CMG;
int i, j, n_CMGwindows;
float newlon;
double CMGrow, CMGcol, current_row_offset;
int irow, icol, ncol, CMGindex;
float32 *tmpflt[CRSGRDSDSS];
uint8 *n_avg, *ozone;
/*uint16 *tmpICM;*/
uint16 ICM;
uint16 *ICMA;

uint8 *AQAA;  /* 07-JUN-04 */
uint8 AQA;
d_aotqa *tmpAQA;

d_cmask *tmpICM;
uint32 *tmpQA;
int nonpole_index;
static int leftlon_set = 0;
float local_leftlon;
int ilocal_leftlon;
extern int sat_type;           /* 0 for Terra, 1 for Aqua */
float lat_diff;
static float old_lat_diff = 0.0;
int dayside;
int n_should_be_mapped;    /* count of all pixels between +60/-60 latitude that should be mapped per line... */
int n_mapped;              /* ... and the number that actually will be mapped */
int16 *band;
int16 solarzenvalue;
int16 *banda;
uint32 *sqc;
float32 *local_lat;
float32 *local_lon;
uint16 *BT;
int16 hourminute;
char hrmn[5];

/*printf("in %s\n", location);*/

/* 04-AUG-05 */
hrmn[0] = geo_input->xxxxx->rbdt[11];
hrmn[1] = geo_input->xxxxx->rbdt[12];
hrmn[2] = geo_input->xxxxx->rbdt[14];
hrmn[3] = geo_input->xxxxx->rbdt[15];
hrmn[4] = '\0';
hourminute = atoi(hrmn);

local_lat = (float32 *)geo_input->data[0]; 
local_lon = (float32 *)geo_input->data[1];

rows_in_CMG = ((ULLAT-LRLAT)*PIXPERDEG);
cols_in_CMG = ((LRLON-ULLON)*PIXPERDEG);

local_currentlyopen = currentlyopen;
local_current_n_CMGwindows = current_n_CMGwindows;  /* both are global */

CMGwindowsize = (int) DEGPERWIN*PIXPERDEG;  /* lines per window */

/* Get some crude idea of which file will be written to */
nonpole_index = get_max_min(local_lat, k, k+NBSPOTS_1KM, &maxlat, &minlat);

if ((maxlat > 90.1F)||(maxlat < -90.1F)||(minlat > 90.1F)||(minlat < -90.1F)) /* error */
     *which_file = -1;
else if ((maxlat > POLELIM)&&(minlat > POLELIM)) *which_file = 0;
else if ((maxlat > POLELIM)&&(minlat <= POLELIM)) *which_file = 3;    /* 0 and 1 */
else if ((maxlat > -POLELIM)&&(minlat > -POLELIM)) *which_file = 1;   
else if ((maxlat > -POLELIM)&&(minlat <= -POLELIM)) *which_file = 4;  /* 1 and 2 */
else *which_file = 2;

if (*which_file > -1) {
   nextmax = (int)((ULLAT-maxlat)/DEGPERWIN);
   nextmin = (int)((ULLAT-minlat)/DEGPERWIN);
   nextopen = nextmax;
   n_CMGwindows = (nextmin - nextopen)+2;    /* 19-APR-05 */
}
else {    /* error condition */
   nextopen = local_currentlyopen;
   n_CMGwindows = local_current_n_CMGwindows;
}


/* added 09-OCT-02 */
/* 21-APR-05: Note: this "lat_diff" is zero at the end of the scan, because there is 
   no "next line" at the end of a scan.  Earlier versions of the code handled this 
   differently (buffering the latitude from the last scan), but we're not doing that 
   now...  Since "lat_diff" changes only very little from scanline to scanline, just
   use "old_lat_diff" in its place. */
lat_diff = 0.0F; 
if ( (k+NBSPOTS_1KM+(NBSPOTS_1KM/2)) <  NBSPOTS_1KM*NBSCANS_1KM) {
   lat_diff = local_lat[k+NBSPOTS_1KM+(NBSPOTS_1KM/2)] - local_lat[k+(NBSPOTS_1KM/2)]; 
   }
else {
   lat_diff = old_lat_diff;   /* 21-APR-05 */
   }

old_lat_diff = lat_diff;      /* 21-APR-05 */

if (lat_diff < 0.0) {   /* THIS NEEDS EXTREME TESTING!!! */
   if (sat_type == 0)  /* Terra */
       dayside = 1;
   else                /* Aqua */
       dayside = 0;
    }
else {
   if (sat_type == 0)  /* Terra */
       dayside = 0;
   else                /* Aqua */
       dayside = 1;
    }


/*if (!leftlon_set) {
   printf("scan %d SATTYPE = %d, LAT max min: %f %f (%f, %d)", k, sat_type, maxlat, minlat, lat_diff, dayside); 
   if (nonpole_index != -1) printf("***\n");
   else                     printf("\n");
}*/


if ((nonpole_index != -1)&&(!leftlon_set)&&(dayside)) {
   local_leftlon = local_lon[k+(NBSPOTS_1KM/2)];  /* approx. midpoint of scan */
   /*  printf("1 LEFT LON *********************** %f ***************************\n", local_leftlon); */
   if ((local_leftlon > -180.1F) && (local_leftlon < 180.1F)) {
     local_leftlon -= (NONPOLEWIDTH*2.0F)/3.0F;     /* move over to account for tilt */
     local_leftlon -= 10.0F;                        /* move over a bit more (empirical) */
     if (local_leftlon < -180.0) local_leftlon += 360.0F;
     
     ilocal_leftlon = (int)(local_leftlon*PIXPERDEG);
     local_leftlon = ((float)ilocal_leftlon)/PIXPERDEG;   /* 21-APR-05: these try to get the BOUNDINGCOORDINATES straight */
     
     leftlon = local_leftlon;
     /*printf("LEFT LON *********************** %f ***************************\n", local_leftlon); */
     leftlon_set = 1;
     }
   }


/* Read that part of the outputs. */
if(!(jumpstatus=setjmp(jumpbuf))) {
  if ((nextopen != local_currentlyopen)||(n_CMGwindows != local_current_n_CMGwindows)){
     /*printf("Writing at %d, n_lines %d\n", local_currentlyopen*CMGwindowsize, 
        local_current_n_CMGwindows*CMGwindowsize);*/
	     
     /* xfer from sds_outputlist[whatever]->fdata to sds_outputlist[whatever]->data */
     if ((transfer_sds(sds_outputlist)) != 0) longjmp(jumpbuf, 11);
     
     for(i=0;i<CRSGRDSDSS;i++) {
        /* flush the current output buffers */
        if ((get_set_sds(2, sds_outputlist[i], local_currentlyopen*CMGwindowsize, 
	               local_current_n_CMGwindows*CMGwindowsize))!=0)
             longjmp(jumpbuf, 10);
	/* read the new stuff */
        if ((get_set_sds(1, sds_outputlist[i], nextopen*CMGwindowsize, 
	               n_CMGwindows*CMGwindowsize))!=0)
             longjmp(jumpbuf, 10);
       }  
     /*printf("Reading at %d, n_lines %d\n", nextopen*CMGwindowsize, 
         n_CMGwindows*CMGwindowsize); */
	 
     /* xfer data from sds_outputlist[whatever]->data to sds_outputlist[whatever]->fdata */
     if ((transfer_sds(sds_outputlist)) != 0) longjmp(jumpbuf, 11);
          
     local_currentlyopen = nextopen;
     local_current_n_CMGwindows = n_CMGwindows;
   }
}	       
if(jumpstatus==10){
    modsmf(MODIS_E_CMG_IO, "Error reading/writing HDF SDS in CMG.", location);
    return(1);
  }
if(jumpstatus==11){
    modsmf(MODIS_E_CMG_TRANSFER, "Memory error in transfer_sds().", location);
    return(1);
  }
current_row_offset = (double)local_currentlyopen*(double)CMGwindowsize;  

/* Map inputs to outputs */
n_avg = (uint8 *)sds_outputlist[CRSGRDN]->data;
for(i=0;i<CRSGRDN;i++) {
   tmpflt[i] = (float32 *)sds_outputlist[i]->fdata;
   }
tmpICM = (d_cmask *) sds_outputlist[CRSGRDICM]->cmask;
tmpQA  = (uint32 *) sds_outputlist[CRSGRDQA]->data;
tmpAQA  = (d_aotqa *) sds_outputlist[CRSGRDAQA]->aotqa;

/* Remember: all data buffers span from -180 to +180, regardless of which file is
   being written to.  The number of columns of data is thus fixed.  Worry about
   mapping it all to and from HDF files in get_set_sds().
 */
ncol = sds_outputlist[0]->dims[1];
n_should_be_mapped = n_mapped = 0;

/*printf("%p %p\n", local_output->data[OZONE], tmpflt[CRSGRDOZ]);*/

for(j=0;j<NBSPOTS_1KM;j++,k++) {

   CMGrow = ((double)ULLAT - (double)local_lat[k])*(double)PIXPERDEG;
  
   if ((local_lat[k] < POLELIM)&&(local_lat[k] > -POLELIM)) {
       if (dayside) n_should_be_mapped++;
       newlon = local_lon[k] - leftlon;    /* New, 19-DEC-03 */
       if (newlon < 0.0) newlon += 360.0F;
       if ((dayside)&&(newlon < NONPOLEWIDTH)) n_mapped++;
       CMGcol = (double)newlon*(double)PIXPERDEG;
       /*printf("%f %d %d\n", local_lon[k], (int)CMGcol, k);*/
       }
   else {
       CMGcol = ((double)local_lon[k] - (double)ULLON)*(double)PIXPERDEG;
       }
   
   if ((CMGrow >= 0.0)&&(CMGrow < rows_in_CMG)&&   /* catches the crazy values */ 
       (CMGcol >= 0.0)&&(CMGcol < cols_in_CMG)) {      
      irow = (int)((CMGrow - current_row_offset)+0.5);
      icol = (int)(CMGcol+0.5);
      CMGindex = irow*ncol + icol; 
       
      band = geo_input->data[5];
      solarzenvalue = band[k];
      if(solarzenvalue != -32767) {    /* 29-SEP-03: only use if MOD03 data at this point is OK */
      
           for(i=CRSGRDB1;i<CRSGRDTAU1;i++) {
	      band = local_output->data[BAND1_1KM+i];
              tmpflt[i][CMGindex] += (float32)band[k];
	      }
	   
 	   band = local_output->data[AOT_1];
           tmpflt[CRSGRDTAU1][CMGindex] += (float32)band[k];
 	   band = local_output->data[AOT_3];
           tmpflt[CRSGRDTAU3][CMGindex] += (float32)band[k];
 	   band = geo_input->data[5];
           tmpflt[CRSGRDSZ][CMGindex] += (float32)band[k];
 	   band = geo_input->data[3];
           tmpflt[CRSGRDVZ][CMGindex] += (float32)band[k];
 	   band = geo_input->data[6];
 	   banda = geo_input->data[4];
           tmpflt[CRSGRDRA][CMGindex] += 100.0F*(PrincipalAngle((float)band[k]/100.0F - (float)banda[k]/100.0F));
 	   ozone = local_output->data[OZONE];
           tmpflt[CRSGRDOZ][CMGindex] += (float32)ozone[k];

           /* 18-MAY-04 */
	   BT = local_output->data[BAND20];
           tmpflt[CRSGRDB20][CMGindex] += (float32)BT[k];
	   BT = local_output->data[BAND21];
           tmpflt[CRSGRDB21][CMGindex] += (float32)BT[k];
	   BT = local_output->data[BAND31];
           tmpflt[CRSGRDB31][CMGindex] += (float32)BT[k];
	   BT = local_output->data[BAND32];
           tmpflt[CRSGRDB32][CMGindex] += (float32)BT[k];

           /* 04-AUG-05 */
           tmpflt[CRSGRDGTM][CMGindex] += (float32)hourminute;
	
           ICMA = local_output->data[AOTCM];
	   ICM = ICMA[k];
	   if (ICM%2)      tmpICM[CMGindex].cloud++;
	   if (ICM&0x0002) tmpICM[CMGindex].clear++;
	   if (ICM&0x0004) tmpICM[CMGindex].ihc++;
	   if (ICM&0x0008) tmpICM[CMGindex].lwc++;
 	   if (ICM&0x0010) tmpICM[CMGindex].snow++;
 	   if (ICM&0x0020) tmpICM[CMGindex].fire++;
 	   if (ICM&0x0040) tmpICM[CMGindex].glint++;
 	   if (ICM&0x0080) tmpICM[CMGindex].dust++;
 	   if (ICM&0x0100) tmpICM[CMGindex].cldshd++;
 	   if (ICM&0x0200) tmpICM[CMGindex].cldedge++;
 	   if (ICM&0x0c00) tmpICM[CMGindex].cirrus++;
 	   if (ICM&0x1000) tmpICM[CMGindex].panflag++;
	
	   sqc = local_output->data[OKMQA];
	   tmpQA[CMGindex] = sqc[k];
      
	   AQAA = local_output->data[AOTQA];  /* 07-JUN-04: this will have to be improved... */
	   AQA = AQAA[k];

	   tmpAQA[CMGindex].aqa_values[AQA]++;

           n_avg[CMGindex]++;
	   }
    }
}  /* for(j=0;j<NBSPOTS_1KM;j++,k++) */


currentlyopen = local_currentlyopen;
current_n_CMGwindows = local_current_n_CMGwindows;

/*printf("out %s\n", location);*/
if (n_should_be_mapped != n_mapped) return(2);
else return(0);

}

static float32 PrincipalAngle(float32 angle)  /* copied from atmcor.c */
{
#define ONE_OVER_360 2.77777778e-3F
        if (angle < 0.0F) angle = -angle;
        /* only the SGI understands the float floorf function */
#ifdef IRIS4
        angle = angle - 360.0F * floorf(angle * ONE_OVER_360);
#else
        angle = angle - 360.0F * (float)floor((double)(angle * ONE_OVER_360));
#endif
        if (angle > 180.0F) angle = 360.0F - angle;
 
        return angle;
}


void preview_coarse_grid(int k, float *lat, int *which_file)
/*
!C*****************************************************************************
!Description: Reads 1 km resolution latitude in allgeoloc, figures out which 
              CMG file(s) it maps to.

!Input Parameters: 
    k                          where to start reading (1 km) 
    allgeoloc
    
!Output Parameters: 
    which_file                0=northpole, 1=nonpole, 2=southpole; 
                              3=northpole AND nonpole, 4=nonpole AND southpole (borders)
			     -1=error (at least one value in allgeoloc.slatP is bad).

!Input/Output Parameters: none

!Returns: none

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes

      Called by which_ids(), above.
      
!END***************************************************************************
*/
{
float maxlat, minlat;
float32 *local_lat;
/*char location[50]="ids.c, preview_coarse_grid\0";
printf("in %s\n", location);*/

local_lat = (float32 *)lat;  /* 12-JUL-02 */

/* Get some crude idea of which file will be written to */
get_max_min(local_lat, k, k+NBSPOTS_1KM, &maxlat, &minlat);

if ((maxlat > 90.1F)||(maxlat < -90.1F)||(minlat > 90.1F)||(minlat < -90.1F)) /* error */
     *which_file = -1;
else if ((maxlat > POLELIM)&&(minlat > POLELIM)) *which_file = 0;
else if ((maxlat > POLELIM)&&(minlat <= POLELIM)) *which_file = 3;    /* 0 and 1 */
else if ((maxlat > -POLELIM)&&(minlat > -POLELIM)) *which_file = 1;   
else if ((maxlat > -POLELIM)&&(minlat <= -POLELIM)) *which_file = 4;  /* 1 and 2 */
else *which_file = 2;
/*printf("out %s\n", location);*/
return;
}


void end_coarse_grid(d_sds **sds_outputlist, int currentlyopen, int current_n_CMGwindows)
/*
!C*****************************************************************************
!Description: frees all sds_outputlist buffers, closes all SDSs in the CMG
                files.

!Input Parameters: 
    currentlyopen 
    current_n_CMGwindows
    
!Output Parameters: none

!Input/Output Parameters: 
    sds_outputlist         updated

!Returns: none

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
       
       Called by routine clean_up() in file modis_srefl.c.
       
!END***************************************************************************
*/
{
int i, j;
int size;
/*char location[50]="ids.c, end_coarse_grid\0";
printf("in %s\n", location);*/

size = (int) DEGPERWIN*PIXPERDEG;  /* lines per window */

transfer_sds(sds_outputlist);  /* if you've gotten this far, presumably it will work */

for(i=0;i<CRSGRDSDSS;i++) {
   /* flush the current buffer */
   get_set_sds(2, sds_outputlist[i], currentlyopen*size, current_n_CMGwindows*size);
   
   /* free */
   if (sds_outputlist[i]->data != NULL) free(sds_outputlist[i]->data);
   if (sds_outputlist[i]->fdata != NULL) free(sds_outputlist[i]->fdata);
   if (sds_outputlist[i]->cmask != NULL) {
      /* This is not necessary, at least in the standalone version of this routine...
      for(j=0;j<sds_outputlist[i]->currentsize;j++)
          if (sds_outputlist[i]->cmask[j] != NULL) free(sds_outputlist[i]->cmask[j]);*/
      free(sds_outputlist[i]->cmask);
   }
   
   /* end */
   for (j=0;j<NCMGFILES;j++) 
     if(sds_outputlist[i]->sdsid[j] != -1)
       SDendaccess(sds_outputlist[i]->sdsid[j]);
   }
for (j=0;j<NCMGFILES;j++)
   if(sds_outputlist[0]->sdid[j] != -1)
         SDend(sds_outputlist[0]->sdid[j]);
   
for(i=0;i<CRSGRDSDSS;i++) if (sds_outputlist[i] != NULL) free(sds_outputlist[i]);
if (sds_outputlist != NULL) free(sds_outputlist);

/*printf("out %s\n", location);*/
return;
}

int32 get_set_sds(int getsetinit, d_sds *sdsinfo, int line, int n_lines)
/*
!C*****************************************************************************
!Description:     Reads/writes to/from CMG files.  Reads/writes any type of data, to/from
                  void *data in sdsinfo structure.

!Input Parameters: 
    getsetinit             getsetinit = 0,    get memory but do NOT read (init)
                           getsetinit = 1,    get memory (if necessary) and read (SDreaddata)
                           getsetinit = 2,    set (SDwritedata)

    line                   line in the CMG where to start reading/writing
    n_lines                how many lines in the CMG you need to read/write
    
!Output Parameters: none

!Input/Output Parameters: 
    sdsinfo                updated

!Returns: 0                success
         -1                error reading from/writing to SDS
         -2                error figuring out where to read from/write to
	-13                memory error

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
       sdsinfo is a pointer to the following:
             typedef struct{
                 void *data;        --- the CMG/program I/O buffer
                 float32 *fdata;    --- a buffer in which data is summed prior to averaging.
		                        These buffers have a fixed number of columns.  But which
					columns are written to/read from depends on the file.
                 char SDSname[256];
                 int32 dims[2];     --- overall dims of each SDS for each CMG file.  Will
		                        vary by CMG file (e. g., 600x7200 for poles, 2400x1800
					for nonpolar files).  NOT the dims specifically used
					for I/O.
                 int32 sdsid[3];    --- pointer to all possible SDSs this data may be written to
                 int32 sdid[3];     --- pointer to all possible SDs this data may be written to
                 int32 type;      
                 int currentsize;   --- size of the data/fdata buffers IN PIXELS, not bytes. Does
		                        NOT equal dims[0]*dims[1] (currently), need not (ever).
               }d_sds;
	       
       Data is mapped to the sdsinfo buffer "fdata" in a simple linear lat/lon manner (in 
       update_coarse_grid(), above). Buffer "fdata" is used to sum the mapped data as float32's.  
       Routine transfer_sds() (below) is used to average the data in "fdata" and write it to 
       "data" -- and routine transfer_sds() is to be called 
           1) prior to calling this routine for writes  (fdata -> data)
	   2) after calling this routine for reads (data -> fdata). 
	   
       This routine (get_set_sds()) writes/reads to/from the I/O buffer ("data") to whichever 
       sdsid's are necessary.  Which sdsid depends upon argument "line" (line in CMG to start
       reading/writing) and "n_lines" (how many lines to read/write), and is figured out in 
       routine which_and_where() (below).
       
       Both buffers ("data" and "fdata") have a constant number of columns, 
       ((LRLON-ULLON)*PIXPERDEG); currently, (180.0-(-180.0))*20.0 or 7200.  All columns are
       written for polar CMGs; only the first NONPOLEWIDTH*PIXPERDEG columns are written for
       non-polar CMGs (currently, 1800 columns).  Routine update_coarse_grid() takes care of
       mapping pixels for both polar and non-polar CMG files.
       
       Both buffers ("data" and "fdata") have a variable number of lines.  The current number 
       of lines is always currentsize/((LRLON-ULLON)*PIXPERDEG), the future number is argument
       n_lines.  Routine which_and_where() (below) figures out how to map this across CMG files.
       
!END***************************************************************************
*/
{
int32 *i32_array;
uint32 *ui32_array;
float64 *f64_array;
float32 *f32_array;
int8 *i8_array;
uint8 *ui8_array;
int16 *i16_array;
uint16 *ui16_array;
char *char_array;
uchar8 *uchar_array;
int i, count;
int n_sds;
int32 j, nlines, offset;
int32 type, ret, sdss[2], dimms[2][2], sstart[2][2];
int which_and_where(int32 sdss[2], int32 dimms[2][2], int32 sstart[2][2], 
                    int line, int n_lines, int32 n_col);
typedef long (*FUNC)();
FUNC funktion;          /* either SDreaddata() or SDwritedata() */
/*char location[50]="ids.c, get_set_sds\0";
printf("in %s\n", location);*/


/*printf("%cET: Line/n_lines in start %d %d: %d\n", (getsetinit < 2 ? 'G' : 'S'), line, n_lines, n_lines*sdsinfo->dims[1]);
n_sds = which_and_where(sdss, dimms, sstart, line, n_lines, sdsinfo->dims[1]);
printf(" (%d %d)", sstart[0][0], dimms[0][0]);
if (n_sds)printf(" (%d %d) **\n", sstart[1][0], dimms[1][0]);
else      printf("\n");
*/

n_sds = which_and_where(sdss, dimms, sstart, line, n_lines, sdsinfo->dims[1]);

if ((n_sds < 0)||(n_sds > 1)) return(-2);

count = n_lines*sdsinfo->dims[1];
type = sdsinfo->type;
offset = 0;

/* set these RIGHT BEFORE the read */
if (sdss[0] == 1) dimms[0][1] = (int32)(NONPOLEWIDTH*PIXPERDEG);
if ((n_sds) && (sdss[1] == 1)) dimms[1][1] = (int32)(NONPOLEWIDTH*PIXPERDEG);

for(i=0;i<=n_sds;i++) {
    sdss[i] = sdsinfo->sdsid[sdss[i]];
    }
    
if (getsetinit == 0)      funktion = (FUNC)SDreaddata;
else if (getsetinit == 1) funktion = (FUNC)SDreaddata;
else                      funktion = (FUNC)SDwritedata;

ret=0;
  
       switch(type) {
           case DFNT_CHAR:
	      char_array = (char *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          char_array = (char *)realloc(char_array, count*sizeof(char));
		  sdsinfo->currentsize = count;
	             }
	      if (char_array != NULL) {
	         if (getsetinit) {
	           for(i=0;i<=n_sds;i++) { 
		      if (sdss[i] == -1) continue;
		      nlines = dimms[i][0];
		      dimms[i][0] = 1;
		      for(j=0;j<nlines;j++) {
                         ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (char_array+offset));
		         offset += sdsinfo->dims[1];
		         sstart[i][0]++;
		         }
	              }
		   }  
	         sdsinfo->data = (void *)char_array;
		 }
	      else ret = -13;
              break;
           case DFNT_UCHAR:
              uchar_array = (uchar8 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          uchar_array = (uchar8 *)realloc(uchar_array, count*sizeof(uchar8));
		  sdsinfo->currentsize = count;
	             }
	      if (uchar_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (uchar_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
		   }
	         sdsinfo->data = (void *)uchar_array;
		 }
	      else ret = -13;
              break;
           case DFNT_FLOAT32:
	      f32_array = (float *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          f32_array = (float32 *)realloc(f32_array, count*sizeof(float32));
		  sdsinfo->currentsize = count;
	             }
	      if (f32_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (f32_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
	            }
	         sdsinfo->data = (void *)f32_array;
		 }
	      else ret = -13;
              break;
           case DFNT_FLOAT64:
              f64_array = (double *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          f64_array = (float64 *)realloc(f64_array, count*sizeof(float64));
		  sdsinfo->currentsize = count;
	             }
	      if (f64_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (f64_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
	           }
	         sdsinfo->data = (void *)f64_array;
		 }
	      else ret = -13;
              break;
           case DFNT_INT8:
              i8_array = (int8 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          i8_array = (int8 *)realloc(i8_array, count*sizeof(int8));
		  sdsinfo->currentsize = count;
	             }
	      if (i8_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (i8_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
	            }
	         sdsinfo->data = (void *)i8_array;
		 }
	      else ret = -13;
              break;
           case DFNT_UINT8:
              ui8_array = (uint8 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          ui8_array = (uint8 *)realloc(ui8_array, count*sizeof(uint8));
		  sdsinfo->currentsize = count;
	             }
	      if (ui8_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (ui8_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
		    }
	         sdsinfo->data = (void *)ui8_array;
		 }
	      else ret = -13;
              break;
           case DFNT_INT16:
              i16_array = (int16 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          i16_array = (int16 *)realloc(i16_array, count*sizeof(int16));
		  sdsinfo->currentsize = count;
	             }
	      if (i16_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		      if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (i16_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
		    }
	         sdsinfo->data = (void *)i16_array;
		 }
	      else ret = -13;
              break;
           case DFNT_UINT16:
              ui16_array = (uint16 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          ui16_array = (uint16 *)realloc(ui16_array, count*sizeof(uint16));
		  sdsinfo->currentsize = count;
	             }
	      if (ui16_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (ui16_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
		    }
	         sdsinfo->data = (void *)ui16_array;
		 }
	      else ret = -13;
              break;
           case DFNT_INT32:
              i32_array = (int32 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          i32_array = (int32 *)realloc(i32_array, count*sizeof(int32));
		  sdsinfo->currentsize = count;
	             }
	      if (i32_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (i32_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
		    }
	         sdsinfo->data = (void *)i32_array;
		 }
	      else ret = -13;
              break;
           case DFNT_UINT32:
              ui32_array = (uint32 *)sdsinfo->data;
              if ((getsetinit < 2)&& 
                  (count != sdsinfo->currentsize)) {
	          ui32_array = (uint32 *)realloc(ui32_array, count*sizeof(uint32));
		  sdsinfo->currentsize = count;
	             }
	      if (ui32_array != NULL) {
	         if (getsetinit) {
	            for(i=0;i<=n_sds;i++) { 
		       if (sdss[i] == -1) continue;
		       nlines = dimms[i][0];
		       dimms[i][0] = 1;
		       for(j=0;j<nlines;j++) {
                          ret=funktion(sdss[i], sstart[i], NULL, dimms[i], (ui32_array+offset));
		          offset += sdsinfo->dims[1];
		          sstart[i][0]++;
		          }
	               }
		    }
	         sdsinfo->data = (void *)ui32_array;
		 }
	      else ret = -13;
              break;
                         }
/*printf("out %s\n", location);*/
return(ret);
}

void ids_zero_out(d_sds *sdsinfo, int n_lines)
/*
!C*****************************************************************************
!Description: sets all the values in buffer "data" in sdsinfo to zero.

!Input Parameters: 
    n_lines                   how many lines to blank out
    
!Output Parameters: none

!Input/Output Parameters: 
    sdsinfo                   updated

!Returns: none

!Revision History:
	Original version 2001, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
!END***************************************************************************
*/
{
int32 *i32_array;
uint32 *ui32_array;
float64 *f64_array;
float32 *f32_array;
int8 *i8_array;
uint8 *ui8_array;
int16 *i16_array;
uint16 *ui16_array;
char *char_array;
uchar8 *uchar_array;
int i,count;
int32 ret,id,type;
/*char location[50]="ids.c, ids_zero_out\0";
printf("in %s\n", location);*/

count = n_lines*sdsinfo->dims[1];
type = sdsinfo->type;

       switch(type) {
           case 4:
              char_array = (char *)sdsinfo->data;
	      memset(char_array, '\0', count*sizeof(char));
              break;
           case 3:
              uchar_array = (uchar8 *)sdsinfo->data;
 	      memset(uchar_array, '\0', count*sizeof(uchar8));
              break;
           case 5:
              f32_array = (float *)sdsinfo->data;
 	      memset(f32_array, '\0', count*sizeof(float));
              break;
           case 6:
              f64_array = (double *)sdsinfo->data;
 	      memset(f64_array, '\0', count*sizeof(double));
              break;
           case 20:
              i8_array = (int8 *)sdsinfo->data;
 	      memset(i8_array, '\0', count*sizeof(int8));
              break;
           case 21:
              ui8_array = (uint8 *)sdsinfo->data;
  	      memset(ui8_array, '\0', count*sizeof(uint8));
              break;
           case 22:
              i16_array = (int16 *)sdsinfo->data;
  	      memset(i16_array, '\0', count*sizeof(int16));
              break;
           case 23:
              ui16_array = (uint16 *)sdsinfo->data;
  	      memset(ui16_array, '\0', count*sizeof(uint16));
              break;
           case 24:
              i32_array = (int32 *)sdsinfo->data;
  	      memset(i32_array, '\0', count*sizeof(int32));
              break;
           case 25:
              ui32_array = (uint32 *)sdsinfo->data;
  	      memset(ui32_array, '\0', count*sizeof(uint32));
              break;
                         }

/*printf("out %s\n", location);*/
return;
}

int transfer_sds(d_sds **sds_output)
/*
!C*****************************************************************************
!Description: Transfers data to/from buffer "data" from/to buffer "fdata" (both
              in structure sds_output), with averaging (if fdata -> data) and
	      "unaveraging" (if data -> fdata).  Transfers from whichever one 
	      is nonNULL to whichever one is NULL.

!Input Parameters: none
    
!Output Parameters: none

!Input/Output Parameters: 
    sds_output                  updated

!Returns: 0  success
          1  either both or neither buffer is NULL
	 -1  memory problem

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
     Designed such that the last sds_output structure in the array holds the N in its
     "data" buffer, and this N is to be applied to the averaging of the data in each
     of the other sds_output structure's "fdata" buffer values.  
    
     New (24-JUL-02), implemented transfers between the "cmask" buffer and the "data"
     buffer for the internal cloud mask SDS; not with averaging, but with taking whatever
     value is in the majority.
!END***************************************************************************
*/
{
int32 *i32_array;
uint32 *ui32_array;
float64 *f64_array;
float32 *f32_array;
int8 *i8_array;
uint8 *ui8_array;
int16 *i16_array;
uint16 *ui16_array;
char *char_array;
uchar8 *uchar_array;
d_cmask *cmask_array;
d_aotqa *aotqa_array;

int32 *i32_array1;
uint32 *ui32_array1;
float64 *f64_array1;
float32 *f32_array1;
int8 *i8_array1;
uint8 *ui8_array1;
int16 *i16_array1;
uint16 *ui16_array1;
char *char_array1;
uchar8 *uchar_array1;
d_cmask *cmask_array1;
d_aotqa *aotqa_array1;

int32 i, j, type, count;
uint8 *n_avg;
uint8 n;
int8 tmpqa, tmpqaindex, tmpaqa;
uint16 tmpcld;
int fdata_is_null, data_is_null;
/*char location[50]="ids.c, transfer_sds\0";
printf("in %s\n", location);*/

count = sds_output[0]->currentsize;

/*for(i=0;i<CRSGRDN;i++) 
   printf("******* %p %p\n", sds_output[i]->fdata, sds_output[i]->data);
*/

fdata_is_null = (sds_output[0]->fdata == NULL);
data_is_null = (sds_output[0]->data == NULL);

if ((!fdata_is_null) && (data_is_null)) {
/* Then transfer data from fdata to data, with the required averaging */
    n_avg = sds_output[CRSGRDN]->data; 
    
    /* experimental, 23-JUL-02 */
    /* Assuming ICM is uint16 */
    cmask_array = (d_cmask *)sds_output[CRSGRDICM]->cmask;
    aotqa_array = (d_aotqa *)sds_output[CRSGRDAQA]->aotqa;
    if ((ui16_array = (uint16 *)calloc(count,sizeof(uint16))) == NULL) return(-1);
    if ((ui8_array = (uint8 *)calloc(count,sizeof(uint8))) == NULL) return(-1);
    for(j=0;j<count;j++) {
       n = n_avg[j]/2;
       tmpcld = 0;
       ui8_array[j] = AOTAQA_FILL;
       if (n_avg[j]) {
          if (cmask_array[j].cloud > n)tmpcld |= 1;  /* if the majority are cloudy, flag this as cloudy... */
          if (cmask_array[j].clear > n)tmpcld |= 2;  /* if the majority are clear, flag this as clear... */
          if (cmask_array[j].ihc > n)  tmpcld |= 4;
          if (cmask_array[j].lwc > n)  tmpcld |= 8;
          if (cmask_array[j].snow > n) tmpcld |= 16;
          if (cmask_array[j].fire > n) tmpcld |= 32;
          if (cmask_array[j].glint > n)tmpcld |= 64;
          if (cmask_array[j].dust > n) tmpcld |= 128;
          if (cmask_array[j].cldshd > n) tmpcld |= 256;
          if (cmask_array[j].cldedge > n) tmpcld |= 512;
          if (cmask_array[j].cirrus > n) tmpcld |= 3072;
          if (cmask_array[j].panflag > n) tmpcld |= 4096;
	  ui16_array[j] = tmpcld;
	  
          tmpqa = -1;
          tmpqaindex = -1;
	  for(i=0;i<N_AOTQA_VALUES+1;i++) {
	     tmpaqa = (int8)(aotqa_array[j].aqa_values[i]);
	     if (( tmpaqa > tmpqa ) && ( tmpaqa != AOTAQA_FILL )){
	         tmpqa = tmpaqa;
                 tmpqaindex = i;
		 }
	     }
	  if (tmpqaindex > -1) ui8_array[j] = (uint8)tmpqaindex;
	  }
         }
    sds_output[CRSGRDICM]->data = (void *)ui16_array;
    sds_output[CRSGRDAQA]->data = (void *)ui8_array;
    free(sds_output[CRSGRDICM]->cmask);
    free(sds_output[CRSGRDAQA]->aotqa);
    sds_output[CRSGRDICM]->cmask = (void *)NULL;
    sds_output[CRSGRDAQA]->aotqa = (void *)NULL;
    /* end, experimental, 23-JUL-02 */
       
    for(i=0;i<CRSGRDQA;i++) {   /* UP TO BUT NOT INCLUDING QA/ICM!!! */
        type = sds_output[i]->type;
        f32_array1 = (float32 *)sds_output[i]->fdata;
        switch(type) {
           case 4:
	      if ((char_array = (char *)calloc(count,sizeof(char))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             char_array[j] = (char)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)char_array;
              break;
           case 3:
	      if ((uchar_array = (uchar8 *)calloc(count,sizeof(uchar8))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             uchar_array[j] = (uchar8)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)uchar_array;
              break;
           case 5:
	      if ((f32_array = (float32 *)calloc(count,sizeof(float32))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             f32_array[j] = (float32)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)f32_array;
              break;
           case 6:
	      if ((f64_array = (float64 *)calloc(count,sizeof(float64))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             f64_array[j] = (float64)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)f64_array;
              break;
           case 20:
              if ((i8_array = (int8 *)calloc(count,sizeof(int8))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             i8_array[j] = (int8)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)i8_array;
	      break;
           case 21:
              if ((ui8_array = (uint8 *)calloc(count,sizeof(uint8))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             ui8_array[j] = (uint8)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)ui8_array;
	      break;
           case 22:
              if ((i16_array = (int16 *)calloc(count,sizeof(int16))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             i16_array[j] = (int16)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)i16_array;
	      break;
           case 23:
              if ((ui16_array = (uint16 *)calloc(count,sizeof(uint16))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             ui16_array[j] = (uint16)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)ui16_array;
	      break;
           case 24:
              if ((i32_array = (int32 *)calloc(count,sizeof(int32))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             i32_array[j] = (int32)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)i32_array;
	      break;
           case 25:
              if ((ui32_array = (uint32 *)calloc(count,sizeof(uint32))) == NULL) return(-1);
	      for(j=0;j<count;j++) {
	         if (n_avg[j])
	             ui32_array[j] = (uint32)(f32_array1[j]/(float)n_avg[j]);
		  }
	      sds_output[i]->data = (void *)ui32_array;
	      break;
	   }
    /*free(f32_array1);*/
    free(sds_output[i]->fdata);
    sds_output[i]->fdata = (void *)NULL;
    }
    return(0);
}
else if ((fdata_is_null) && (!data_is_null)) {
/* Then transfer data from "data" to "fdata", and set "data" to NULL; */
    n_avg = sds_output[CRSGRDN]->data; 
    
    /* experimental, 23-JUL-02 */
    /* Assuming ICM is uint16 */
    if ((sds_output[CRSGRDICM]->cmask = (d_cmask *)malloc(count*sizeof(d_cmask))) == NULL) { 
	return(-1);
	}
    if ((sds_output[CRSGRDAQA]->aotqa = (d_aotqa *)malloc(count*sizeof(d_aotqa))) == NULL) { 
	return(-1);
	}
    cmask_array1 = (d_cmask *)sds_output[CRSGRDICM]->cmask;
    aotqa_array1 = (d_aotqa *)sds_output[CRSGRDAQA]->aotqa;
    ui16_array = (uint16 *)sds_output[CRSGRDICM]->data;
    ui8_array = (uint8 *)sds_output[CRSGRDAQA]->data;
    for(j=0;j<count;j++) {
       for(i=0;i<N_AOTQA_VALUES+1;i++) aotqa_array1[j].aqa_values[i] = 0;
       cmask_array1[j].cloud = 0;
       cmask_array1[j].clear = 0;
       cmask_array1[j].ihc = 0;
       cmask_array1[j].lwc = 0;
       cmask_array1[j].snow = 0;
       cmask_array1[j].fire = 0;
       cmask_array1[j].glint = 0;
       cmask_array1[j].dust = 0;
       cmask_array1[j].cldshd = 0;
       cmask_array1[j].cldedge = 0;
       cmask_array1[j].cirrus = 0;
       cmask_array1[j].panflag = 0;

       tmpcld = ui16_array[j];
       tmpqa = ui8_array[j];
       n = n_avg[j]/2 + 1;  /* the majority... */
       if (n_avg[j]) {
          if (tmpcld%2)      cmask_array1[j].cloud = n;  /* ...are set to cloudy */
          if (tmpcld&0x0002) cmask_array1[j].clear = n;  /* ...are set to clear, etc. */
          if (tmpcld&0x0004) cmask_array1[j].ihc = n;
          if (tmpcld&0x0008) cmask_array1[j].lwc = n;
          if (tmpcld&0x0010) cmask_array1[j].snow = n;
          if (tmpcld&0x0020) cmask_array1[j].fire = n;
          if (tmpcld&0x0040) cmask_array1[j].glint = n;
          if (tmpcld&0x0080) cmask_array1[j].dust = n;
          if (tmpcld&0x0100) cmask_array1[j].cldshd = n;
          if (tmpcld&0x0200) cmask_array1[j].cldedge = n;
          if (tmpcld&0x0c00) cmask_array1[j].cirrus = n;
          if (tmpcld&0x1000) cmask_array1[j].panflag = n;
          
	  if (tmpqa <= N_AOTQA_VALUES) 
	     aotqa_array1[j].aqa_values[tmpqa] = n;
	  else 
	     for(i=0;i<N_AOTQA_VALUES+1;i++) 
	        aotqa_array1[j].aqa_values[i] = AOTAQA_FILL;

	  }

         }
    free(ui16_array);
    free(ui8_array);
    sds_output[CRSGRDICM]->data = (void *)NULL;
    sds_output[CRSGRDAQA]->data = (void *)NULL;
    /* end, experimental, 23-JUL-02 */
       
    for(i=0;i<CRSGRDQA;i++) {
        type = sds_output[i]->type;
        if ((sds_output[i]->fdata = (float32 *)calloc(count,sizeof(float32))) == NULL) { 
	    return(-1);
	    }
        f32_array1 = sds_output[i]->fdata;
        switch(type) {
          case 4:
              char_array = (char *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(char_array[j]*n_avg[j]);
              free(char_array);
              break;
           case 3:
              uchar_array = (uchar8 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(uchar_array[j]*n_avg[j]);
              free(uchar_array);
              break;
           case 5:
              f32_array = (float32 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(f32_array[j]*n_avg[j]);
              free(f32_array);
              break;
           case 6:
              f64_array = (float64 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(f64_array[j]*n_avg[j]);
              free(f64_array);
              break;
           case 20:
              i8_array = (int8 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(i8_array[j]*n_avg[j]);
              free(i8_array);
	      break;
           case 21:
              ui8_array = (uint8 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(ui8_array[j]*n_avg[j]);
              free(ui8_array);
	      break;
           case 22:
              i16_array = (int16 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(i16_array[j]*n_avg[j]);
              free(i16_array);
	      break;
           case 23:
              ui16_array = (uint16 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(ui16_array[j]*n_avg[j]);
              free(ui16_array);
	      break;
           case 24:
              i32_array = (int32 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(i32_array[j]*n_avg[j]);
              free(i32_array);
	      break;
           case 25:
              ui32_array = (uint32 *)sds_output[i]->data;
	      for(j=0;j<count;j++) 
	             f32_array1[j] = (float32)(ui32_array[j]*n_avg[j]);
              free(ui32_array);
	      break;
           }
       /*free(sds_output[i]->data);*/
       sds_output[i]->data = (void *)NULL;
       /*sds_output[i]->fdata = (void *)f32_array1;*/
   }
   /*printf("out %s\n", location);*/
   return(0);
}
else {
   /*printf("out %s\n", location);*/
   return(1);	
   }		 
}

int get_max_min(float *array, int start, int length, float *max, float *min)
/*
!C*****************************************************************************
!Description:   Gets the maximum/minimum of values in "array".  Meant for getting 
                max/min of latitude values. 
		 
!Input Parameters: 
     array                     any array of floating point values
     start                     where to start looking for max/min
     length                    start + how many data values to look at
    
!Output Parameters: 
     max                       max and
     min                               min of whatever is in array - but within 
                               limits meaningful for latitude (+/- 90.0F).  If
			       no value in array is within those limits, routine
			       returns -999.0 for both.

!Input/Output Parameters: none


!Returns: Returns index of last value in "array" that is between +/-POLELIM 
          (+/-60.0 degrees), or -1.  (If "array" does not contain latitude,
	  return value won't mean anything.)

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
    
!END***************************************************************************
*/
{
int i, ret, maxminset;
float lmax, lmin;
float tmp;
/*char location[50]="ids.c, get_max_min\0";
printf("in %s\n", location);*/

ret = -1;
lmax = -999.0F;
lmin = -999.0F;
maxminset = 0;

for (i=start;i<length;i++){
  tmp = array[i];
  if ((tmp > -90.1F) && (tmp < 90.1F)) {
     if (!maxminset) {
        lmax = tmp;
	lmin = tmp;
	maxminset = 1;
        }
     if (maxminset) {
        if (tmp<lmin)lmin=tmp;
        if (tmp>lmax)lmax=tmp;
        if ((tmp<POLELIM)&&(tmp>-POLELIM))ret=i;
	}
     }
   }
*max = lmax;
*min = lmin;

/*printf("out %s\n", location);*/
return(ret);
}

int which_and_where(int32 sdss[2], int32 dimms[2][2], int32 sstart[2][2], 
                    int line, int n_lines, int32 n_col)
/*
!C*****************************************************************************
!Description:   Figures out which CMG file to write to, sets "dimms" and "sstart"
                accordingly.
		 
!Input Parameters: 
    line                   line in the CMG where to start reading/writing
    n_lines                how many lines in the CMG you need to read/write
    n_col                  number of columns to put into dimms[*][1] (will be
                                 overwritten later).
    
!Output Parameters: 
     sdss                  array of CMG files to access (read/write):
                                    0=northpole, 1=non-polar, 2=southpole.
				    NOT sds values, but pointer into 
				    sds_info->sdsid[*] array.
     dimms                 where to end accessing each line/col in the sds
     sstart                where to start accessing each line/col in the sds
     

!Input/Output Parameters: none


!Returns: the length of the array "sdss" (success), or 
           -1 (error)

!Revision History:
	Original version Feb. 2002, J. Ray (jim@cratmos.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Jim Ray
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		jim@cratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
   This routine is meant to map lines from data space to output file space.  In the
   data "line" will vary from 0 to 3400, and "n_lines" can be up to ~600.  In the output 
   files, the North Pole file (sds index 0) takes lines 0 to 599, the "middle" file 
   (sds index 1) takes 600 to 2999, and the South Pole file (sds index 2) takes from 
   3000 to 3599 (all at POLELIM = 60.0, PIXPERDEG = 20.0, ULLAT = 90.0, LRLAT = -90.0). 

!END***************************************************************************
*/
{ 
int n_sds, endline, maxlines, endofN, startofS;
char location[50] = "ids.c, which_and_where\0";
char msg[256];
 
maxlines = (int)((ULLAT-LRLAT)*PIXPERDEG);
endofN = (int)((ULLAT-POLELIM)*PIXPERDEG);
startofS = (int)((2*POLELIM)*PIXPERDEG) + endofN;
 
/* 28-APR-05 */
endline = line+n_lines;
if (endline > maxlines) {
   n_lines = maxlines - line;
   endline = maxlines;
   }
 
if ((line < 0) || (endline < 0) || (n_lines < 0) ||
    (line > maxlines) || (endline > maxlines)){
   sprintf(msg, "problem with variables 'line' (%d) or 'n_lines' (%d)", line, n_lines);
   modsmf((PGSt_SMF_code)MODIS_E_FUNCTION_ERROR, msg, location);
   /*printf("out %s\n", location);*/
   return(-1);
   }
n_sds = 0;
sstart[0][1] = 0;
sstart[1][1] = 0;
dimms[0][1] = n_col;
dimms[1][1] = n_col;

if (line < endofN) {
   sdss[n_sds] = 0;
   sstart[0][0] = line;
   if (endline >= endofN) {
      dimms[0][0] = endofN - line;
      n_sds++;
      sdss[n_sds] = 1;
      sstart[1][0] = 0;
      dimms[1][0] = endline - endofN;
      }
   else {
      dimms[0][0] = n_lines;
      }
   }
else if ((line >= endofN) && (line < startofS)) {
   sdss[n_sds] = 1;
   sstart[0][0] = line - endofN; 
   if (endline >= startofS) {
      dimms[0][0] = startofS - line;
      n_sds++;
      sdss[n_sds] = 2;
      sstart[1][0] = 0;
      dimms[1][0] = endline - startofS;
      }
   else {
      dimms[0][0] = n_lines;
      }
   }
else {
   sdss[n_sds] = 2;
   sstart[0][0] = line - startofS; 
   dimms[0][0] = n_lines;
   }

if (n_sds > 1) {
   modsmf(MODIS_F_CODING_ERROR, "\0", location);
   }
/*printf("out %s\n", location);*/
return(n_sds);

}
