/****************************************************************************
!C

!File: modis_srefl_io.c

!Description:   Routines for taking care of I/O for MOD_PR09.exe.

!Input Parameters: none
    
!Output Parameters: none

!Input/Output Parameters: none

!Returns: nothing

!Revision History:

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

    MODIS Science Team Members:
      Chris Justice
      MODIS Land Science Team           University of Maryland
      justice@kratmos.gsfc.nasa.gov     NASA/GSFC Code 923
      phone: 301-286-7372               Greenbelt, MD 20771

      Eric F Vermote
      Univerisity of MD.
      NASA's GSFC Code 923
      Greenbelt, MD , 20771
      eric@kratmos.gsfc.nasa.gov
      301-286-6232

      Jim Ray
      Science Systems and Applications Inc.
      NASA's Goddard Space Flight Center Code 923
      Greenbelt MD, 20771
      jim@kratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes
     
     Both inputs and outputs are represented by a single structure: dims_struct. 

*	typedef struct {
*	  void *data[ALLBANDS];
*	  char *filename;
*	  int OK;                         -1 = uninit, 0 = OK, 1 = error 
*	  int32 nblayers,nblines,nbcols;
*	  int32 ver, n_sds;
*	  int32 sds_offset[ALLBANDS];     -- for 3D SDSs, the number of the layer the parameter is found in
*	  int32 band_name[ALLBANDS];      -- for 3D SDSs, the actual identity of that layer 
*	  float32 radscale[ALLBANDS], radoffset[ALLBANDS];
*	  float32 refscale[ALLBANDS], refoffset[ALLBANDS];
*	  int32 lun, sdid, sdsid[ALLBANDS];
*	  int32 type[ALLBANDS];
*	  input_metadata *xxxxx;
*	} dims_struct;

     There is one dims_struct variable for each file being read from or written to -- NOT
     INCLUDING the IDS files, which require geolocating, and thus get special handling 
     (see ids.c).  
     
     Routines in this file do not access or set any global variables.
     
     Actual SDSreads and SDSwrites are done in routines get_a_scan() and write_a_scan()
     in file atmcor.c.  

!END***************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "modis_srefl.h"
#include "HdfEosDef.h"
#include "protos.h"
#ifdef SUBSET
char method[256];
#endif
int setup_inputs(dims_struct ****inputs, int *n_in, int *input_luns)
/*
!C****************************************************************************

!Routine: setup_inputs()

!Description: Allocates memory for the array inputs and for each structure in 
              the array inputs, and fills each structure with information 
	      about the inputs.

!Input Parameters:
  int *n_in               The number of each type of input file
  int *input_luns         The LUNs for each type of input file

!Input/Output Parameters:
  dims_struct ****inputs  Pointer to a 2-D array of input structures (updated)

!Output Parameters:
  none

!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
         called by routine main() (modis_srefl.c)

!END**************************************************************************/
{
int i, j, maxn, n_OK[ALL_L1B_FILES];
PGSt_integer n, n_inputs;
char msg[256];
char location[100]="modis_srefl_io.c, setup_inputs\0";

maxn = n_in[0];
n_OK[0] = 0;
for(i=1;i<ALL_L1B_FILES;i++) {
   if (n_in[i] > maxn) maxn = n_in[i];
   n_OK[i] = 0;
   }

/* get memory for the input structures, only  */
get_input_str_2D_memory(inputs, ALL_L1B_FILES, maxn);

for(j=0;j<ALL_L1B_FILES;j++) {
   for(i=0;i<n_in[j];i++) {  
     /* get all necessary metadata from each file  */
     if(get_info(input_luns[j], i+1, (*inputs)[j][i])!= 0) {
         sprintf(msg, "Check file at LUN %d, version %d", (int32)input_luns[j], i+1);
         modsmf((PGSt_SMF_code)MODIS_E_FILE_ACCESS, msg, location);
        }
	if ( (*inputs)[j][i]->OK == 0  ) n_OK[j]++;
     }
  }

/*if ( (n_OK[RAD500_FILE] == 0) || (n_OK[RAD1_FILE] == 0) || (n_OK[GEO_FILE] == 0) ) {
     sprintf(msg, "for either 500m files (%d read successfully), 1km files (%d read successfully) or geolocation files (%d read successfully),"
                  " check the files...",  n_OK[RAD500_FILE], n_OK[RAD1_FILE], n_OK[GEO_FILE]);
     modsmf(MODIS_F_GETFILEINFO, msg, location);
   }
 23-MAR-04: 
   */
if ( (n_OK[RAD1_FILE] == 0) || (n_OK[GEO_FILE] == 0) ) {
     sprintf(msg, "for either 1km files (%d read successfully) or geolocation files (%d read successfully),"
                  " check the files...",  n_OK[RAD1_FILE], n_OK[GEO_FILE]);
     modsmf(MODIS_F_GETFILEINFO, msg, location);
   }

return(0);
}

void get_input_str_2D_memory(dims_struct ****inputs, int dim1, int dim2)
/*
!C****************************************************************************

!Routine: get_input_str_2D_memory()

!Description: Allocates memory for the array inputs and for each structure in 
              the array inputs.  Initializes each structure also.

!Input Parameters:
   dim1                   first dim
   dim2                   second dim

!Input/Output Parameters:
  dims_struct ****inputs  Pointer to a 2-D array of input structures (updated)
                          Dimensions: dim1 by dim2.

!Output Parameters:
  none

!Returns:
  nothing
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
     All errors are fatal memory errors.
     called by routine sort_inputs() (modis_srefl_io.c)
     called by routine setup_inputs() (modis_srefl_io.c)

!END**************************************************************************/
{
int i, j;
char msg[256];
char location[100]="modis_srefl_io.c, get_input_str_2D_memory\0";

/* allocate one dimension of dims_struct's */
if (( (*inputs) = (dims_struct ***)malloc(dim1*sizeof(dims_struct **))) == NULL) {
    sprintf(msg, "%d bytes", dim1*sizeof(dims_struct **));
    modsmf(MODIS_F_MEM_FAILURE, msg, location);
  }
/*printf("%d bytes, %p, %s\n", dim1*sizeof(dims_struct **), *inputs, location);*/

for(j=0;j<dim1;j++) {

   /* get memory for another dimension of dims_struct's */
   if (( (*inputs)[j] = (dims_struct **)malloc(dim2*sizeof(dims_struct *))) == NULL) {
      sprintf(msg, "%d bytes", dim2*sizeof(dims_struct *));
      modsmf(MODIS_F_MEM_FAILURE, msg, location);
     }
   /*printf("%d * %d bytes, %p, %s\n",  dim2, dim2*sizeof(dims_struct *), (*inputs)[j], location);*/

   for(i=0;i<dim2;i++) {   /* and for each... */

     /* get memory for each dims_struct */
     if (( (*inputs)[j][i] = (dims_struct *)malloc(sizeof(dims_struct))) == NULL) {
         sprintf(msg, "%d bytes", sizeof(dims_struct));
         modsmf(MODIS_F_MEM_FAILURE, msg, location);
       }
     /*printf("%d bytes, %p, %s\n", sizeof(dims_struct), (*inputs)[j][i], location);*/
    
     /* initialize each structure */
     init_struct((*inputs)[j][i], location);
     /*printf("%d %d init_struct %p %p\n",  j,i,(*inputs)[j][i]->xxxxx, (*inputs)[j][i]);*/

       }
     }
}


void get_input_str_1D_memory(dims_struct ***inputs, int dim1)
/*
!C****************************************************************************

!Routine: get_input_str_1D_memory()

!Description: Allocates memory for the array inputs and for each structure in 
              the array inputs.  Initializes each structure also.

!Input Parameters:
  int dim1                the dimension

!Input/Output Parameters:
  dims_struct ****inputs  Pointer to a 2-D array of input structures (updated)
                          Dimension: dim1.

!Output Parameters:
  none

!Returns:
  nothing
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
     All errors are fatal memory errors.

!END**************************************************************************/
{
int i, j;
char msg[256];
char location[100]="modis_srefl_io.c, get_input_str_1D_memory\0";

/* allocate one dimension of dims_struct's */
if (( (*inputs) = (dims_struct **)malloc(dim1*sizeof(dims_struct *))) == NULL) {
    sprintf(msg, "%d bytes", dim1*sizeof(dims_struct *));
    modsmf(MODIS_F_MEM_FAILURE, msg, location);
  }
/*printf("%d bytes, %p, %s\n", dim1*sizeof(dims_struct *), *inputs, location);*/

for(j=0;j<dim1;j++) {

   /* get memory for another dimension of dims_struct's */
   if (( (*inputs)[j] = (dims_struct *)malloc(sizeof(dims_struct))) == NULL) {
      sprintf(msg, "%d bytes", sizeof(dims_struct *));
      modsmf(MODIS_F_MEM_FAILURE, msg, location);
     }
   /*printf("%d * %d bytes, %p, %s\n",  1, sizeof(dims_struct *), (*inputs)[j], location);*/

   /* initialize each structure */
   init_struct((*inputs)[j], location);
   /*printf("%d %d init_struct %p %p\n",  j,i,(*inputs)[j]->xxxxx, (*inputs)[j]);*/

     }
}


int sort_inputs(dims_struct ****inputs, int *n_in)
{
/*
!C****************************************************************************
!Routine: sort_inputs()

!Description: Sorts the inputs into temporal order.

!Input Parameters:
   none
   
!Input/Output Parameters:
  int *n_in             INPUT: number of inputs (dimension ALL_L1B_FILES)
                        OUTPUT: number of inputs (dimension ALL_TIMES)
  dims_struct *inputs   Pointer to the array of input structures (sorted)
 
!Output Parameters:
  none
 
!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  06-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
           called by routine main() (modis_srefl.c)

!END**************************************************************************/
dims_struct ***sorted_inputs;
int i,j, ii, maxn, ret, n_unique_tais;
extern PGSt_double orbit_start_tai, orbit_end_tai;   /* read already in modis_srefl.c, init_processing() */
PGSt_double granule_start_tai, granule_tai, granule_end_tai;
PGSt_double *all_granule_start_tais;       /* all the UNIQUE granule start tais */
char location[100]="modis_srefl_io.c, sort_inputs\0";
char configParamVal[PGSd_PC_VALUE_LENGTH_MAX] = {'\0'};

maxn = n_in[0];
for(i=1;i<ALL_L1B_FILES;i++) 
   if (n_in[i] > maxn) maxn = n_in[i];
   
/* Get earliest granule start time, last granule end time from the input metadata.
   Make sure everything is within the orbit start, end times  */
ret = get_granule_times(inputs, n_in, orbit_start_tai, orbit_end_tai, &granule_start_tai, &granule_end_tai);
/*printf("%d %f %f\n", ret, granule_start_tai, granule_end_tai);*/

if (ret == -1) {
   modsmf(MODIS_F_NONCOMPT_INPUT, ": no input files are appropriate for processing.", location);
   }
   

/* Get a sorted list of all unique granule start times */
all_granule_start_tais = (PGSt_double *)AllocMem(maxn*ALL_L1B_FILES*sizeof(PGSt_double), location);
ret = get_unique_granule_times(inputs, n_in, &all_granule_start_tais, &n_unique_tais);

/*for(i=0;i<n_unique_tais;i++) printf("%p -----------> ** %d %f \n", all_granule_start_tais,i, all_granule_start_tais[i]);*/

/* Now, make ANOTHER 2-D array of dims_struct structures ... */
get_input_str_2D_memory(&sorted_inputs, (int)ALL_L1B_FILES, n_unique_tais);

/*for(i=0;i<n_unique_tais;i++) printf("%p -----------> ** %d %f \n", all_granule_start_tais,i, all_granule_start_tais[i]);*/

/* and copy inputs into sorted_inputs according to time */
for(j=0;j<n_unique_tais;j++) {
  for(i=0;i<ALL_L1B_FILES;i++) {
    for(ii=0;ii<n_in[i];ii++) {
 
 /*printf("1 ** %d %d %s %s %s\n", i,ii,(*inputs)[i][ii]->filename, (*inputs)[i][ii]->xxxxx->daynightflag, (*inputs)[i][ii]->xxxxx->rbdt);*/
 /*printf("2 %d %f %f --> %d\n", j, (*inputs)[i][ii]->xxxxx->rbdt_tai, all_granule_start_tais[j],  ((*inputs)[i][ii]->xxxxx->rbdt_tai ==
         all_granule_start_tais[j]) );*/
 
       if ((*inputs)[i][ii]->xxxxx->rbdt_tai == all_granule_start_tais[j]) {
           structcpy( (sorted_inputs[i][j]), ((*inputs)[i][ii]) );
	   }
 
 /*printf("3 ** %d %d %d %s %s %s\n", j,i,ii,sorted_inputs[i][j]->filename, sorted_inputs[i][j]->xxxxx->daynightflag, sorted_inputs[i][j]->xxxxx->rbdt);*/
  
         }
      }
   }
free(all_granule_start_tais);

/* and free old input structures to prevent memory leaks... */
close_io_struct((*inputs), ALL_L1B_FILES,  maxn);

/* when I substitute the new one for the old. */
(*inputs) = sorted_inputs;    

for(i=0;i<ALL_L1B_FILES;i++) 
   n_in[i] = n_unique_tais;

return(0);
}

int get_unique_granule_times(dims_struct ****inputs, int *n_in, PGSt_double **all_granule_start_tais, 
                             int *n_unique_tais)
/*
!C****************************************************************************
!Routine: get_unique_granule_times()

!Description: Gets all unique granule start times AND SORTS THEM INTO TEMPORAL ORDER.

!Input Parameters:
  int *n_in                             number of inputs (dimension ALL_L1B_FILES)
  dims_struct *inputs                   Pointer to the array of input structures 
    
!Input/Output Parameters:
  none
 
!Output Parameters:
  PGSt_double *all_granule_start_tais   Pointer to the array of granule start TAIs (sorted)
  int *n_unique_tais                    number of unique TAI values.
  
!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  06-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
        called by routine sort_inputs() (modis_srefl_io.c)

!END**************************************************************************/
{
int i,j, ii, n_u, unique;

*n_unique_tais = 0;
n_u = 0;

for(i=0;i<ALL_L1B_FILES;i++) {
  for(j=0;j<n_in[i];j++) {
      if (n_u == 0) { 
          if ((*inputs)[i][j]->OK == 0)
	      (*all_granule_start_tais)[n_u++] = (*inputs)[i][j]->xxxxx->rbdt_tai;
         }
      else {
         unique = 1;
         for(ii=0;ii<n_u;ii++) 
	    if ( ((*inputs)[i][j]->OK == 0) &&
	       ( (*all_granule_start_tais)[ii] == (*inputs)[i][j]->xxxxx->rbdt_tai )) unique = 0;
	 if (unique) 
            if ((*inputs)[i][j]->OK == 0)
              (*all_granule_start_tais)[n_u++] = (*inputs)[i][j]->xxxxx->rbdt_tai;
         }
   }
}

/* Sort all the unique granule start times into temporal order
 * NOTE: for dmalloc processing, must comment this out or somehow work-around it;
 * dmalloc doesn't seem to understand realloc.   */
#ifndef DEBUG_MALLOC
*all_granule_start_tais = (PGSt_double *)realloc(*all_granule_start_tais, n_u*sizeof(PGSt_double));
#endif 
unique = n_u;
qsort(*all_granule_start_tais, unique, sizeof(PGSt_double), compare_tai);
*n_unique_tais = n_u;

return(0);
}


int compare_tai(const void *i, const void *j)
/*
!C*****************************************************************************
!Description:
	Compare two PGSt_double values. For use with ANSI qsort().
	See qsort() man pages for generic description.

!Input Parameters:
i -- pointer to a PGSt_double value
j -- pointer to a PGSt_double value

!Output Parameters: none

!Input/Output Parameters: none

!Returns: -1 if *i is less than *j
          1 if *i is greater than *j

!Revision History:
	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

   Paul Fisher
   Science Systems and Applications Inc.
   NASA's Goddard Space Flight Center Code 923
   Greenbelt MD, 20771
   fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes
    called by routine get_unique_granule_times() (modis_srefl_io.c)

!END***************************************************************************
*/
{
  if(*(PGSt_double *)i<*(PGSt_double *)j)return(-1);
  else if(*(PGSt_double *)i>*(PGSt_double *)j)return(1);
  else return(0);
}

int get_granule_times(dims_struct ****inputs, int *n_in, PGSt_double orbit_start_tai, PGSt_double orbit_end_tai,
                      PGSt_double *granule_start_tai, PGSt_double *granule_end_tai)
/*
!C****************************************************************************
!Routine: get_granule_times()

!Description: Gets granule start/end times in TAI from "inputs" metadata:
              checks granule start/end times against those of the orbit.

!Input Parameters:
  int *n_in                       number of inputs (dimension ALL_L1B_FILES)
  PGSt_double  orbit_start_tai
  PGSt_double  orbit_end_tai      start, end time of orbit (in TAI)
  
!Input/Output Parameters:
  dims_struct *inputs             Pointer to the array of input structures 
                                  (field inputs->xxxxx->r{be}dt_tai are updated)
 
!Output Parameters:
  PGSt_double  granule_start_tai
  PGSt_double  granule_end_tai    minimal start, maximal end time of all granules (in TAI)
 
!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  06-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
      called by routine sort_inputs() (modis_srefl_io.c)

!END**************************************************************************/
{
int i,j;
char location[100]="modis_srefl_io.c, get_granule_times\0";
char configParamVal[PGSd_PC_VALUE_LENGTH_MAX] = {'\0'};
int ret;
PGSt_double granule_tai;

*granule_start_tai=FLT_MAX;
*granule_end_tai=0.0;
for(i=0;i<ALL_L1B_FILES;i++) {
  for(j=0;j<n_in[i];j++) {
     sprintf(configParamVal, "%s.0000Z\0", (*inputs)[i][j]->xxxxx->rbdt);
     if(PGS_TD_UTCtoTAI(configParamVal, &granule_tai)!=PGS_S_SUCCESS)
	modsmf(MODIS_E_SDP_ERROR, "PGS_TD_UTCtoTAI()", location);
     /*printf("%d %d %s %f\n", i,j,(*inputs)[i][j]->xxxxx->rbdt, granule_tai);*/
     if (granule_tai < orbit_start_tai) {
        modsmf(MODIS_W_OUT_OF_ORBIT, configParamVal, location);
	(*inputs)[i][j]->OK = 1;
	}
     else {
        (*inputs)[i][j]->xxxxx->rbdt_tai = granule_tai;
        if (granule_tai < *granule_start_tai)*granule_start_tai = granule_tai;
        }

     sprintf(configParamVal, "%s.0000Z\0", (*inputs)[i][j]->xxxxx->redt);
     if(PGS_TD_UTCtoTAI(configParamVal, &granule_tai)!=PGS_S_SUCCESS)
	modsmf(MODIS_E_SDP_ERROR, "PGS_TD_UTCtoTAI()", location);
     /*printf("%d %d %s %f\n", i,j,(*inputs)[i][j]->xxxxx->redt, granule_tai);*/
     if (granule_tai > orbit_end_tai) {
        sprintf(configParamVal, "%s.0000Z\0", (*inputs)[i][j]->xxxxx->rbdt);
        modsmf(MODIS_W_OUT_OF_ORBIT, configParamVal, location);
	(*inputs)[i][j]->OK = 1;
	}
     else  { 
        (*inputs)[i][j]->xxxxx->redt_tai = granule_tai;
        if (granule_tai > *granule_end_tai)  *granule_end_tai = granule_tai;
	}
   }
}
if ((*granule_start_tai == FLT_MAX)||(*granule_end_tai == 0.0)) 
   return(-1);
else
    return(0);   
}

int get_info(int lun, PGSt_integer version, dims_struct *inputs)
/*
!C****************************************************************************

!Routine: get_info()

!Description: Gets all the relevant information from the input file at
              lun/version and puts it into the structure input.

!Input Parameters:
  int lun
  PGSt_integer version
 
!Input/Output Parameters:
  dims_struct *inputs   Pointer to the array of input structures (updated)

!Output Parameters:
  none

!Returns:
  Success: 0
  Failure: nonzero
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
        called by routine setup_inputs() (modis_srefl_io.c)

!END**************************************************************************/
{
PGSt_integer local_version;
char location[100]="modis_srefl_io.c, get_info";
char filename[PGSd_PC_FILE_PATH_MAX]={'\0'};
char c;
char looking_for[100]; 
char esdt[20];
int ret, jday;
extern int sat_type;
PGSt_SMF_status getECSmetadata(PGSt_PC_Logical fileid, PGSt_integer version, input_metadata *input);

inputs->lun = (int32)lun; 
inputs->ver = (int32)version; 

/* Get filename */
local_version = version;
if(PGS_PC_GetReference((PGSt_PC_Logical)lun, &local_version, filename)!=PGS_S_SUCCESS){
    sprintf(msg, "LUN %d, version %d", (int32)lun, version);
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
    return(MODIS_E_GETREF_FAIL);
  }

local_version = version;
if(getECSmetadata((PGSt_PC_Logical)lun, local_version, inputs->xxxxx)!=MODIS_S_SUCCESS){ 
    sprintf(msg, "Error in getECSmetadata(), LUN %d, version %d", (int32)lun, version);
    modsmf(MODIS_W_GET_METADATA, msg, location);
    inputs->OK = 2; 
    return(inputs->OK);
  }

/* 27-JAN-05: Get the file's satellite ID, and make damned sure it agrees with 
   the sat_type that has been read from the PCF!!! */
if ( (ret = lgid_parser(inputs->xxxxx->localgranuleid, esdt, &jday)) == -1 ) {
    sprintf(msg, "Error in lgid_parser(), LUN %d, version %d", (int32)lun, version);
    modsmf(MODIS_W_GET_METADATA, msg, location);
    inputs->OK = 3; 
    return(inputs->OK);
  }

if (ret != sat_type) {
    sprintf(msg, "File at LUN %d, version %d is a %s file, while the PCF and/or MCFs are for %s files", 
        (int32)lun, version, (ret == 0 ? "Terra" : "Aqua"), (sat_type == 0 ? "Terra" : "Aqua") );
    modsmf(MODIS_F_INPUT_NOMATCH, msg, location);
 }

inputs->filename = (char *)AllocMem(strlen(filename)+1, location);
strcpy(inputs->filename, filename);

/* This next step is very weird; LINUX boxes HATE this and give immediate Segmentation fault right here.
   But HPUX machines (e. g., tracer) need this.  */
#ifdef HPUX
strcat(inputs->filename, '\0');
#endif

inputs->OK = 0;   /* OK */
return(inputs->OK);
}


int lgid_parser(char *lgid, char *esdt, int *jday)
/*
!C********************************************************************
!Description: lgid_parser() 

     Chops the LocalGranuleID into parts.
     
!Input Parameters:
     char lgid     the LocalGranuleID
     
!Output Parameters:   
     char esdt     its ESDT
     int jday      the julian day in the LGID.

!Input/Output Parameters:
     none

!Returns: the satellite ID in the LocalGranuleID:
      -1 error
      0 TERRATYPE (Terra)
      1 AQUATYPE (Aqua) 
      
!Revision History:
   Revision 1.0  
   Jim Ray/SSAI
   Initial revision.
 
!Team-unique Header: 

  This software is created by the MODIS Land Science Team and Goddard
  Space Flight Center Code 923 under contract NAS5-31752/002.

!References and Credits:
  Written by Jim Ray
  Science Systems and Applications Inc.
  MODIS Land Science Team -- UMD
  4321 Hartwick Rd., Suite 310
  College Park, MD 20742
  jim@cratmos.gsfc.nasa.gov
  phone: (301) 405-8120


!Design Notes:
    Copied from MOD_PR09A's code (v5.0.1) on 27-JAN-05, slightly modified.
    
!END  
*********************************************************************/
{
int i, j, n, n_dots = 0;
int satid;
char x[5];
char location[]="srefl3_io.c, lgid_parser\0";
#define TERRATYPE 0
#define AQUATYPE 1

if ((n = strlen(lgid)) <= 0) return(-1);

for(i=0;i<n;i++) 
   if(lgid[i] == '.') n_dots++;

if (n_dots < 5) return(-1);

for(i=0;i<n;i++) {
   esdt[i] = lgid[i];
   if(lgid[i] == '.') break;
   }
esdt[i] = '\0';

if ((esdt[0] != 'M')||(esdt[2] != 'D')) return(-1);
if (esdt[1] == 'O') satid = TERRATYPE;
else if (esdt[1] == 'Y') satid = AQUATYPE;
else  {
      sprintf(msg, "satellite type in ESDT '%s' not recognized", esdt);
      modsmf(MODIS_E_GETFILEINFO, msg, location);
      return(-1);
      }

i = i + 2;
for(j=0;j<4;j++,i++) x[j] = lgid[i];
x[j] = '\0';
/*  *fourdigityear = atoi(x);  Why bother checking the year? */

for(j=0;j<3;j++,i++) x[j] = lgid[i];
x[j] = '\0';
*jday = atoi(x);

return(satid);
}


void init_struct(dims_struct *inputs, char *call)
/*
!C****************************************************************************

!Routine: init_struct()

!Description: Generic initializer for I/O structures.  Also initializes
               the metadata structure within the dims_struct structure.

!Input Parameters:
  none
  
!Input/Output Parameters:
  dims_struct **inputs  Pointer to the I/O structure 

!Output Parameters:
  none

!Returns:
  none

!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
       called by routine open_outputs() (modis_srefl_io.c)
       called by routine open_coarse_outputs() (modis_srefl_io.c)
       called by routine get_input_str_memory() (modis_srefl_io.c)
       called by routine atmospheric_correction() (atmcor.c)

!END**************************************************************************/
{
int i;

inputs->OK = -1;
inputs->sdid = (int32)-1;
inputs->lun = (int32)-1;
inputs->ver = (int32)-1;
inputs->n_sds = (int32)-1;
for(i=0;i<BANDPROCESS;i++) {
   inputs->refscale[i] = (float32)-1.0;
   inputs->refoffset[i] = (float32)-1.0;
   inputs->radscale[i] = (float32)-1.0;
   inputs->radoffset[i] = (float32)-1.0;
   inputs->sds_offset[i] = (int32)-1;
   inputs->band_name[i] = (int32)-1;
   }
for(i=0;i<ALLBANDS;i++) {
   inputs->type[i] = (int32)-1;
   inputs->data[i] = (void *)NULL;
   inputs->sdsid[i] = (int32)-1;
    }
/* new 13-APR-06 */    
for(i=0;i<TEMPS;i++) {
    inputs->temp_fd[i] = (FILE *)NULL;
    inputs->temp_lun[i] = (int32)-1;
   }

inputs->filename = (char *)NULL;
inputs->nblayers = (int32)-1;
inputs->nblines = (int32)-1;
inputs->nbcols = (int32)-1;
inputs->xxxxx = (input_metadata *)malloc(sizeof(input_metadata));
/*printf("%p %s %d\n", inputs->xxxxx, call, sizeof(input_metadata));*/
initECSmetadata(inputs->xxxxx);
}

int open_inputs(dims_struct **local_input, int *nlines_1km, int *ncols_1km, int *local_process_bands)
/*
!C****************************************************************************

!Routine: open_inputs()

!Description: Figures out what to read in all input files.

!Input Parameters:
  none
  
!Input/Output Parameters:
  dims_struct **local_input     Pointer to the input structure 
  int local_process_bands       flag for processing/not processing individual bands
  
!Output Parameters:
  int nlines_1km
  int ncols_1km

!Returns:
  0 success
  1 skip this time-point (to be run once for each time-point; inputs are all for the same
                          time).

!Revision History:
  Revision 1.0  11-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
         called by routine atmospheric_correction() (atmcor.c)

!END**************************************************************************/
{
int32 i,j;
int n_SDSs_to_look_for;
char what_to_get[10][25];
char start[4] = "EV_\0";
char end[7] = "_RefSB\0";
int8 band_index[] = {0,2,9,BANDPROCESS,0,0,0};
extern int n_bands_res[3]; /* 250m, 500m, 1km */
char location[100]="modis_srefl_io.c, open_inputs\0";
char filetype[15] = {'\0'};
char msg[256] = {'\0'};
extern int32 dead_bands[BANDPROCESS][4 * NBSCANS_1KM];
extern int32 noisy_bands[BANDPROCESS][4 * NBSCANS_1KM];

#ifdef SUBSET
extern char method[256];
#endif

*nlines_1km = *ncols_1km = 0;
if ((local_input[RAD1_FILE]->OK != 0)||(local_input[GEO_FILE]->OK != 0)) return(1);

#ifdef SUBSET
strcpy(method, "\0");
#endif

for(i=0;i<ALL_L1B_FILES;i++) {
    if (local_input[i]->OK != 0) {
        for(j=band_index[i];j<band_index[i+1];j++) local_process_bands[j] = 0;
	if (i==RAD250_FILE) n_bands_res[0] = 0;
	else if (i==RAD500_FILE) n_bands_res[1] = 0;
	else if (i==RAD1_FILE) n_bands_res[2] = 0;
        continue;
	}
    
    switch (i) {
       case RAD250_FILE:
           n_SDSs_to_look_for = 1;
	   sprintf(what_to_get[0], "%s250%s\0", start, end);
	   sprintf(filetype, "250m\0");
           break;
       case RAD500_FILE:
           n_SDSs_to_look_for = 2;
	   sprintf(what_to_get[0], "%s500%s\0", start, end);
	   sprintf(what_to_get[1], "%s250_Aggr500%s\0", start, end);
	   sprintf(filetype, "500m\0");
           break;
       case RAD1_FILE:
           n_SDSs_to_look_for = 4;
	   sprintf(what_to_get[0], "%s1KM%s\0", start, end);
	   sprintf(what_to_get[1], "%s1KM_Emissive\0", start);
	   sprintf(what_to_get[2], "%s500_Aggr1km%s\0", start, end);
	   sprintf(what_to_get[3], "%s250_Aggr1km%s\0", start, end);
	   sprintf(filetype, "1km\0");
           break;
       case GEO_FILE:
           n_SDSs_to_look_for = 8;
	   strcpy(what_to_get[0], "Latitude\0");
	   strcpy(what_to_get[1], "Longitude\0");
	   strcpy(what_to_get[2], "Height\0");
	   strcpy(what_to_get[3], "SensorZenith\0");
	   strcpy(what_to_get[4], "SensorAzimuth\0");
	   strcpy(what_to_get[5], "SolarZenith\0");
	   strcpy(what_to_get[6], "SolarAzimuth\0");
	   strcpy(what_to_get[7], "Land/SeaMask\0");
 	   sprintf(filetype, "geolocation\0");
           break;
       case CLD_FILE:
           n_SDSs_to_look_for = 1;
	   strcpy(what_to_get[0], "Cloud_Mask\0");
 	   sprintf(filetype, "cloudmask\0");
           break;
       }

    if ( get_L1B_sds(i, &(local_input[i]), what_to_get, n_SDSs_to_look_for, local_process_bands) == 2) {
        sprintf(msg, " File '%s' (LUN %d, version %d) does not have inputs expected of a %s file\0", 
	    local_input[i]->filename, local_input[i]->lun, local_input[i]->ver, filetype); 
        modsmf((PGSt_SMF_code)MODIS_F_LEVEL1B_READ, msg, location);
      }

    if (i == RAD500_FILE) { /* 01-FEB-05: read dead/noisy detector data */
        read_dead_det(local_input[i]->sdid, dead_bands, noisy_bands);
        }

    }
*nlines_1km = local_input[RAD1_FILE]->nblines;
*ncols_1km = local_input[RAD1_FILE]->nbcols;
return(0);
}

int get_L1B_sds(int index, dims_struct **local_input, char what_to_get[10][25], int n_SDSs_to_look_for, int *local_process_bands)
/*
!C****************************************************************************

!Routine: get_L1B_sds()

!Description: SDstarts all input files.

!Input Parameters:
  int index                     what type of file is being processed 
                                  (RAD250_FILE, RAD500_FILE, RAD1_FILE, GEO_FILE, CLD_FILE... )
  char what_to_get              names of the SDSs to read (dimension n_SDSs_to_look_for)
  int n_SDSs_to_look_for        number of SDSs to read 
  
!Input/Output Parameters:
  dims_struct **local_input     Pointer to the input structure (updated) 
  int *local_process_bands      whether to process this band or not
  
!Output Parameters:
  none

!Returns:
  0 success
 -1 nonfatal error  (error message is printed out here)
  2 fatal error -- file does not have any of the expected SDSs (handled in calling routine)

!Revision History:
  Revision 1.0  11-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 

    Tries to read SDS id's, scales, offsets, etc into I/O structure variables 
    in the BANDPROCESS order: e. g., 
     BAND1_250M, BAND2_250M, BAND1_500M, BAND2_500M, BAND3_500M,
     BAND4_500M, BAND5_500M, BAND6_500M, BAND7_500M, BAND1_1KM, 
     BAND2_1KM,  BAND3_1KM,  BAND4_1KM,  BAND5_1KM,  BAND6_1KM, 
     BAND7_1KM,  BAND8,      BAND18,     BAND19,     BAND20,  
     BAND21,     BAND22,     BAND23,     BAND26,     BAND29,     
     BAND31,     BAND32,   BANDPROCESS
       
     called by routine open_inputs() (modis_srefl_io.c)

!END**************************************************************************/
{
int32 i, ii, j, jj, n_sets, ret, sdid, sds, rank, dims[5], start[3]; 
int32 number_type, count, n_gattr, nattr, n_bands;
int8 match, at_least_one_match;
char name[256];
char attrib_name[256];
char band_names[100];
int32 band_offset[3] = {-1, 1, 8};
int32 bands[BANDPROCESS];
int8 band_index[] = {0,2,9,BANDPROCESS,0,0,0};
float64 geoscale;
float32 refoffset[BANDPROCESS], refscale[BANDPROCESS], radoffset[BANDPROCESS], radscale[BANDPROCESS];
char tmp[6];
char location[100]="modis_srefl_io.c, get_L1B_sds\0";
char msg[256] = {'\0'};
#ifdef SUBSET
extern char method[256];
char local_method[256] = {'\0'}; 
int32 attr_number;
#endif

if ( (sdid = SDstart((*local_input)->filename, DFACC_RDONLY)) == -1) {
   sprintf(msg, "Error, SDstart, file '%s'", (*local_input)->filename);
   modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
   for(j=band_index[index];j<band_index[index+1];j++) local_process_bands[j] = 0;
   return(-1);
      }
      
ret = SDfileinfo(sdid, &n_sets, &n_gattr);
if ((ret != 0) || (n_sets <= 0)) {
   sprintf(msg, "Cannot read file '%s'", (*local_input)->filename);
   modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
   for(j=band_index[index];j<band_index[index+1];j++) local_process_bands[j] = 0;
   return(-1);
      }

#ifdef SUBSET
attr_number = SDfindattr(sdid, "SubsettingMethod\0");
SDreadattr(sdid, attr_number, &local_method);
if (strlen(method) == 0) strcpy(method, local_method);
else {
    if (strcmp(method, local_method))  {
      sprintf(msg, "Subsetting method for %s does not match others", (*local_input)->filename);
      modsmf((PGSt_SMF_code)MODIS_F_LEVEL1B_READ, msg, location);
        }
     }
#endif

at_least_one_match = 0;
for (j=0;j<n_sets;j++) {
   sds = SDselect(sdid, j);
   ret = SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
   if ((ret != 0) || (rank <= 0)) {
      sprintf(msg, "Cannot read SDS %s in '%s'", (*local_input)->filename, name);
      modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
      for(j=band_index[index];j<band_index[index+1];j++) local_process_bands[j] = 0;
      return(-1);
         }
	 
   /* check each SDS for the one we want */
   match = -1;
   for (ii=0;ii<n_SDSs_to_look_for;ii++) {
       if ( !strcmp(name, what_to_get[ii]) ) match = ii;
      }
   if (match != -1) {
         at_least_one_match = 1;
	 if ((*local_input)->sdid == -1)    (*local_input)->sdid = sdid;
	 if ((*local_input)->n_sds == -1)   (*local_input)->n_sds = 0;
	 
	 if (rank == 2) {  /* geolocation */
	    get_2D_scale(sds, &geoscale);
	    if ((*local_input)->nblines == -1) (*local_input)->nblines = dims[0];
	    if ((*local_input)->nbcols == -1)  (*local_input)->nbcols = dims[1];
	    if ((*local_input)->n_sds < BANDPROCESS)  {
	      (*local_input)->sdsid[(*local_input)->n_sds] = sds;
	      (*local_input)->type[(*local_input)->n_sds] = number_type;
	      (*local_input)->refscale[(*local_input)->n_sds] = (float32)geoscale;
	      (*local_input)->radscale[(*local_input)->n_sds] = (float32)geoscale;
	      (*local_input)->n_sds++;
	       }
	     }
	 else if (rank == 3) {  
	    if ((*local_input)->nblayers == -1) (*local_input)->nblayers = dims[0];
	    if ((*local_input)->nblines == -1) (*local_input)->nblines = dims[1];
	    if ((*local_input)->nbcols == -1)  (*local_input)->nbcols = dims[2];
	     
   	    if ((n_SDSs_to_look_for == 1) && !strcmp(name, "Cloud_Mask\0"))  {   /* cloudmask file */
	       if ((*local_input)->n_sds < BANDPROCESS)  {
                   for(ii=0;ii<dims[0];ii++) {
	             (*local_input)->sdsid[(*local_input)->n_sds] = sds;
	             (*local_input)->type[(*local_input)->n_sds] = number_type;
	             (*local_input)->sds_offset[(*local_input)->n_sds] = ii;
	             (*local_input)->n_sds++;
		     }
	          }
	       }	   
	    else  {
	
	       /* get band_names, offsets, scales */
	       get_bands_and_scales(sds, bands, &n_bands, refoffset, refscale, radoffset, radscale);
	       if (n_bands != dims[0]) {
                   sprintf(msg, "Cannot read SDS in '%s' ", (*local_input)->filename);
                   modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
                   for(j=band_index[index];j<band_index[index+1];j++) local_process_bands[j] = 0;
                   return(-1);
                   }
		   		 
               for(ii=0;ii<n_bands;ii++) {
	       
		 jj = -1;
		 if (index == RAD250_FILE) { /* 250m data; band names must be below 3 */
		     if (bands[ii] < 3) {
			 jj = band_offset[0] + bands[ii];
		        }
	             else {
                         sprintf(msg, "Error, 250m L1B file '%s' has corrupted metadata\n", (*local_input)->filename);
                         modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
                         return(-1);
		        }
	             }
		 else if (index == RAD500_FILE) { /* 500m data; band names must be below 8 */
		     if (bands[ii] < 8) {
			 jj = band_offset[1] + bands[ii];
		        }
	             else {
                         sprintf(msg, "Error, 500m L1B file '%s' has corrupted metadata\n", (*local_input)->filename);
                         modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
                         return(-1);
		        }
	             }
		  else if (index == RAD1_FILE) { /* 1km data */
		     if (bands[ii] < 8)   
	                 jj = band_offset[2] + bands[ii];
		     else {    /* other data */
		         switch (bands[ii]) {
			     case 8 :
			         jj = BAND8;
				 break;
			     case 9 :
			         jj = BAND9;
				 break;
			     case 10 :
			         jj = BAND10;
				 break;
			     case 11 :
			         jj = BAND11;
				 break;
			     case 12 :
			         jj = BAND12;
				 break;
			     case 13 :
			         jj = BAND13;
				 break;
			     case 14 :
			         jj = BAND14;
				 break;
			     case 15 :
			         jj = BAND15;
				 break;
			     case 16 :
			         jj = BAND16;
				 break;
			     case 18 :
			         jj = BAND18;
				 break;
			     case 19 :
			         jj = BAND19;
				 break;
			     case 20 :
			         jj = BAND20;
				 break;
			     case 21 :
			         jj = BAND21;
				 break;
			     case 22 :
			         jj = BAND22;
				 break;
			     case 23 :
			         jj = BAND23;
				 break;
			     case 26 :
			         jj = BAND26;
				 break;
			     case 29 :
			         jj = BAND29;
				 break;
			     case 31 :
			         jj = BAND31;
				 break;
			     case 32 :
			         jj = BAND32;
				 break;
			     default:
				 break;
			     }
		         } 
                      }
		     if (jj > -1) {
	                (*local_input)->type[jj] = number_type;
	                (*local_input)->sds_offset[jj] = ii;
	                (*local_input)->band_name[jj] = bands[ii];
	                (*local_input)->refoffset[jj] = refoffset[ii];
	                (*local_input)->radoffset[jj] = radoffset[ii];
	                (*local_input)->refscale[jj] = refscale[ii];
	                (*local_input)->radscale[jj] = radscale[ii];
	                (*local_input)->sdsid[jj] = sds;
	                (*local_input)->n_sds++;
			if (jj > 7) local_process_bands[jj] = 1;  /* turn on band 8/26 and emissive band processing! */
                        }
	          }  /*  for(ii=0;ii<n_bands;ii++) */
	      } 
          }  /* if (rank == 3) */
       }
       else 
          SDendaccess(sds);	  
	  
   }  /* for (j=0;j<n_sets;j++) */

if (at_least_one_match == 0) return(2);
else return(0);
}


void get_bands_and_scales(int32 sds, int32 *bands, int32 *n_bands, float32 *refoffset, float32 *refscale, 
                          float32 *radoffset, float32 *radscale)
/*
!C****************************************************************************

!Routine: get_bands_and_scales()

!Description: Reads band_names, radiance_scales, radiance_offsets, reflectance_scales, reflectance_offsets
              from L1B SDSs.

!Input Parameters:
  int32 sds                    the SDSs to read 
  
!Input/Output Parameters:
  none 
  
!Output Parameters:
  int32 bands
  int32 n_bands
  float32 refoffset
  float32 refscale
  float32 radoffset
  float32 radscale

!Returns:
  none

!Revision History:
  Revision 1.0  11-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
   Values in attribute 'band_names' are converted to integers and written to 'bands'.  
   I am ignoring "13hi, 14hi".  The internal band number for 13hi, 14hi is 113, 114,
   which is ignored in later code (e. g., get_L1B_sds()).
   
   called by routine get_L1B_sds() (modis_srefl_io.c)

!END**************************************************************************/
{
int32 i, ii, j, jj, n_sets, rank, dims[5]; 
int32 number_type, count, n_gattr, nattr, iband, inhi;
int8 errorflag, match;
char name[256];
char attrib_name[256];
char band_names[100];
char tmp[6];

iband=0;
errorflag = 0;

errorflag |= SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
   for(i=0;i<nattr;i++) {
       errorflag |= SDattrinfo(sds, i, attrib_name, &number_type, &count);
       
       if (!strcmp(attrib_name, "band_names\0")) {
           errorflag |= SDreadattr(sds, i, band_names);
	   
	   /* Convert to integers */
	   band_names[count] =  ',';
	   band_names[count+1] =  '\0';
	   iband=0;
	   inhi=0;
	   for(ii=0,j=0;ii<strlen(band_names);ii++) {
	       if (band_names[ii] == ',') {
		  tmp[j] = '\0';
	          j = 0;
		  bands[iband] = atoi( tmp );
		  if (inhi) bands[iband] += 100;
		  inhi = 0;
		  iband++;
	       }
	       else tmp[j++] = band_names[ii];
	       
	       if (ii>0) {
	          if ((band_names[ii-1] == 'h') && (band_names[ii] == 'i')) inhi = 1; 
	          }
	     }
	   
	   /* A quick sanity check for band names */
	   for(ii=0;ii<iband;ii++) {
	       /*printf("band %d\n", bands[ii]);*/
               if ((bands[ii] < 0)/*||(bands[ii] > 36)*/)errorflag |= 1;
	     }
	   }
		
        if (!strcmp(attrib_name, "reflectance_scales\0")) {
             errorflag |= SDreadattr(sds, i, refscale);
           }
        if (!strcmp(attrib_name, "reflectance_offsets\0")) {
             errorflag |= SDreadattr(sds, i, refoffset);
           }
        if (!strcmp(attrib_name, "radiance_scales\0")) {
             errorflag |= SDreadattr(sds, i, radscale);
           }
        if (!strcmp(attrib_name, "radiance_offsets\0")) {
             errorflag |= SDreadattr(sds, i, radoffset);
           }
 
       }
if (errorflag == 0)
    *n_bands = iband;
}

void get_2D_scale(int32 sds, float64 *scale)
/*
!C****************************************************************************

!Routine: get_2D_scale()

!Description: Reads scale factor metadata from 2D SDSs.

!Input Parameters:
  int32 sds                    the SDSs to read 
  
!Input/Output Parameters:
  none 
  
!Output Parameters:
  float32 scale

!Returns:
  none

!Revision History:
  Revision 1.0  11-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
           called by routine get_L1B_sds() (modis_srefl_io.c)

!END**************************************************************************/
{
int32 i, ii, j, jj, n_sets, rank, dims[5]; 
int32 number_type, count, n_gattr, nattr;
int8 errorflag;
char name[256];
char attrib_name[256];

*scale = 1.0;
errorflag = 0;

errorflag |= SDgetinfo(sds, name, &rank, dims, &number_type, &nattr);
   for(i=0;i<nattr;i++) {
       errorflag |= SDattrinfo(sds, i, attrib_name, &number_type, &count);
       if (!strcmp(attrib_name, "scale_factor\0")) {
             errorflag |= SDreadattr(sds, i, scale);
           }
       }
       
if (errorflag != 0) *scale = 1.0;

}


int open_outputs(int mode, int output_lun, dims_struct **output, dims_struct *input, int nlines_1km, int ncols_1km, 
                 int *local_process_bands, int *band_numbers)
/*
!C****************************************************************************

!Routine: open_outputs()

!Description: Throws open the full resolution output file at output_lun/version.

!Input Parameters:
  int mode                      mode of processing, either 1 (AOT retrieval) or 0 (AOT use)
  int output_lun
  int version
  int local_process_bands       flag for processing/not processing individual bands
  int nlines_1km
  int ncols_1km
  int *band_numbers             array of band numbers (dimension BANDPROCESS) -- the actual
                                number of the band for array BANDPROCESS.
				
!Input/Output Parameters:
  dims_struct **local_output    Pointer to the I/O structure 
  
!Output Parameters:
  none

!Returns:
  0 success
  1 problem with PGS_PC_GetReference.
 -1 problem with SDcreate  
 (SDstart problem is FATAL).

!Revision History:
  Revision 1.0  11-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
         called by routine atmospheric_correction() (atmcor.c)

         This routine sets (*output)->type[i] for array, when array is to be set up
	 *FOR OUTPUT*.  If array is necessay but is NOT to be written anywhere,
	 (*output)->type[i] won't be set, so routines that allocate memory for
	 the array have to be told the type some other way.
   
!END**************************************************************************/
{
int i, j;
int32 dimid, ret;
int8 qa_flag;
PGSt_integer local_version, version;
int32 sdid, sdsid, rank, number_type, n_attr, dims[2], qadims[2];
int32 factor, nblines, nbcols, index;
char prefix[5];
char location[100]="modis_srefl_io.c, open_outputs\0";
char name[PGSd_PC_FILE_PATH_MAX]={'\0'};
char filename[PGSd_PC_FILE_PATH_MAX]={'\0'};
char dimname[60];
extern output_type_t output_type[ALLBANDS];
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
int bandindex, n_bands[3]; /* 250m, 500m, 1km */

/* New, 20-DEC-05 */
int make_eos = 1;
void HDFEOSstart(char *filename, int nlines_1km, int ncols_1km, int *band_numbers, int compress);
int altHDFEOSstart(char *filename, int nlines_1km, int ncols_1km, int *band_numbers, int compress);
int make_chunked = 1;
void set_chunking(int mode, int32 sds, int32 lines, int32 cols);

/* 13-APR-06 */
int32 lun;
long total_n_pixels;
PGSt_IO_Gen_FileHandle *fd;
extern int time_point;   /* this is global in file atmcor.c */
int16 blanks[NBSPOTS_1KM*NBSCANS_1KM];

/* Get filename */
local_version = input->ver;
version = input->ver;
if(PGS_PC_GetReference((PGSt_PC_Logical)output_lun, &local_version, name)!=PGS_S_SUCCESS){
    sprintf(msg, "LUN %d, version %d", (int32)output_lun, version);
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
    return(1);
  }
strcpy(filename, name);  
if (rmpath(filename)!=MODIS_S_SUCCESS)
    modsmf((PGSt_SMF_code)MODIS_E_FUNCTION_ERROR, "Failure from rmpath()", location);

if (mode == 1) {

    if (make_eos) {
      /* 20-DEC-05: Changes to make full-resolution outputs HDFEOS-compatible
       ************************************************************************
       */
      /* create the file with HDF open */
      if((sdid = Hopen((char *) name, DFACC_CREATE, 0)) == -1) {
          sprintf(msg, "LUN %ld, version %d", (int32)output_lun, input->ver);
          modsmf((PGSt_SMF_code)MODIS_F_OPEN_HDF_FILE, msg, location);
          }
    
      /* close the file */
      Hclose(sdid);
   
      /* set up all the HDFEOS stuff */
      altHDFEOSstart(name, nlines_1km, ncols_1km, band_numbers, make_chunked);

      if( (sdid=SDstart(name, DFACC_RDWR)) == -1 )  {
         sprintf(msg, "SDstart, LUN %d, version %d", (int32)output_lun, version);
          modsmf((PGSt_SMF_code)MODIS_F_OPEN_HDF_FILE, msg, location);
        }
      /****************************
       * end of 20-DEC-05 changes
       */           
       }   
    else {                                         
      if( (sdid=SDstart(name, DFACC_CREATE)) == -1 ) {
          sprintf(msg, "SDstart, LUN %d, version %d", (int32)output_lun, version);
           modsmf((PGSt_SMF_code)MODIS_F_OPEN_HDF_FILE, msg, location);
          }
        }
    sprintf(msg, "(AOTR mode): file %s, GRANULENUMBER %d, RANGEBEGINNINGDATETIME %s", 
            filename , input->xxxxx->granulenumber, input->xxxxx->rbdt);
    modsmf(MODIS_S_SREFL_L2GRAN, msg, location);

   }
else {
   if( (sdid=SDstart(name, DFACC_RDWR)) == -1 ) {
       sprintf(msg, "SDstart, LUN %d, version %d", (int32)output_lun, version);
       modsmf((PGSt_SMF_code)MODIS_F_OPEN_HDF_FILE, msg, location);
     }
   else {
       sprintf(msg, "(AOTU mode): file %s, GRANULENUMBER %d, RANGEBEGINNINGDATETIME %s", 
            filename , input->xxxxx->granulenumber, input->xxxxx->rbdt);
       modsmf(MODIS_S_SREFL_L2GRAN, msg, location);
    }
   }

/* initialize the output structure * /
init_struct(*output, location);*/
(*output)->sdid = sdid;
rank = 2;
(*output)->filename = (char *)malloc(strlen(name)+1);
strcpy((*output)->filename, name);

#ifdef HPUX
strcat((*output)->filename, '\0');
#endif

bandindex = 0;
for (i=0;i<3;i++) n_bands[i] = 0;
for (i=0;i<BAND8;i++) {
    if ((i==BAND1_500M)||(i==BAND1_1KM))bandindex++;
    if (local_process_bands[i] == 1) n_bands[bandindex]++;
    }

/* Set up the HDF interface */


if (mode == 1) {
   nblines = (int32)nlines_1km;
   nbcols  = (int32)ncols_1km;
   (*output)->nblines = nblines;
   (*output)->nbcols = nbcols;
   dims[0] = nblines;
   dims[1] = nbcols;
   strcpy(prefix, "1km\0");
   /*for (i=AOT_1;i<SURFP;i++)*/  /* will put H2O into first pass, not second */
   for (i=AOT_1;i<=WATERVAPOR;i++){
   
    if (i==SURFP) continue;
    if (i==SLEVP) continue;
    
    if  ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH)) {
   
      sprintf(name, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
#ifdef OLDSDSNAMES 
      if(i == WATERVAPOR)sprintf(name, "1km water_vapor\0");
#endif           
      /* 20-DEC-05: SDS will have been created in HDFEOSstart(), just SDselect() it here */      
      if (make_eos) {
         if ((sdsid = SDselect(sdid, SDnametoindex(sdid, name)))==-1) return(-1);
	 }
      else {
         if ((sdsid = SDcreate(sdid, name, output_type[i].type, rank, dims))==-1) return(-1);
	 
         sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
         if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
         sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
         if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);
	 	 
         if (make_chunked)  set_chunking(mode, sdsid, NBSCANS_1KM, NBSPOTS_1KM);
	    
	 }
 
      (*output)->sdsid[i] = sdsid;
      (*output)->type[i] = output_type[i].type;
        
      
      }
      
   }
   for (i=AOTCM;i<=AOTQA;i++) {
    if  ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH)) {
      sprintf(name, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
      
      /* 20-DEC-05: SDS will have been created in HDFEOSstart(), just SDselect() it here */      
      if (make_eos) {
         if ((sdsid = SDselect(sdid, SDnametoindex(sdid, name)))==-1) return(-1);
	 }
      else {
         if ((sdsid = SDcreate(sdid, name, output_type[i].type, rank, dims))==-1) return(-1);
         sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
         if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
         sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
         if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);
	 
         if (make_chunked)  set_chunking(mode, sdsid, NBSCANS_1KM, NBSPOTS_1KM);
	    
      }
      
      (*output)->sdsid[i] = sdsid;
      (*output)->type[i] = output_type[i].type;
        
      
      }
   }
   /* 31-MAR-06 */
#ifndef TOTEMP
   for (i=TRBAND1;i<=TGBAND16;i++) {
    if  ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH)) {
      sprintf(name, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
      
      /* 20-DEC-05: SDS will have been created in HDFEOSstart(), just SDselect() it here */      
      if (make_eos) {
         if ((sdsid = SDselect(sdid, SDnametoindex(sdid, name)))==-1) return(-1);
	 }
      else {
         if ((sdsid = SDcreate(sdid, name, output_type[i].type, rank, dims))==-1) return(-1);
         sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
         if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
         sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
         if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);
	 
         if (make_chunked)  set_chunking(mode, sdsid, NBSCANS_1KM, NBSPOTS_1KM);
	    
      }
      
      (*output)->sdsid[i] = sdsid;
      (*output)->type[i] = output_type[i].type;
        
      
      }
   }
#endif
   
   /* 13-APR-06 */
#ifdef TOTEMP
   for (i=0;i<TEMPS;i++) {
      lun = (PGSt_PC_Logical)(TEMPLUN + (time_point*4) + i);
      (*output)->temp_lun[i] = (int32)lun;
      
      ret = PGS_IO_Gen_Temp_Open( PGSd_IO_Gen_NoEndurance, lun, PGSd_IO_Gen_Write, &fd );
      
      if( ret != PGS_S_SUCCESS ) {
          sprintf(msg, "opening temp file for %s, GRANULENUMBER %d, RANGEBEGINNINGDATETIME %s", 
                  filename , input->xxxxx->granulenumber, input->xxxxx->rbdt);
          modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
       }
      /*printf("TP = %d, Opening file at lun %d for writing\n", time_point, lun);*/
      fseek(fd, 0L, SEEK_SET);
      
      /* And now initialize the whole thing with zeros */
      total_n_pixels = (long)(NBSCANS_1KM*NBSPOTS_1KM);
      memset(blanks,'\0', NBSCANS_1KM*NBSPOTS_1KM * sizeof(int16));      
      for(j=0;j<((nblines/NBSCANS_1KM)*16);j++) {
	 fwrite(blanks, 2, (size_t)total_n_pixels, fd);
         /*fseek(fd, total_n_pixels*2, SEEK_CUR);*/
         }
      fseek(fd, 0L, SEEK_SET);
      (*output)->temp_fd[i] = fd;

      }
#endif   

}

else {
  for (i=0;i<BANDPROCESS;i++) {
   if (i < BAND1_500M) {
       factor = 4;
       strcpy(prefix, "250m\0");
       }
   else if (i < BAND1_1KM) {
       factor = 2;
       strcpy(prefix, "500m\0");
       }
   else {
       factor = 1;
       strcpy(prefix, "1km\0");
       }
   
   nblines = (int32)nlines_1km*factor;
   nbcols  = (int32)ncols_1km*factor;
   (*output)->nblines = nblines;
   (*output)->nbcols = nbcols;
   dims[0] = nblines;
   dims[1] = nbcols;

   if ((local_process_bands[i] == 1) &&
       ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH))    /* here only worry about fine/coarse */
				    ){
      if ((i < BAND20) || (i == BAND26))sprintf(name, "%s Surface Reflectance Band %d\0", prefix, band_numbers[i]);	 
      else if ((i == LAT) || (i == LON))sprintf(name, "%s\0", sdsname[i-BAND20ALB]);
      else                              sprintf(name, "%s Brightness Temperature Band %d\0", prefix, band_numbers[i]);

#ifdef OLDSDSNAMES 
      if(i == BAND20)sprintf(name, "BAND20\0");
      if(i == BAND31)sprintf(name, "BAND31\0");
      if(i == BAND32)sprintf(name, "BAND32\0");
#endif           

      /* new, 04-FEB-05: */
      if ((mode == 2) || (make_eos == 1) ){  /* ASSUMING we've already run this in mode==0 mode */
         if ((sdsid = SDselect(sdid, SDnametoindex(sdid, name)))==-1) return(-1);
         if (make_eos == 1)  set_attributes(sdsid, name, prefix, output_type[i], DFNT_FLOAT, 0);
        }
      else {
        if ((sdsid = SDcreate(sdid, name, output_type[i].type, rank, dims))==-1) return(-1);
      
        sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
        if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
        sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
        if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);
	    
        set_attributes(sdsid, name, prefix, output_type[i], DFNT_FLOAT, 0);
      }

      if (make_chunked && (1==0) /* disabled */)  set_chunking(1, sdsid, NBSCANS_1KM*factor, NBSPOTS_1KM*factor);
      
      (*output)->sdsid[i] = sdsid;
      (*output)->type[i] = output_type[i].type;
    }
    (*output)->type[i] = output_type[i].type;
    
     }
     
     /* Here, now: dims[0] and dims[1] will contain data for 1km resolution; the following depends upon that */
     
  for (i=BAND20ALB;i<ALLBANDS;i++) {
  
   factor = 1;
   qa_flag = 0;

#ifdef TOTEMP
   if (i > CRSCM) continue;
#endif
   
   if  ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH)) {
   
      if (i==QKMQA) {
         factor = 4;
         strcpy(prefix, "250m\0");
	 if (n_bands[0] == 0) continue;
	 qa_flag = 1;
	 }
      else if (i==HKMQA) {
         factor = 2;
         strcpy(prefix, "500m\0");
	 if (n_bands[1] == 0) continue;
	 qa_flag = 1;
	 }
      else {
         factor = 1;
         strcpy(prefix, "1km\0");
 	 }

      if ((i==STATE)||(i==OKMQA)||(i==AOTCM)||(i==AOTQA)) qa_flag = 1;
      
      sprintf(name, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
      if ((i==LAT)||(i==LON)) sprintf(name, "%s\0", sdsname[i-BAND20ALB]);
      qadims[0] = dims[0]*factor;
      qadims[1] = dims[1]*factor;
      
#ifdef OLDSDSNAMES 
      if(i == WATERVAPOR)sprintf(name, "1km water_vapor\0");
      if(i == BAND20ALB)sprintf(name, "BAND20ALBEDO\0");
#endif           
      if ((i==AOT_1)||(i==AOT_3)||(i==AOT_8)||(i==AOT_M)||(i==AOTCM)||(i==AOTQA)||(i==WATERVAPOR)){
        if ((index = SDnametoindex(sdid, name))==-1) return(-1);
	else {
           if ((sdsid = SDselect(sdid, index))==-1) return(-1);
           else {
              (*output)->sdsid[i] = sdsid;
              (*output)->type[i] = output_type[i].type;
             }
        }
      }
      else {

         /* 20-DEC-05: SDS will have been defined in HDFEOSstart(); just SDselect() it here */
         if ((mode == 2) || (make_eos == 1)) {    
            if ((sdsid = SDselect(sdid, SDnametoindex(sdid, name)))==-1) return(-1);
           }
         else {
            if ((sdsid = SDcreate(sdid, name, output_type[i].type, rank, qadims))==-1) return(-1);
      
            sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
            if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
            sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
            if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);
	    
 	    
         }
	 
         (*output)->sdsid[i] = sdsid;
         (*output)->type[i] = output_type[i].type;

      }
            
      if (make_chunked && (1==0) /* disabled */)  {
        if (i < TRBAND1)
            set_chunking(1, sdsid, NBSCANS_1KM*factor, NBSPOTS_1KM*factor); 
        else
	    set_chunking(0, sdsid, NBSCANS_1KM*factor, NBSPOTS_1KM*factor);
         }
      
      if (i < QKMQA)  set_attributes(sdsid, name, prefix, output_type[i], DFNT_FLOAT, 0);
      else            set_attributes(sdsid, name, prefix, output_type[i], DFNT_FLOAT, qa_flag);
      
      }
      (*output)->type[i] = output_type[i].type;
     
  }
     
   /* 13-APR-06 */
#ifdef TOTEMP
   for (i=0;i<TEMPS;i++) {
      lun = (PGSt_PC_Logical)(TEMPLUN + (time_point*4) + i);
      (*output)->temp_lun[i] = (int32)lun;
      
      ret = PGS_IO_Gen_Temp_Open( PGSd_IO_Gen_NoEndurance, lun, PGSd_IO_Gen_Read, &fd );
      
      if( ret != PGS_S_SUCCESS ) {
          sprintf(msg, "opening temp file for %s, GRANULENUMBER %d, RANGEBEGINNINGDATETIME %s", 
                  filename , input->xxxxx->granulenumber, input->xxxxx->rbdt);
          modsmf(MODIS_F_OPEN_TEMP_FILE, msg,  location );
       }
      /*printf("TP = %d, Opening file at lun %d for reading\n", time_point, lun);*/
      (*output)->temp_fd[i] = fd;
      }
#endif   


}
return(0);
}

void set_chunking(int mode, int32 sds, int32 lines, int32 cols)
/*
!C****************************************************************************

!Routine: set_chunking()

!Description: Sets generic chunking parameters.

!Input Parameters:
  int32 sds                     the SDS to chunk, and
  int32 lines                   its dimensions
  int32 cols
				
!Input/Output Parameters:
  none
  
!Output Parameters:
  none

!Returns:
  nothing

!Revision History:
  Revision 1.0  22-JAN-2006
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
   
!END**************************************************************************/
{
HDF_CHUNK_DEF cdef;
int32 max_cache, cache_flag;

cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = lines;
cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = cols;
cdef.comp.comp_type = COMP_CODE_DEFLATE;
cdef.comp.cinfo.deflate.level = 8;
max_cache = 8;
cache_flag = 0;
 
if (mode == 1) SDsetchunk(sds, cdef, HDF_CHUNK | HDF_COMP);	 /* for writing to */
else           SDsetchunkcache(sds, max_cache, cache_flag);      /* for reading from */

return;
}


/*****************************************************************************************************/
void datatypestring(char typestring[256], int32 type)
{
switch (type) {
   case DFNT_CHAR:
      strcpy(typestring, "DFNT_CHAR\0");
      break;
   case DFNT_UCHAR:
      strcpy(typestring, "DFNT_UCHAR\0");
      break;
   case DFNT_FLOAT:
      strcpy(typestring, "DFNT_FLOAT\0");
      break;
   case DFNT_FLOAT64:
      strcpy(typestring, "DFNT_FLOAT64\0");
      break;
   case DFNT_INT8:
      strcpy(typestring, "DFNT_INT8\0");
      break;
   case DFNT_UINT8:
      strcpy(typestring, "DFNT_UINT8\0");
      break;      
   case DFNT_INT16:
      strcpy(typestring, "DFNT_INT16\0");
      break;
   case DFNT_UINT16:
      strcpy(typestring, "DFNT_UINT16\0");
      break;      
   case DFNT_INT32:
      strcpy(typestring, "DFNT_INT32\0");
      break;
   case DFNT_UINT32:
      strcpy(typestring, "DFNT_UINT32\0");
      break;     
   case DFNT_INT64:
      strcpy(typestring, "DFNT_INT64\0");
      break;
   case DFNT_UINT64:
      strcpy(typestring, "DFNT_UINT64\0");
      break;
   }
return;   
}

void insert_datafield(char *tmpstr, int n_datafields, char xname[256], int32 type, char ldatadim[256], char cdatadim[256])
{
char typestring[256];
void datatypestring(char typestring[256], int32 type);

sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tOBJECT=DataField_%d\n", n_datafields);
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tDataFieldName=\"%s\"\n", xname);
datatypestring( typestring,  type );
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tDataType=%s\n", typestring);
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tDimList=(\"%s\",\"%s\")\n", ldatadim, cdatadim);
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tEND_OBJECT=DataField_%d\n", n_datafields);
return;
}

void insert_dimension(char *tmpstr, int n_dims, char name[256], int32 size)
{
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tOBJECT=Dimension_%d\n", n_dims);
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tDimensionName=\"%s\"\n", name);
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tSize=\"%d\"\n", size);
sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tEND_OBJECT=Dimension_%d\n", n_dims);
}

int altHDFEOSstart(char *name, int nlines_1km, int ncols_1km, int *band_numbers, int compress)
{
#define SWATH_NAME "MODIS SWATH TYPE L2"
char location[]="modis_srefl_io.c, altHDFEOSstart\0";
int32 sdid, sdsid, dimid, dims[2], rank = 2;
int32 vfile_id, vgid[4];
char xname[256];
char ldatadim250m[256] = "\0";
char ldatadim500m[256] = "\0";
char ldatadim1km[256] = "\0";
char cdatadim250m[256] = "\0";
char cdatadim500m[256] = "\0";
char cdatadim1km[256] = "\0";
char ldatadim[256] = "\0";
char cdatadim[256] = "\0";
int32 i, l, c, n_dims, n_datafields, factor, isgeo;
char prefix[5];
extern output_type_t output_type[ALLBANDS];
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
extern int32 process_bands[BANDPROCESS];
extern int n_bands_res[3]; /* 250m, 500m, 1km */
char *tmpstr;
char typestring[256];
void insert_datafield(char *tmpstr, int n_datafields, char xname[256], int32 type, char ldatadim[256], char cdatadim[256]);
void insert_dimension(char *tmpstr, int n_dims, char name[256], int32 size);
void set_chunking(int mode, int32 sds, int32 lines, int32 cols);


  if ( (sdid = SDstart(name, DFACC_RDWR)) == -1) return(-1);

  l = (int32)nlines_1km;
  c = (int32)ncols_1km;



  /* Start Vgroup stuff 
   ***********************/
  if ( (vfile_id = Hopen(name, DFACC_RDWR, 0)) == -1) return(-1);

  /* Start the Vgroup access */  
  if (Vstart(vfile_id) == -1) return -1;
  
  /* Create Root Vgroup for Grid */
  vgid[0] = Vattach(vfile_id, -1, "w");
  if (vgid[0] == -1) return -1;

  /* Set Name and Class (SWATH) */
  if (Vsetname(vgid[0], SWATH_NAME) == -1) return -1;
  if (Vsetclass(vgid[0], "SWATH") == -1) return -1;

  /* Create Geolocation Fields Vgroup */
  vgid[1] = Vattach(vfile_id, -1, "w");
  if (vgid[1] == -1)  return -1;
  if (Vsetname(vgid[1], "Geolocation Fields") == -1) return -1;
  if (Vsetclass(vgid[1], "SWATH Vgroup") == -1) return -1;
  if (Vinsert(vgid[0], vgid[1]) == -1) return -1;

  /* Create Data Fields Vgroup */
  vgid[2] = Vattach(vfile_id, -1, "w");
  if (vgid[2] == -1)  return -1;
  if (Vsetname(vgid[2], "Data Fields") == -1) return -1;
  if (Vsetclass(vgid[2], "SWATH Vgroup") == -1) return -1;
  if (Vinsert(vgid[0], vgid[2]) == -1) return -1;

  /* Create Attributes Vgroup */
  vgid[3] = Vattach(vfile_id, -1, "w");
  if (vgid[3] == -1)  return -1;
  if (Vsetname(vgid[3], "Swath Attributes") == -1) return -1;
  if (Vsetclass(vgid[3], "SWATH Vgroup") == -1) return -1;
  if (Vinsert(vgid[0], vgid[3]) == -1) return -1;
  
  /* The above is pretty much exactly what SWcreate() does... */

  /* Start off the StructMetadata */
  tmpstr = (char *)malloc(80000*sizeof(char));

  strcpy(tmpstr,"GROUP=SwathStructure\n");
  strcat(tmpstr,"\tGROUP=SWATH_1\n");
  sprintf(&tmpstr[strlen(tmpstr)],"\t\tSwathName=\"%s\"\n", SWATH_NAME);
  strcat(tmpstr,"\t\tGROUP=Dimension\n");

  n_dims = 1;

  if (n_bands_res[0] > 0) {
     strcpy(prefix, "250m\0");
     sprintf(ldatadim250m, "%s %s:%s\0", prefix, LHDF_REF_DLINE, SWATH_NAME);
     insert_dimension(tmpstr, n_dims, ldatadim250m, l*4);
     n_dims++;
   
     strcpy(prefix, "250m\0");
     sprintf(cdatadim250m, "%s %s:%s\0", prefix, LHDF_REF_MAX_SAMP, SWATH_NAME);
     insert_dimension(tmpstr, n_dims, cdatadim250m, c*4);
     n_dims++;
  }
  if (n_bands_res[1] > 0) {
     strcpy(prefix, "500m\0");   
     sprintf(ldatadim500m, "%s %s:%s\0", prefix, LHDF_REF_DLINE, SWATH_NAME);
     insert_dimension(tmpstr, n_dims, ldatadim500m, l*2);
     n_dims++;
   
     strcpy(prefix, "500m\0");
     sprintf(cdatadim500m, "%s %s:%s\0", prefix, LHDF_REF_MAX_SAMP, SWATH_NAME);
     insert_dimension(tmpstr, n_dims, cdatadim500m, c*2);
     n_dims++;
  }
  if (n_bands_res[2] > 0) {
     strcpy(prefix, "1km\0");  
     sprintf(ldatadim1km, "%s %s:%s\0", prefix, LHDF_REF_DLINE, SWATH_NAME);
     insert_dimension(tmpstr, n_dims, ldatadim1km, l);
     n_dims++;
   
     strcpy(prefix, "1km\0");
     sprintf(cdatadim1km, "%s %s:%s\0", prefix, LHDF_REF_MAX_SAMP, SWATH_NAME);
     insert_dimension(tmpstr, n_dims, cdatadim1km, c);
     n_dims++;
  }
  
  strcat(tmpstr,"\t\tEND_GROUP=Dimension\n");
  strcat(tmpstr,"\t\tGROUP=DimensionMap\n");
  strcat(tmpstr,"\t\tEND_GROUP=DimensionMap\n");
  strcat(tmpstr,"\t\tGROUP=IndexDimensionMap\n");
  strcat(tmpstr,"\t\tEND_GROUP=IndexDimensionMap\n");
  strcat(tmpstr,"\t\tGROUP=GeoField\n");

  /* Set up lat/lon as GeoFields */	  
  i = 1;
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tOBJECT=GeoField_%d\n",i);
  sprintf(xname, "%s\0", sdsname[LAT-BAND20ALB]);
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tGeoFieldName=\"%s\"\n", xname);
  strcat(tmpstr,"\t\t\t\tDataType=DFNT_FLOAT\n");
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tDimList=(\"%s\",\"%s\")\n", ldatadim1km, cdatadim1km);
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tEND_OBJECT=GeoField_%d\n",i);

  i = 2;
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tOBJECT=GeoField_%d\n",i);
  sprintf(xname, "%s\0", sdsname[LON-BAND20ALB]);
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tGeoFieldName=\"%s\"\n", xname);
  strcat(tmpstr,"\t\t\t\tDataType=DFNT_FLOAT\n");
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\t\tDimList=(\"%s\",\"%s\")\n", ldatadim1km, cdatadim1km);
  sprintf(&tmpstr[strlen(tmpstr)],"\t\t\tEND_OBJECT=GeoField_%d\n",i);
	
  strcat(tmpstr,"\t\tEND_GROUP=GeoField\n");
  strcat(tmpstr,"\t\tGROUP=DataField\n");
	
  n_datafields = 1;
	
  /* Make these first.  Perhaps consider re-ordering SDSs to collection 4 order */
  strcpy(prefix, "1km\0");
  for (i=AOT_1;i<=AOT_M;i++) {
         sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
         dims[0] = l;
         dims[1] = c;
         sdsid = SDcreate(sdid, xname, output_type[i].type, rank, dims);
         if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, ldatadim1km);     
         if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, cdatadim1km);
         if (compress) set_chunking(1, sdsid, NBSCANS_1KM, NBSPOTS_1KM);

    	 if (Vaddtagref(vgid[2], DFTAG_NDG, SDidtoref(sdsid)) == -1) return(-2); 

         SDendaccess(sdsid);
	 
         insert_datafield(tmpstr, n_datafields, xname, output_type[i].type, ldatadim1km, cdatadim1km);
         n_datafields++;
     }
     
  i = WATERVAPOR;
  sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
#ifdef OLDSDSNAMES 
  sprintf(xname, "1km water_vapor\0");
#endif           
  dims[0] = l;
  dims[1] = c;
  sdsid = SDcreate(sdid, xname, output_type[i].type, rank, dims);
  if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, ldatadim1km);     
  if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, cdatadim1km);
  if (compress) set_chunking(1, sdsid, NBSCANS_1KM, NBSPOTS_1KM);
  if (Vaddtagref(vgid[2], DFTAG_NDG, SDidtoref(sdsid)) == -1) return(-2); 
  SDendaccess(sdsid);

  insert_datafield(tmpstr, n_datafields, xname, output_type[i].type, ldatadim1km, cdatadim1km);
  n_datafields++;


  i = AOTCM;
  sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
  dims[0] = l;
  dims[1] = c;
  sdsid = SDcreate(sdid, xname, output_type[i].type, rank, dims);
  if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, ldatadim1km);     
  if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, cdatadim1km);
  if (compress) set_chunking(1, sdsid, NBSCANS_1KM, NBSPOTS_1KM);
  if (Vaddtagref(vgid[2], DFTAG_NDG, SDidtoref(sdsid)) == -1) return(-2); 
  SDendaccess(sdsid);
	 
  insert_datafield(tmpstr, n_datafields, xname, output_type[i].type, ldatadim1km, cdatadim1km);
  n_datafields++;

  i = AOTQA;
  sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
  dims[0] = l;
  dims[1] = c;
  sdsid = SDcreate(sdid, xname, output_type[i].type, rank, dims);
  if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, ldatadim1km);     
  if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, cdatadim1km);
  if (compress) set_chunking(1, sdsid, NBSCANS_1KM, NBSPOTS_1KM);
  if (Vaddtagref(vgid[2], DFTAG_NDG, SDidtoref(sdsid)) == -1) return(-2); 
  SDendaccess(sdsid);
	 
  insert_datafield(tmpstr, n_datafields, xname, output_type[i].type, ldatadim1km, cdatadim1km);
  n_datafields++;


  for (i=0;i<ALLBANDS;i++) {
     isgeo = 0;
     if (i==BANDPROCESS) continue;
     if ((i==AOT_1)||(i==AOT_3)||(i==AOT_8)||(i==AOT_M)||(i==AOTCM)||(i==AOTQA)||(i==WATERVAPOR)) continue;
     if ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH)) {   /* here only worry about fine/coarse */

#ifdef TOTEMP
     if (i > CRSCM) continue;
#endif  
   
     if ( (i < BAND1_500M)||(i==QKMQA)) {
       if (n_bands_res[0] == 0) continue;
       else {
          strcpy(prefix, "250m\0");
          strcpy(ldatadim, ldatadim250m);
          strcpy(cdatadim, cdatadim250m);
	  factor = 4;
         }
       }
     else if ((i < BAND1_1KM)||(i==HKMQA)) {
       if (n_bands_res[1] == 0) continue;
       else {
          strcpy(prefix, "500m\0");
          strcpy(ldatadim, ldatadim500m);
          strcpy(cdatadim, cdatadim500m);
	  factor = 2;
	  }
       }
     else  {
       if (n_bands_res[2] == 0) continue;
       else {
          strcpy(ldatadim, ldatadim1km);
          strcpy(cdatadim, cdatadim1km);
          strcpy(prefix, "1km\0");
 	  factor = 1;
	  }
       }
     dims[0] = l*factor;
     dims[1] = c*factor;

     if ((i < BAND20) || (i == BAND26))sprintf(xname, "%s Surface Reflectance Band %d\0", prefix, band_numbers[i]);
     else if (i<BAND20ALB)             sprintf(xname, "%s Brightness Temperature Band %d\0", prefix, band_numbers[i]);
     else if ((i==LAT)||(i==LON))      sprintf(xname, "%s\0", sdsname[i-BAND20ALB]);
     else                              sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
   
#ifdef OLDSDSNAMES 
     if(i == WATERVAPOR)sprintf(xname, "1km water_vapor\0");
     if(i == BAND20)sprintf(xname, "BAND20\0");
     if(i == BAND31)sprintf(xname, "BAND31\0");
     if(i == BAND32)sprintf(xname, "BAND32\0");
     if(i == BAND20ALB)sprintf(xname, "BAND20ALBEDO\0");
#endif  
    
     if ((i==LAT)||(i==LON)) isgeo = 1;

     sdsid = SDcreate(sdid, xname, output_type[i].type, rank, dims);
     if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, ldatadim);     
     if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, cdatadim);
     if (compress) set_chunking(1, sdsid, NBSCANS_1KM*factor, NBSPOTS_1KM*factor);
     
     if (isgeo) {
         if (Vaddtagref(vgid[1], DFTAG_NDG, SDidtoref(sdsid)) == -1) return(-2); 
	 }
     else {
         if (Vaddtagref(vgid[2], DFTAG_NDG, SDidtoref(sdsid)) == -1) return(-2); 
	 }
     SDendaccess(sdsid);
 	 
     if (!isgeo) {	 /* because we've already inserted information about lat/lon as geofields */ 
       insert_datafield(tmpstr, n_datafields, xname, output_type[i].type, ldatadim, cdatadim);
       n_datafields++;
       }
     }
   }
    
  /* Detach Vgroups */
  if (Vdetach(vgid[1]) == -1) return -1;
  if (Vdetach(vgid[2]) == -1) return -1;
  if (Vdetach(vgid[3]) == -1) return -1;		
  if (Vdetach(vgid[0]) == -1) return -1;
  if (Vend(vfile_id) == -1)   return -1;
  Hclose(vfile_id);
  	
  /* Finish off the StructMetadata */	
  strcat(tmpstr,"\t\tEND_GROUP=DataField\n");
  strcat(tmpstr,"\t\tGROUP=MergedFields\n");
  strcat(tmpstr,"\t\tEND_GROUP=MergedFields\n");
  strcat(tmpstr,"\tEND_GROUP=SWATH_1\n");
  strcat(tmpstr,"END_GROUP=SwathStructure\n");
  strcat(tmpstr,"GROUP=GridStructure\n");
  strcat(tmpstr,"END_GROUP=GridStructure\n");
  strcat(tmpstr,"GROUP=PointStructure\n");
  strcat(tmpstr,"END_GROUP=PointStructure\n");
  strcat(tmpstr,"END\n");
	
  /* Write the HDFEOSversion and the  StructMetadata */	
  SDsetattr(sdid,"HDFEOSVersion",DFNT_CHAR8,11,(VOIDP)"HDFEOS_V2.9");
  SDsetattr(sdid,"StructMetadata.0",DFNT_CHAR8,strlen(tmpstr),(VOIDP)tmpstr);

  /* Cleanup */
  free(tmpstr);
  SDend(sdid); 
  
  return(0);   
}

/*****************************************************************************************************/
/*****************************************************************************************************/
/*****************************************************************************************************/


void HDFEOSstart(char *name, int nlines_1km, int ncols_1km, int *band_numbers, int compress)
{
char location[]="modis_srefl_io.c, HDFEOSstart\0";
int32 swid, swathid;
char xname[256];
char datadim250m[256];
char datadim500m[256];
char datadim1km[256];
char datadim[256];
int32 i, l, c, compcode;
int compparm[5];
int32 chunkdims[2];
char prefix[5];
extern output_type_t output_type[ALLBANDS];
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
extern int32 process_bands[BANDPROCESS];
extern int n_bands_res[3]; /* 250m, 500m, 1km */

compparm[0] = 1;
compcode = HDFE_COMP_DEFLATE;

l = (int32)nlines_1km;
c = (int32)ncols_1km;

swid = SWopen(name, DFACC_CREATE);
swathid = SWcreate(swid, "MODIS SWATH TYPE L2");

if (n_bands_res[0] > 0) {
   strcpy(prefix, "250m\0");
   sprintf(xname, "%s %s", prefix, LHDF_REF_DLINE);
   strcpy(datadim250m, xname);
   strcat(datadim250m, ",\0");
   SWdefdim(swathid, xname, l*4);
   sprintf(xname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
   strcat(datadim250m, xname);
   strcat(datadim250m, "\0");
   SWdefdim(swathid, xname, c*4);
   chunkdims[0] = l*4;
   chunkdims[1] = c*4;
}

if (n_bands_res[1] > 0) {
   strcpy(prefix, "500m\0");
   sprintf(xname, "%s %s", prefix, LHDF_REF_DLINE);
   strcpy(datadim500m, xname);
   strcat(datadim500m, ",\0");
   SWdefdim(swathid, xname, l*2);
   sprintf(xname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
   strcat(datadim500m, xname);
   strcat(datadim500m, "\0");
   SWdefdim(swathid, xname, c*2);
   chunkdims[0] = l*2;
   chunkdims[1] = c*2;
}

if (n_bands_res[2] > 0) {
   strcpy(prefix, "1km\0");
   sprintf(xname, "%s %s", prefix, LHDF_REF_DLINE);
   strcpy(datadim1km, xname);
   strcat(datadim1km, ",\0");
   SWdefdim(swathid, xname, l);
   sprintf(xname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
   strcat(datadim1km, xname);
   strcat(datadim1km, "\0");
   SWdefdim(swathid, xname, c);
   chunkdims[0] = l;
   chunkdims[1] = c;
}

/* Make these first.  Perhaps consider re-ordering SDSs to collection 4 order */
for (i=AOT_1;i<=AOT_M;i++) {
    strcpy(datadim, datadim1km);
    strcpy(prefix, "1km\0");
    sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
    if (compress) SWdefcomp(swathid, compcode, compparm);
    SWdefdatafield(swathid, xname, datadim, output_type[i].type, HDFE_NOMERGE);
   }
i = WATERVAPOR;
strcpy(datadim, datadim1km);
strcpy(prefix, "1km\0");
sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
#ifdef OLDSDSNAMES 
if(i == WATERVAPOR)sprintf(xname, "1km water_vapor\0");
#endif           
if (compress) SWdefcomp(swathid, compcode, compparm);
SWdefdatafield(swathid, xname, datadim, output_type[i].type, HDFE_NOMERGE);
i = AOTCM;
strcpy(datadim, datadim1km);
strcpy(prefix, "1km\0");
sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
if (compress) SWdefcomp(swathid, compcode, compparm);
SWdefdatafield(swathid, xname, datadim, output_type[i].type, HDFE_NOMERGE);
i = AOTQA;
strcpy(datadim, datadim1km);
strcpy(prefix, "1km\0");
sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
if (compress) SWdefcomp(swathid, compcode, compparm);
SWdefdatafield(swathid, xname, datadim, output_type[i].type, HDFE_NOMERGE);

for (i=0;i<ALLBANDS;i++) {
   if (i==BANDPROCESS) continue;
   if ((i==AOT_1)||(i==AOT_3)||(i==AOT_8)||(i==AOT_M)||(i==AOTCM)||(i==AOTQA)||(i==WATERVAPOR)) continue;
   if ((output_type[i].FC == FINE_ONLY)||(output_type[i].FC == FCBOTH)) {   /* here only worry about fine/coarse */
   
   
   if ( (i < BAND1_500M)||(i==QKMQA)) {
       if (n_bands_res[0] == 0) continue;
       else {
          strcpy(datadim, datadim250m);
          strcpy(prefix, "250m\0");
          }
       }
   else if ((i < BAND1_1KM)||(i==HKMQA)) {
       if (n_bands_res[1] == 0) continue;
       else {
          strcpy(datadim, datadim500m);
          strcpy(prefix, "500m\0");
	  }
       }
   else  {
       if (n_bands_res[2] == 0) continue;
       else {
          strcpy(datadim, datadim1km);
          strcpy(prefix, "1km\0");
	  }
       }
   if ((i < BAND20) || (i == BAND26))sprintf(xname, "%s Surface Reflectance Band %d\0", prefix, band_numbers[i]);
   else if (i<BAND20ALB)             sprintf(xname, "%s Brightness Temperature Band %d\0", prefix, band_numbers[i]);
   else                              sprintf(xname, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
   
#ifdef OLDSDSNAMES 
    if(i == WATERVAPOR)sprintf(xname, "1km water_vapor\0");
    if(i == BAND20)sprintf(xname, "BAND20\0");
    if(i == BAND31)sprintf(xname, "BAND31\0");
    if(i == BAND32)sprintf(xname, "BAND32\0");
    if(i == BAND20ALB)sprintf(xname, "BAND20ALBEDO\0");
#endif           
     
    if (compress) SWdefcomp(swathid, compcode, compparm);
       
    SWdefdatafield(swathid, xname, datadim, output_type[i].type, HDFE_NOMERGE);
    }
   }
    
SWdetach(swathid);
SWclose(swid);
}




int open_coarse_outputs(int output_lun, dims_struct **output, int version, int nlines_1km, int ncols_1km, 
                        int *local_process_bands, int *band_numbers, int daynightflag)
/*
!C****************************************************************************

!Routine: open_coarse_outputs()

!Description: Throws open the coarse resolution output file at output_lun/version.

!Input Parameters:
  int output_lun
  int version
  int local_process_bands       flag for processing/not processing individual bands
  int nlines_1km
  int ncols_1km
  int *band_numbers             array of band numbers (dimension BANDPROCESS) -- the actual
                                number of the band for array BANDPROCESS.
  int daynightflag              1 for day or both, 0 for night

!Input/Output Parameters:
  dims_struct **local_output    Pointer to the I/O structure 
  
!Output Parameters:
  none

!Returns:
  0 success
  1 problem with PGS_PC_GetReference.
 -1 problem with SDcreate  
 (SDstart problem is FATAL).

!Revision History:
  Revision 1.0  11-FEB-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
        called by routine atmospheric_correction() (atmcor.c)

!END**************************************************************************/
{
#include "qa_map.h"
int i;
int32 dimid;
PGSt_integer local_version;
int32 sdid, sdsid, rank, number_type, n_attr, dims[2];
int32 factor, nblines, nbcols;
char prefix[20];
char rez[20];
char location[100]="modis_srefl_io.c, open_coarse_outputs\0";
char filename[PGSd_PC_FILE_PATH_MAX]={'\0'};
char dimname[60];
extern output_type_t output_type[ALLBANDS];
extern char sdsname[ALLBANDS-BANDPROCESS+1][50];
char message[1200];

/* Get filename */
local_version = version;
if(PGS_PC_GetReference((PGSt_PC_Logical)output_lun, &local_version, filename)!=PGS_S_SUCCESS){
    sprintf(msg, "LUN %d, version %d", (int32)output_lun, version);
    modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
    return(1);
  }
  
if( (sdid=SDstart(filename, DFACC_CREATE)) == -1 ) {
    sprintf(msg, "SDstart, LUN %d, version %d", (int32)output_lun, version);
    modsmf((PGSt_SMF_code)MODIS_F_OPEN_HDF_FILE, msg, location);
  }
else {
    sprintf(msg, " file %s, version %d", filename, version);
    modsmf(MODIS_S_SREFLCRS_L2GRAN, msg, location);
   }   
/* initialize the output structure * /
init_struct(*output, location);*/
(*output)->sdid = sdid;
strcpy(prefix, "Coarse Resolution\0");
strcpy(rez, "5km\0");
(*output)->filename = (char *)malloc(strlen(filename)+1);
strcpy((*output)->filename, filename);

if(nlines_1km % SREFL_COARSESIZE == 0) nblines = nlines_1km/SREFL_COARSESIZE;
else nblines=((int32)(nlines_1km/SREFL_COARSESIZE))+1;
nbcols = ((int32)(NBSPOTS_1KM/SREFL_COARSESIZE)) +1;
(*output)->nblines = nblines;
(*output)->nbcols = nbcols;
dims[0] = nblines;
dims[1] = nbcols;


/* Set up the HDF interface */
for (i=BAND1_1KM;i<BAND8;i++) {
   
   if (daynightflag == 0) continue;

   if (local_process_bands[i] == 1){
      sprintf(filename, "%s Surface Reflectance Band %d\0", prefix, band_numbers[i]);
      if ((sdsid = SDcreate(sdid, filename, output_type[i].type, 2, dims))==-1) return(-1);
      else {
         (*output)->sdsid[i] = sdsid;
         (*output)->type[i] = output_type[i].type;
        }

      set_attributes(sdsid, filename, rez, output_type[i], DFNT_FLOAT, 0);
      
      sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
      if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
      sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
      if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);


    }
   }
for (i=BAND8;i<BANDPROCESS;i++) {   

   if ((local_process_bands[i] == 1)    &&    /* local_process_bands is valid only up to  BANDPROCESS */
       ((output_type[i].FC == COARSE_ONLY)||(output_type[i].FC == FCBOTH)) &&  
       (((daynightflag == 1)&&(output_type[i].DN == DAY_ONLY)||(output_type[i].DN == DNBOTH)) || 
       ((daynightflag == 0)&&(output_type[i].DN == NIGHT_ONLY)||(output_type[i].DN == DNBOTH)))   
				    ){
      if ((i == BAND8) || (i == BAND26))sprintf(filename, "%s Surface Reflectance Band %d\0", prefix, band_numbers[i]);
      else                              sprintf(filename, "%s Brightness Temperature Band %d\0", prefix, band_numbers[i]);
      
      if ((sdsid = SDcreate(sdid, filename, output_type[i].type, 2, dims))==-1) return(-1);
      else {
         (*output)->sdsid[i] = sdsid;
         (*output)->type[i] = output_type[i].type;
       }

      set_attributes(sdsid, filename, rez, output_type[i], DFNT_FLOAT, 0);

      sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
      if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
      sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
      if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);


    }
   }
for (i=BAND20ALB;i<ALLBANDS;i++) {
   if  (((output_type[i].FC == COARSE_ONLY)||(output_type[i].FC == FCBOTH)) &&  
       (((daynightflag == 1)&&((output_type[i].DN == DAY_ONLY)||(output_type[i].DN == DNBOTH))) || 
       ((daynightflag == 0)&&((output_type[i].DN == NIGHT_ONLY)||(output_type[i].DN == DNBOTH))))   
				    ){
   
      sprintf(filename, "%s %s\0", prefix, sdsname[i-BAND20ALB]);
         
      if ((sdsid = SDcreate(sdid, filename, output_type[i].type, 2, dims))==-1) return(-1);
      else {
         (*output)->sdsid[i] = sdsid;
         (*output)->type[i] = output_type[i].type;
        }
      /*printf("%d %s %d %d\n", i, filename, (*output)->sdsid[i], (*output)->type[i]);*/

      if ((i==CRSQA)||(i==CRSCM))
         set_attributes(sdsid, filename, rez, output_type[i], DFNT_FLOAT, 1);
      else
         set_attributes(sdsid, filename, rez, output_type[i], DFNT_FLOAT, 0);

      if (i==CRSQA) {
         strcpy(message, QAMAP_COARSE);
         SDsetattr(sdsid, "QA index\0", DFNT_CHAR, strlen(message), (void *)message);
         }

      sprintf(dimname, "%s %s", prefix, LHDF_REF_DLINE);
      if((dimid=SDgetdimid(sdsid, 0))!=-1) SDsetdimname(dimid, dimname);
      
      sprintf(dimname, "%s %s", prefix, LHDF_REF_MAX_SAMP);
      if((dimid=SDgetdimid(sdsid, 1))!=-1) SDsetdimname(dimid, dimname);
      
      }
}

return(0);
}


int close_HDF_interface(dims_struct *input)
/*
!C****************************************************************************

!Routine: close_HDF_interface()

!Description: Runs SDendaccess and SDend on the inputs.

!Input Parameters:
   none
 
!Input/Output Parameters:
  dims_struct *inputs   Pointer to the array of input structures (updated)

!Output Parameters:
  none

!Returns:
  Success: 0
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
    called by routine atmospheric_correction() (atmcor.c)
  
!END**************************************************************************/
{
int j;
for (j=0;j<ALLBANDS;j++) {
   if (input->sdsid[j] != -1) SDendaccess(input->sdsid[j]);
   }
if (input->sdid != -1) SDend(input->sdid);

/* 13-APR-06 */
for (j=0;j<TEMPS;j++) {
   if (input->temp_fd[j] != (FILE *)NULL) {
        if(  PGS_IO_Gen_Close(input->temp_fd[j]) != PGS_S_SUCCESS ) {
	    modsmf(MODIS_F_CLOSE_TEMP_FILE, "temp file", "modis_srefl_io.c, close_HDF_interface");
	   }
        /*PGS_IO_Gen_Temp_Delete( input->temp_lun[j] );*/
        /*printf("Closing file at lun %d\n", input->temp_lun[j]);*/
        input->temp_fd[j] = (FILE *)NULL;
        input->temp_lun[j] = (int32)-1;
	}
   }

return(0);

}


int close_io_struct(dims_struct ***inputs, int dim1, int dim2)
/*
!C****************************************************************************

!Routine: close_io_struct()

!Description: Frees memory for a two-dimensional array of I/O structures.

!Input Parameters:
   dim1                   first dim
   dim2                   second dim
 
!Input/Output Parameters:
  dims_struct ***inputs   Pointer to the array of input structures (updated)

!Output Parameters:
  none

!Returns:
  Success: 0
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
      called by routine main() (modis_srefl.c)
      called by routine sort_inputs() (modis_srefl_io.c)

!END**************************************************************************/
{
int i,j;

for(i=0;i<dim1;i++) {
  for(j=0;j<dim2;j++) {
     /* printf("*** %p\n", inputs[i][j]->filename); */ 
     if (inputs[i][j]->filename != NULL) free(inputs[i][j]->filename); 
     /*printf("CLS IO %d %d *** %p %p\n", i,j,inputs[i][j]->xxxxx,inputs[i][j]); */
     if (inputs[i][j]->xxxxx != NULL) freeECSmetadata(inputs[i][j]->xxxxx); 
     /* printf("*** %p\n", inputs[i][j]);  */
     if (inputs[i][j] != NULL) free(inputs[i][j]); 
     }
     /*  printf("** %p\n", inputs[i]);  */
   if (inputs[i] != NULL) free(inputs[i]); 
   }
     /*  printf("* %p\n", inputs);  */
if (inputs != NULL) free(inputs); 

return(0);

}

int close_io_struct2(dims_struct **inputs, int dim1)
/*
!C****************************************************************************

!Routine: close_io_struct2()

!Description: Frees memory for one-dimensional array of I/O structures.

!Input Parameters:
   dim1                 dim
 
!Input/Output Parameters:
  dims_struct **inputs  Pointer to the array of input structures (updated)

!Output Parameters:
  none

!Returns:
  Success: 0
  
!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
      called by routine atmospheric_correction() (atmcor.c)

!END**************************************************************************/
{
int i,j;

for(i=0;i<dim1;i++) {
      /*printf("CLS IO2 %d *** %p \n", i, inputs[i]->filename); */
    if (inputs[i]->filename != NULL) free(inputs[i]->filename); 
      /*printf("CLS IO2 %d *** %p %p \n", i, inputs[i]->xxxxx, inputs[i]); */
    if (inputs[i]->xxxxx != NULL) freeECSmetadata(inputs[i]->xxxxx); 
      /*printf("*** %p\n", inputs[i]); */
    if (inputs[i] != NULL) free(inputs[i]); 
    }
      /*printf("* %p\n", inputs);  */

if (inputs != NULL) free(inputs); 

return(0);

}

void structcpy( dims_struct *to, dims_struct *from )
/*
!C****************************************************************************

!Routine: structcpy()

!Description: copies dim_struct data at "from" to "to", because memcpy 
              doesn't quite work.

!Input Parameters:
  dims_struct *from
  
!Input/Output Parameters:
  dims_struct *to  (updated)

!Output Parameters:
  none

!Returns:
  nothing

!Revision History:
  Revision 1.0  08-JUL-2002
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
      called by routine atmospheric_correction() (atmcor.c)
      called by routine sort_inputs() (modis_srefl_io.c)

      Does NOT copy the data contained in from[n]->data.

!END**************************************************************************/
{
int i;
char location[100]="modis_srefl_io.c, structcpy\0";

to->OK = from->OK;
to->sdid = from->sdid;
to->lun = from->lun;
to->ver = from->ver;
to->n_sds = from->n_sds;
for(i=0;i<BANDPROCESS;i++) {
   to->refscale[i] = from->refscale[i];
   to->refoffset[i] = from->refoffset[i];
   to->radscale[i] = from->radscale[i];
   to->radoffset[i] = from->radoffset[i];
   to->sdsid[i] = from->sdsid[i];
   to->sds_offset[i] = from->sds_offset[i];
   to->band_name[i] = from->band_name[i];
   to->type[i] = from->type[i];
   to->data[i] = (void *)NULL;
   }
to->filename = (char *)NULL;
if (from->filename != NULL) {
   if (to->filename == NULL) to->filename = (char *)malloc(strlen(from->filename)+1);
   else {   
       if (strlen(to->filename) != strlen(from->filename));
           to->filename = (char *)realloc(to->filename, strlen(from->filename)+1);
	}
   strcpy(to->filename, from->filename);
/* Dmalloc on HPUX machines (e. g., tracer) need this.  */
#ifdef HPUX
   strcat(to->filename, '\0');
#endif  
   }
   
/* new, 13-APR-06 */   
for(i=0;i<TEMPS;i++) {
    to->temp_fd[i] = from->temp_fd[i];
    to->temp_lun[i] = from->temp_lun[i];
   }
   
to->nblayers = from->nblayers;
to->nblines = from->nblines;
to->nbcols = from->nbcols;

ECScpy(to->xxxxx, from->xxxxx, 0);

}

int set_attributes(int32 sdsid, char *longname, char *rez, output_type_t output_type, int32 type, int8 qa_flag)
/*
!C****************************************************************************

!Routine: set_attributes()

!Description: sets local SDS attributes.

!Input Parameters:
  int32 sdsid                  SDS identifier
  char *longname               attribute "long_name"
  char *rez                    attribute "Nadir Data Resolution"
  output_type_t output_type    structure containing information on attributes "units", "scale_factor", etc.
  int32 type                   attribute "calibrated_nt"
  int8 qa_flag                 Flag for whether SDS is a QA SDS (=1) or not (=0)
  
!Input/Output Parameters:
  none

!Output Parameters:
  none

!Returns:
  0 for success, nonzero for error.

!Revision History:
  Revision 1.0  28-APR-2003
  Jim Ray

!Team-unique Header:
  This software was developed by:

    MODIS Science Data Support Team for the National Aeronautics and Space 
    Administration, Goddard Space Flight Center, under contract NAS5-32373.

  Developer:
      Jim Ray
      Science Systems and Applications, Inc.
      MODIS Land Science Team
      NASA Goddard Space Flight Center
      Greenbelt, Maryland 20771
      jim@cratmos.gsfc.nasa.gov

!References and Credits 

!Design Notes: 
      called by routine open_outputs() (modis_srefl_io.c)
      called by routine open_coarse_outputs() (modis_srefl_io.c)

!END**************************************************************************/
{
int d; /* 0 = int16, 1 = float, 2 = uint16, 3 = uint8, 4 = uint32 */
int32 ret = 0;
float64 zero = 0.0;
char bmp[12] = "bitmap\0";
int16 range[2], fill;
float frange[2], ffill;
uint16 urange[2], ufill;
uint8 u8range[2], u8fill;
uint32 u32range[2], u32fill;

d = 0;
range[0] = output_type.lowrange;
range[1] = output_type.hirange;
fill     = output_type.fillvalue;

if (output_type.type == DFNT_FLOAT) {
   d = 1;
   frange[0] = (float)output_type.lowrange;
   frange[1] = (float)output_type.hirange;
   ffill = (float)output_type.fillvalue;
    }
else if (output_type.type == DFNT_UINT16) {
   d = 2;
   urange[0] = (uint16)output_type.lowrange;
   urange[1] = (uint16)output_type.hirange;
   ufill = (uint16)output_type.fillvalue;
    }
else if (output_type.type == DFNT_UINT8) {
   d = 3;
   u8range[0] = (uint8)output_type.lowrange;
   u8range[1] = (uint8)output_type.hirange;
   u8fill = (uint8)output_type.fillvalue;
    }
else if (output_type.type == DFNT_UINT32) {
   d = 4;
   u32range[0] = (uint32)0;
   u32range[1] = (uint32)0xffffffff;
   u32fill = (uint32)3;
    }

ret |= SDsetattr(sdsid, "long_name\0", DFNT_CHAR, strlen(longname), longname);

if (qa_flag) ret |= SDsetattr(sdsid, "units\0", DFNT_CHAR, strlen(bmp), bmp);
else         ret |= SDsetattr(sdsid, "units\0", DFNT_CHAR, strlen(output_type.units), output_type.units);

ret |= SDsetattr(sdsid, "Nadir Data Resolution\0", DFNT_CHAR, strlen(rez), rez);

switch (d) {
   case (0) :
    ret |= SDsetattr(sdsid, "valid_range\0", DFNT_INT16, 2, range);
    ret |= SDsetattr(sdsid, "_FillValue\0", DFNT_INT16, 1, &fill);
    break;
   case (1) :
    ret |= SDsetattr(sdsid, "valid_range\0", DFNT_FLOAT, 2, frange);
    ret |= SDsetattr(sdsid, "_FillValue\0", DFNT_FLOAT, 1, &ffill);
    break;
   case (2) :
    ret |= SDsetattr(sdsid, "valid_range\0", DFNT_UINT16, 2, urange);
    ret |= SDsetattr(sdsid, "_FillValue\0", DFNT_UINT16, 1, &ufill);
    break;
   case (3) :
    ret |= SDsetattr(sdsid, "valid_range\0", DFNT_UINT8, 2, u8range);
    ret |= SDsetattr(sdsid, "_FillValue\0", DFNT_UINT8, 1, &u8fill);
    break;
   case (4) :
    ret |= SDsetattr(sdsid, "valid_range\0", DFNT_UINT32, 2, u32range);
    ret |= SDsetattr(sdsid, "_FillValue\0", DFNT_UINT32, 1, &u32fill);
    break;
}

if (qa_flag) return(ret);  /* short-circuit */

ret |= SDsetattr(sdsid, "scale_factor\0", DFNT_FLOAT64, 1, &(output_type.scale));
ret |= SDsetattr(sdsid, "scale_factor_err\0", DFNT_FLOAT64, 1, &zero);
ret |= SDsetattr(sdsid, "add_offset\0", DFNT_FLOAT64, 1, &zero);
ret |= SDsetattr(sdsid, "add_offset_err\0", DFNT_FLOAT64, 1, &zero);
ret |= SDsetattr(sdsid, "calibrated_nt\0", DFNT_INT32, 1, &type);

return(ret);
}

int32 read_dead_det(int32 sdid, int32 dead_bands[BANDPROCESS][4 * NBSCANS_1KM], int32 noisy_bands[BANDPROCESS][4 * NBSCANS_1KM])
/*
!C*****************************************************************************
!Description: reads the dead detector array from the MODIS L1B file and sets
              a switch structure indicating dead fields.
  
!Input Parameters:
sdid:int32 file handle

!Output Parameters: 
dead-bands: file switch structureof size [bands][detector of band]

!Input/Output Parameters: none

!Returns: 0 on failure
          1 on success

!Revision History:
        Revision 1, October 1998, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).
	1.  Bug fix to read the correct values for the fire bands

	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980

!Design Notes

1.  dead_bands is initialized to zero, indicating a non-dead detector.
The structure is then set to whatever value is held within the dead
detector attribute (currently a value of 1 is used to
indicate a dead detector)

!END****************************************************************************/
{

  int32 i, j;
  int8 dead_detectors[490];
  int32 p=0;
  int32 attr_number;
  extern int sat_type;

  for(i=0;i<BANDPROCESS;i++){
     for(j=0;j<NBSCANS_250M;j++){
        dead_bands[i][j]=0;
           }
        }
	
  for(i=0;i<490;i++)dead_detectors[i]=0;
  
  if ( (attr_number = SDfindattr(sdid, "Dead Detector List\0")) == -1) 
     return(-1);
  if (  SDreadattr(sdid, attr_number, dead_detectors) != PGS_S_SUCCESS)
     return(0);
 
  
  
  for(i=BAND1_250M;i<BAND1_500M;i++){
    for(j=0;j<(4*NBSCANS_1KM);j++)dead_bands[i][j]=(int32)dead_detectors[p++];
  }
  for(i=BAND3_500M;i<BAND1_1KM;i++){
    for(j=0;j<(2*NBSCANS_1KM);j++)dead_bands[i][j]=(int32)dead_detectors[p++];
  }
   
      
  /* 19-DEC-06 * /   
  
  No, don't do this (20-DEC-06) 
  
  for(i=BAND1_250M;i<BAND1_500M;i++){
    if (sat_type == 1) p += (40*(i+1));
    for(j=0;j<(4*NBSCANS_1KM);j++){
       dead_bands[i][j]=(int32)dead_detectors[p];
       if (sat_type == 1) p--;
       else               p++;
       }
  }
  if (sat_type == 1) p += 60;
  
  for(i=BAND3_500M;i<BAND1_1KM;i++){
    for(j=0;j<(2*NBSCANS_1KM);j++) {
       dead_bands[i][j]=(int32)dead_detectors[p];
       if (sat_type == 1) p--;
       else               p++;
       }
    if (sat_type == 1) p += 40;
  }
  
  / * And if so, need to do the same for 1km */ 
  
/*  p+=150;  20-SEP-00: we had beed reading dead detectors for bands 20, 21 and 22 
   incorrectly all along. */
  p+=140;

  for(j=0;j<NBSCANS_1KM;j++)dead_bands[BAND20][j]=(int32)dead_detectors[p++];
  for(j=0;j<NBSCANS_1KM;j++)dead_bands[BAND21][j]=(int32)dead_detectors[p++];
  for(j=0;j<NBSCANS_1KM;j++)dead_bands[BAND22][j]=(int32)dead_detectors[p++];

/*  p+=70;*/
  p+=80;

  for(j=0;j<NBSCANS_1KM;j++)dead_bands[BAND31][j]=(int32)dead_detectors[p++];
  for(j=0;j<NBSCANS_1KM;j++)dead_bands[BAND32][j]=(int32)dead_detectors[p++];

  p = 0;
  for(i=0;i<490;i++)dead_detectors[i]=0;
  
  if ( (attr_number = SDfindattr(sdid, "Noisy Detector List\0")) == -1) 
     return(-1);
  if (  SDreadattr(sdid, attr_number, dead_detectors) != PGS_S_SUCCESS)
     return(0);
 
  
  for(i=BAND1_250M;i<BAND1_500M;i++){
    for(j=0;j<(4*NBSCANS_1KM);j++)noisy_bands[i][j]=(int32)dead_detectors[p++];
  }
  for(i=BAND3_500M;i<BAND1_1KM;i++){
    for(j=0;j<(2*NBSCANS_1KM);j++)noisy_bands[i][j]=(int32)dead_detectors[p++];
  }

/*  p+=150;  20-SEP-00: we had beed reading dead detectors for bands 20, 21 and 22 
   incorrectly all along. */
  p+=140;

  for(j=0;j<NBSCANS_1KM;j++)noisy_bands[BAND20][j]=(int32)dead_detectors[p++];
  for(j=0;j<NBSCANS_1KM;j++)noisy_bands[BAND21][j]=(int32)dead_detectors[p++];
  for(j=0;j<NBSCANS_1KM;j++)noisy_bands[BAND22][j]=(int32)dead_detectors[p++];

/*  p+=70;*/
  p+=80;

  for(j=0;j<NBSCANS_1KM;j++)noisy_bands[BAND31][j]=(int32)dead_detectors[p++];
  for(j=0;j<NBSCANS_1KM;j++)noisy_bands[BAND32][j]=(int32)dead_detectors[p++];

  
  return(1);
  
}


