/*
!C******************************************************************

!Description: 
		polarization.c

         Routines to apply polarization correction to band 8.  

!Input Parameters:
!Output Parameters: 
!Revision History:
!Team-unique Header:
!References and Credits 
!Design Notes

         The following routines in this file are meant to be called from
	 main(), in modis_srefl.c:
	 
	      int read_polarization_data(char *name)
	      int read_Rayleigh_data(char *name)
              void free_polarization_data(pol_param_t param)
	      
	 and from atmcor_band8(), in atmcor.c:
	 
	      polarization_coefficient( float32 SUN, float32 THETA, 
                                       float32 PHI, int N_DET, float32 ZAOI, 
				       int N_SIDE, 
				       float32 *E_RAD, float32 *C_RAD)
     
	 All the rest of the routines in this file are "private".
	 

!END*************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>

#include "PGS_MODIS_37092.h"

#include "modis_srefl.h"
#include "protos.h"
/*#include "pol_driver.h"*/
#include "smf.h"
#include "mfhdf.h"

/*#include "proc_define.h"*/

#define WIND 7.0F                /* meters/second */

extern pol_param_t aoi;          /* all these "extern"s refer to variables */
extern pol_param_t am12_0;       /* that are declared in pol_global.h, which */
extern pol_param_t am13_0;       /* is included global to main(). */
extern pol_param_t am12_1;
extern pol_param_t am13_1;
extern pol_param_t polphs;
extern pol_param_t polmag;
extern pol_param_t tau;
extern pol_param_t dep;
extern pol_param_t ang;
extern pol_param_t sun;
extern pol_param_t sigma;
extern pol_param_t eye;
extern pol_param_t cue;
extern pol_param_t you;
extern int n_bands;
extern int n_AOIs;
extern int n_dets;
extern int n_sides;
extern int n_lambda;
extern int n_rad_ray;
extern int n_sun_ray;
extern int n_wind_ray;
extern int n_order_ray;

int read_polarization_data(int lun1, int lun2)
{
/*
!C******************************************************************

!Description: This function reads polarization look-up tables.

!Input Parameters:
lun1:        the LUN for one file to be read (209570, which points to "new_modis_pol_corr5a.hdf").
lun2:        the LUN for the other file to be read (209569, which points to "new_modis_pol_corr5b.hdf").

   (If you know the LUNs already, why are they arguments?  Just in case the LUNs are changed.)

!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01

Adapted for second polarization look-up table on 23-JAN-03.

!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    All parameters to read are globally defined in file pol_driver.h.
    To be called from main() at the very beginning of the program.

!END****************************************************************************
*/
int32 sd;
int ret;
int pol_ver;
int get_polarization_data(int32 sds, pol_param_t *param, char *name);
int32 n_sets1, n_gattr1;
char location[100]="polarization.c, read_polarization_data\0";
char name[256];
char msg[256];

/* Assuming lun1 points to "new_modis_pol_corr5a.hdf", which 
   I further assume has the data for "north of the sun" pixels, which
   is to be read into am12_0, am13_0. */
pol_ver = 1;
if(PGS_PC_GetReference(lun1, &pol_ver, name)!=PGS_S_SUCCESS){
  sprintf(msg, "LUN %d, version %d (polarization correction tables)", (int32)lun1, pol_ver);
  modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
       }

if ( (sd = SDstart(name, DFACC_RDONLY)) == -1) {
   modsmf((PGSt_SMF_code)MODIS_E_OPEN_HDF_FILE, name, location);
   return(2);
   }

ret=SDfileinfo(sd, &n_sets1, &n_gattr1);
if ((ret == -1)||(n_sets1 == 0)) {
   modsmf((PGSt_SMF_code)MODIS_E_HDF_READ, "polarization data file (a) is unreadable", location);
   SDend(sd);
   return(2);
   }

ret = get_polarization_data(sd, &aoi, "aoi\0");  /* AOI ***MUST!!!*** be read first */
ret += get_polarization_data(sd, &am12_0, "am12\0");
ret += get_polarization_data(sd, &am13_0, "am13\0");
/* These are not really needed (at least for now) */
/*ret += get_polarization_data(sd, &polphs, "pol_phs\0");*/
/*ret += get_polarization_data(sd, &polmag, "pol_mag\0");*/

if (ret > 0){
   modsmf((PGSt_SMF_code)MODIS_E_HDF_READ, "Error accessing polarization data (a)", location);
   SDend(sd);
   return(2);
   }
   
ret=SDend(sd);
if (ret == -1){
   modsmf((PGSt_SMF_code)MODIS_E_CLOSE_HDF_FILE, name, location);
   return(2);
   }

/* Assuming lun2 points to "new_modis_pol_corr5b.hdf", which 
   I further assume has the data for "south of the sun" pixels, which
   is to be read into am12_1, am13_1. */
pol_ver = 1;

/* will initialize HERE: this way, if the second file ISN'T staged, 
   the polarization_coefficient routine will know to NOT use data
   in am12_1, am13_1, which will be NULL. */
init_polarization_data(&am12_1, "am12\0");
init_polarization_data(&am13_1, "am13\0");

if(PGS_PC_GetReference(lun2, &pol_ver, name)!=PGS_S_SUCCESS){
   sprintf(msg, "LUN %d, version %d (polarization correction tables)", (int32)lun2, pol_ver);
   modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
   return(0);  /* no fault if it is not staged (FOR NOW) */
       }

if ( (sd = SDstart(name, DFACC_RDONLY)) == -1) {
   modsmf((PGSt_SMF_code)MODIS_E_OPEN_HDF_FILE, name, location);
   return(2);
   }
   
ret=SDfileinfo(sd, &n_sets1, &n_gattr1);
if ((ret == -1)||(n_sets1 == 0)) {
   modsmf((PGSt_SMF_code)MODIS_E_HDF_READ, "polarization data file (b) is unreadable", location);
   SDend(sd);
   return(2);
   }
   
/* For now (23-JAN-03), will not read aoi, because 
  the aoi in one file is identical to the aoi in the other anyway. */
ret += get_polarization_data(sd, &am12_1, "am12\0");
ret += get_polarization_data(sd, &am13_1, "am13\0");

if (ret > 0){
   modsmf((PGSt_SMF_code)MODIS_E_HDF_READ, "Error accessing polarization data (b)", location);
   SDend(sd);
   return(2);
   }
   
ret=SDend(sd);
if (ret == -1){
   modsmf((PGSt_SMF_code)MODIS_E_CLOSE_HDF_FILE, name, location);
   return(2);
   }

return(0);
}

int get_polarization_data(int32 sd, pol_param_t *param, char *name)
{
/*
!C******************************************************************

!Description: This function reads polarization tables within polarization files.

!Input Parameters:
sd:        HDF file tag (return value of SDstart).
param:     structure into which the parameter is read (global)
name:      name of the parameter to read


!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    All parameters to read are globally defined in file pol_driver.h.

!END****************************************************************************
*/
int i,j;
int iband,iside,iaoi,idet,idim;
int32 sds, dimid, ret;
int32 start[MAX_POL_DIMS], dims[MAX_POL_DIMS];
char SDSname[100], dimname[100];
int32 rank1, value, count, dims1[MAX_POL_DIMS], number_type1, nattr1;
float32 *scratch;
char msg[PGS_SMF_MAX_MSG_SIZE];
char location[100]="polarization.c, get_polarization_data\0";
void init_polarization_data(pol_param_t *param, char *name);

init_polarization_data(param, name);

if ( (sds = SDselect(sd, SDnametoindex(sd, param->name))) == -1) {
   sprintf(msg, " SDselecting polarization parameter '%s' ", param->name);
   modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
   return(1);
   }

/* get dimension info */
for (i=0;i<MAX_POL_DIMS;i++)  {
   dimid = SDgetdimid(sds,i);
   if (dimid == -1) continue;
   
   SDdiminfo(dimid, dimname, &value, &number_type1, &nattr1);
   for(j=0;j<strlen(dimname);j++)dimname[j] = toupper(dimname[j]);
   
   if (strstr(dimname,"BAND")) {
       if (n_bands == -1) n_bands = value;
       else if (n_bands != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"DET")){
       if (n_dets == -1) n_dets = value;
       else if (n_dets != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"AOI")){
       if (n_AOIs == -1) n_AOIs = value;
       else if (n_AOIs != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"SIDE")) {
       if (n_sides == -1) n_sides = value;
       else if (n_sides != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   }

/* get parameter info */
SDgetinfo(sds, SDSname, &rank1, dims1, &number_type1, &nattr1);
if (number_type1 != DFNT_FLOAT32) {
    sprintf(msg, " SDS %s does not contain floating point data; check file ", param->name);
    modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
    return(4);
    }
    
if (rank1 > MAX_POL_DIMS) {
    sprintf(msg, " SDS %s has more dimensions than can be handled at present; check file ", param->name);
    modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
    return(4);
    }
    
idim=1;
/* printf("Processing %s, %d", SDSname, dims1[0]);    */
for (i=0;i<rank1;i++){
   /* if (i>0) printf(" by %d", dims1[i]);    */
   start[i] = 0;
   dims[i] = dims1[i];
   param->dims[i] = dims1[i];
   idim *= dims[i];
   }
   
/* printf("\ntotal number of pixels %d\n", idim);  */

if ( (scratch = (float32 *)malloc(idim*sizeof(float32))) == NULL) {
    sprintf(msg, " memory problem; cannot allocate %d floats ", idim);
    modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
    SDendaccess(sds);
    return(3);
}    
  
ret = SDreaddata(sds, start, NULL, dims, (void *)scratch); 
if (ret == -1){
    sprintf(msg, " reading data from SDS %s ", param->name);
    modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
    SDendaccess(sds);
    return(2);
   }
SDendaccess(sds);

/* Translate pointer-based array into array-of-array-of-array..
   types of arrays, to make indexing slightly easier. Variables
   n_sides, n_AOIs, n_dets, n_bands are global. */
idim=0;
if ((n_AOIs > -1)&&(n_sides==-1)&&(n_dets==-1)&&(n_bands==-1)) {
   /* Then you are reading AOI, which is one-dimensional.  Must be the first. */
   if ( (param->one_d_data = (float *)malloc(n_AOIs*sizeof(float))) == NULL) {
     sprintf(msg, " memory problem; cannot allocate %d floats ", n_AOIs);
     modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
     return(3);
    }    
   for(iaoi=0;iaoi<n_AOIs;iaoi++) 
      param->one_d_data[iaoi] = scratch[idim++];
}
else {

   if ( (param->data = (float ****)malloc(n_bands*sizeof(float ***))) == NULL) {
       modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
       return(3);
   }
   for(iband=0;iband<n_bands;iband++) {
     if ( (param->data[iband] = (float ***)malloc(n_dets*sizeof(float **))) == NULL) {
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
        return(3);
     }    
     for(idet=0;idet<n_dets;idet++) {
       if ( (param->data[iband][idet] = (float **)malloc(n_AOIs*sizeof(float *))) == NULL) {
          modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
          return(3);
       }    
     
       for(iaoi=0;iaoi<n_AOIs;iaoi++) {
         if ( (param->data[iband][idet][iaoi] = (float *)malloc(n_sides*sizeof(float))) == NULL) {
           modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
           return(3);
         }    
       }
     }
   }
   
   for(iside=0;iside<n_sides;iside++) {
     for(iaoi=0;iaoi<n_AOIs;iaoi++) {
        for(idet=0;idet<n_dets;idet++) {
           for(iband=0;iband<n_bands;iband++) {
	     param->data[iband][idet][iaoi][iside] = scratch[idim++];
                  }
              }
          }
       }
   param->dims[0] = n_bands;
   param->dims[1] = n_dets;
   param->dims[2] = n_AOIs;
  }


free(scratch);
return(ret);
}

void init_polarization_data(pol_param_t *param, char *name)
{
/*
!C******************************************************************

!Description: This function initializes structures that hold polarization parameters.

!Input Parameters:
param:     structure into which the parameter is read (global)
name:      name of the parameter to read


!Output Parameters:
none.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    All parameters to read are globally defined in file pol_driver.h.

!END****************************************************************************
*/
pol_param_t *scratch;
int i;

scratch = (pol_param_t *)malloc(sizeof(pol_param_t));
scratch->one_d_data = NULL;
scratch->data = NULL;
for (i=0;i<MAX_POL_DIMS;i++) scratch->dims[i] = -1;
strcpy(scratch->name, name);
memcpy(param, scratch, sizeof(pol_param_t));
}

void free_polarization_data(pol_param_t *param)
{
/*
!C******************************************************************

!Description: This function frees memory in structures that hold polarization parameters.

!Input Parameters:
param:     structure with memory to be freed (global)


!Output Parameters:
none.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    All parameters to read are globally defined in file pol_driver.h.
    To be called from main() at very end of program.

!END****************************************************************************
*/
int i,j,k;

if (param->data != NULL) {
  for(i=0;i<param->dims[0];i++) {
    for(j=0;j<param->dims[1];j++) {
      for(k=0;k<param->dims[2];k++) {
           if (param->data[i][j][k] != NULL) free(param->data[i][j][k]);
      }  
      if (param->data[i][j] != NULL) free(param->data[i][j]);
    } 
    if (param->data[i] != NULL) free(param->data[i]);
  }
  free(param->data);
}
if (param->one_d_data != NULL) free(param->one_d_data);

/* dmalloc complains that this isn't being freed.  But when I try to 
   free it -- outside of dmalloc -- I get a segmentation fault, even 
   if I make sure it isn't NULL first.  One of the params must be 
   getting corrupted somewhere.
if (param != NULL) free(param);*/

}


int read_Rayleigh_data(int lun)
{
/*
!C******************************************************************

!Description: This function reads Rayleigh files.

!Input Parameters:
lun:        LUN for the file to be read (e. g., 209571, which points to "rayleigh_modis_412_iqu2.hdf").


!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    All parameters to read are globally defined in file pol_driver.h.
    To be called from main() at the very beginning of the program.

!END****************************************************************************
*/
int32 sd;
int ret;
int pol_ver;
int get_Rayleigh_data(int32 sds, pol_param_t *param, char *name);
int32 n_sets1, n_gattr1;
char location[100]="polarization.c, read_Rayleigh_data\0";
char name[256];
char msg[256];

pol_ver = 1;
if(PGS_PC_GetReference(lun, &pol_ver, name)!=PGS_S_SUCCESS){
  sprintf(msg, "LUN %d, version %d (Rayleigh data tables)", (int32)lun, pol_ver);
  modsmf((PGSt_SMF_code)MODIS_E_GETREF_FAIL, msg, location);
       }

if ( (sd = SDstart(name, DFACC_RDONLY)) == -1) {
   modsmf((PGSt_SMF_code)MODIS_E_OPEN_HDF_FILE, name, location);
   return(2);
   }
   
ret=SDfileinfo(sd, &n_sets1, &n_gattr1);
if ((ret == -1)||(n_sets1 == 0)) {
   modsmf((PGSt_SMF_code)MODIS_E_HDF_READ, "Rayleigh data file is unreadable", location);
   SDend(sd);
   return(2);
   }

ret = get_Rayleigh_data(sd, &tau, "tau\0");  /* these ***MUST!!!*** be read first */
ret += get_Rayleigh_data(sd, &dep, "dep\0");
ret += get_Rayleigh_data(sd, &ang, "ang\0");
ret += get_Rayleigh_data(sd, &sun, "sun\0");
ret += get_Rayleigh_data(sd, &sigma, "sigma\0");
ret += get_Rayleigh_data(sd, &eye, "i\0");
ret += get_Rayleigh_data(sd, &cue, "q\0");
ret += get_Rayleigh_data(sd, &you, "u\0");

if (ret > 0){
   modsmf((PGSt_SMF_code)MODIS_E_HDF_READ, "Error accessing Rayleigh data", location);
   SDend(sd);
   return(2);
   }
   
ret=SDend(sd);
if (ret == -1){
   modsmf((PGSt_SMF_code)MODIS_E_CLOSE_HDF_FILE, name, location);
   return(2);
   }
return(0);
}

int get_Rayleigh_data(int32 sd, pol_param_t *param, char *name)
{
/*
!C******************************************************************

!Description: This function reads Rayleigh tables within Rayleigh files.

!Input Parameters:
sd:        HDF file tag (return value of SDstart).
param:     structure into which the parameter is read (global)
name:      name of the parameter to read


!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    All parameters to read are globally defined in file pol_driver.h.

!END****************************************************************************
*/
int i,j;
int ilambda,irad,isun,iwind,iorder,idim;
int32 sds, dimid, ret;
int32 start[MAX_POL_DIMS], dims[MAX_POL_DIMS];
char SDSname[100], dimname[100];
int32 rank1, value, count, dims1[MAX_POL_DIMS], number_type1, nattr1;
float32 *scratch;
char msg[PGS_SMF_MAX_MSG_SIZE];
char location[100]="polarization.c, get_Rayleigh_data\0";
void init_polarization_data(pol_param_t *param, char *name);

init_polarization_data(param, name);

if ( (sds = SDselect(sd, SDnametoindex(sd, param->name))) == -1) {
   sprintf(msg, " SDselecting Rayleigh parameter '%s' ", param->name);
   modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
   return(1);
   }

/* get dimension info */
for (i=0;i<MAX_POL_DIMS;i++)  {
   dimid = SDgetdimid(sds,i);
   if (dimid == -1) continue;
   
   SDdiminfo(dimid, dimname, &value, &number_type1, &nattr1);
   for(j=0;j<strlen(dimname);j++)dimname[j] = toupper(dimname[j]);
   
   if (strstr(dimname,"LAMBDA")) {
       if (n_lambda == -1) n_lambda = value;
       else if (n_lambda != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"RAD")){
       if (n_rad_ray == -1) n_rad_ray = value;
       else if (n_rad_ray != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"SUN")){
       if (n_sun_ray == -1) n_sun_ray = value;
       else if (n_sun_ray != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"WIND")){
       if (n_wind_ray == -1) n_wind_ray = value;
       else if (n_wind_ray != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  
   if (strstr(dimname,"ORDER")){
       if (n_order_ray == -1) n_order_ray = value;
       else if (n_order_ray != value) {
          sprintf(msg, " problem with the dimension %s of SDS %s ", dimname, param->name);
          modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
          SDendaccess(sds);
          return(2);
       }
      }  

   }

/* printf("%d %d %d %d %d\n",n_lambda,n_rad_ray,n_sun_ray,n_wind_ray,n_order_ray); */


/* get parameter info */
SDgetinfo(sds, SDSname, &rank1, dims1, &number_type1, &nattr1);
if (number_type1 != DFNT_FLOAT32) {
    sprintf(msg, " SDS %s does not contain floating point data; check file ", param->name);
    modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
    return(4);
    }
    
if (rank1 > MAX_POL_DIMS) {
    sprintf(msg, " SDS %s has more dimensions than can be handled at present; check file ", param->name);
    modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
    return(4);
    }
    
idim=1;
/* printf("Processing %s, %d", SDSname, dims1[0]);   */
for (i=0;i<rank1;i++){
   /* if (i>0) printf(" by %d", dims1[i]); */  
   start[i] = 0;
   dims[i] = dims1[i];
   param->dims[i] = dims1[i];
   idim *= dims[i];
   }
   
/* printf("\ntotal number of pixels %d\n", idim);   */

if ( (scratch = (float32 *)malloc(idim*sizeof(float32))) == NULL) {
    sprintf(msg, " memory problem; cannot allocate %d floats ", idim);
    modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
    SDendaccess(sds);
    return(3);
}    
  
ret = SDreaddata(sds, start, NULL, dims, (void *)scratch); 
if (ret == -1){
    sprintf(msg, " reading data from SDS %s ", param->name);
    modsmf((PGSt_SMF_code)MODIS_E_HDF_ERROR, msg, location);
    SDendaccess(sds);
    return(2);
   }
SDendaccess(sds);

/* Translate pointer-based array into array-of-array-of-array..
   types of arrays, to make indexing slightly easier. Variables
   n_sides, n_AOIs, n_dets, n_bands are global. */
idim=0;
if ((n_lambda==-1)||(n_rad_ray==-1)||(n_sun_ray==-1)||
    (n_wind_ray==-1)||(n_order_ray==-1)) {

   /* Then you are reading one-dimensional SDSs.  Must be the first. */
   
   if (strstr(SDSname,"tau")&&(n_lambda != -1)) {
      if ( (param->one_d_data = (float *)malloc(n_lambda*sizeof(float))) == NULL) {
        sprintf(msg, " memory problem; cannot allocate %d floats ", n_lambda);
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
        return(3);
      }    
      for(ilambda=0;ilambda<n_lambda;ilambda++) 
          param->one_d_data[ilambda] = scratch[idim++];
      }
   else if (strstr(SDSname,"dep")&&(n_lambda != -1)) {
      if ( (param->one_d_data = (float *)malloc(n_lambda*sizeof(float))) == NULL) {
        sprintf(msg, " memory problem; cannot allocate %d floats ", n_lambda);
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
        return(3);
      }    
      for(ilambda=0;ilambda<n_lambda;ilambda++) 
          param->one_d_data[ilambda] = scratch[idim++];
      }
   else if (strstr(SDSname,"ang")&&(n_rad_ray != -1)) {
      if ( (param->one_d_data = (float *)malloc(n_rad_ray*sizeof(float))) == NULL) {
        sprintf(msg, " memory problem; cannot allocate %d floats ", n_rad_ray);
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
        return(3);
      }    
      for(irad=0;irad<n_rad_ray;irad++) 
          param->one_d_data[irad] = scratch[idim++];
      }
   else if (strstr(SDSname,"sun")&&(n_sun_ray != -1)) {
      if ( (param->one_d_data = (float *)malloc(n_sun_ray*sizeof(float))) == NULL) {
        sprintf(msg, " memory problem; cannot allocate %d floats ", n_sun_ray);
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
        return(3);
      }    
      for(isun=0;isun<n_sun_ray;isun++) 
          param->one_d_data[isun] = scratch[idim++];
      }
   else if (strstr(SDSname,"sigma")&&(n_wind_ray != -1)) {
      if ( (param->one_d_data = (float *)malloc(n_wind_ray*sizeof(float))) == NULL) {
        sprintf(msg, " memory problem; cannot allocate %d floats ", n_wind_ray);
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, msg, location);
        return(3);
      }    
      for(iwind=0;iwind<n_wind_ray;iwind++)
          param->one_d_data[iwind] = scratch[idim++];
      }
}
else {

   if ( (param->data = (float ****)malloc(n_rad_ray*sizeof(float ***))) == NULL) {
      modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
      return(3);
   }
   for(irad=0;irad<n_rad_ray;irad++) {
     if ( (param->data[irad] = (float ***)malloc(n_order_ray*sizeof(float **))) == NULL) {
        modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
        return(3);
     }    
     for(iorder=0;iorder<n_order_ray;iorder++) {
       if ( (param->data[irad][iorder] = (float **)malloc(n_sun_ray*sizeof(float *))) == NULL) {
          modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
          return(3);
       }    
     
       for(isun=0;isun<n_sun_ray;isun++) {
         if ( (param->data[irad][iorder][isun] = (float *)malloc(n_wind_ray*sizeof(float))) == NULL) {
            modsmf((PGSt_SMF_code)MODIS_E_CALLOC_FAIL, " ", location);
            return(3);
         }    
       }
     }
   }

   for(iwind=0;iwind<n_wind_ray;iwind++) {
     for(isun=0;isun<n_sun_ray;isun++) {
        for(iorder=0;iorder<n_order_ray;iorder++) {
           for(irad=0;irad<n_rad_ray;irad++) {
	     param->data[irad][iorder][isun][iwind] = scratch[idim++];
                  }
              }
          }
       }
   param->dims[0] = n_rad_ray;
   param->dims[1] = n_order_ray;
   param->dims[2] = n_sun_ray;
  }


free(scratch);
return(ret);
}

int polarization_coefficient( float32 SUN, float32 THETA, 
                             float32 PHI, int N_DET, float32 ZAOI, float scan_line_orientation,
		             int N_SIDE, int N_SUN, float32 *RAD)
{
/*
!C******************************************************************

!Description: This function applies values read from polarization and Rayleigh
              tables to "ERRONIOUS RADIANCES" (radiance values in E_RAD),
	      returning a coefficient for correcting the radiance.  Code is 
	      taken from PGE09, files new_modis_pol_corr_sub.f (correcting 
	      radiances) and from rayleigh_rough.f (calculating Rayleigh 
	      parameters).

!Input Parameters:
SUN:       solar zenith angle, in degrees
THETA:     view zenith angle, in degrees
PHI:       relative azimuth angle, in degrees
N_DET:     whatever detector the pixel is from (0-9)
ZAOI:      where along the scan line you are: e. g., 

      if inpixsiz = number of pixels across scan
      and     ipx = current pixel across scan (1 ... inpixsiz)

      ZAOI = 10.5 + real(ipx-1)/real(inpixsiz)*(65.5-10.5);
      
scan_line_orientation:
           equivalent to factor ALPHA from wang3.f, which is set as follows:
	   
      ALPHA = PHISAT - AZ_POL - 90.
	   
	   scan_line-orientation is in degrees.
	   
N_SUN:     which "side of the sun" you are on, which requires different 
           "new_modis_pol_corrX.hdf" files (new, 22-JAN-03).
	    N_SUN = 0, pixel is "north of sun", need data from file new_modis_pol_corr5a.hdf
	    N_SUN = 1, pixel is "south of sun", need data from file new_modis_pol_corr5b.hdf

N_SIDE:    the side of the mirror.

!Input-Output Parameters:
RAD:     bad radiances (input), corrected radiances (output)

!Output Parameters:
none; returns 0 for success, nonzero for error.


!Revision History:

Original version
J. Ray (jim@cratmos.gsfc.nasa.gov), SSAI: 04-APR-01


!Team-unique Header:

MODIS TEAM MEMBER, Christopher O. Justice justice@kratmos.gsfc.nasa.gov, 301-286-7372
		   NASA/GSFC Code 923
		   Greenbelt, MD , 20771

DEVELOPERS:  Eric F Vermote, eric@kratmos.gsfc.nasa.gov, 301-286-6232
	     Robert Wolfe,   rwolfwe@kratmos.gsfc.nasa.gov, 301-286-4585
	     Jeff Privette,  privette@kratmos.gsfc.nasa.gov, 301-286-5340
	     Nazmi El Saleous,  nazmi@kratmos.gsfc.nasa.gov, 301-286-0758


!References and Credits


!Design Notes
    To be run once for each band 8 pixel to be corrected.  Band 8 is the only 
    band that is currently corrected.  The "erronious" band 8 value must be 
    scaled and put as value RAD.  
    
    All polarization/Rayleigh parameters that are used are globally 
    defined in file pol_driver.h.
    
    After this routine is listed the original FORTRAN code from which this code was
    taken, commented out.  That code, and the C versions of that code in this routine, 
    must be accompanied by the following banner:
    
C Copyright 1988-2001 by Rosenstiel School of Marine and Atmospheric Science,
C University of Miami, Miami, Florida.
C
C                       All Rights Reserved
C
C Permission to use, copy, modify, and distribute this software and its
C documentation for non-commercial purposes and without fee is hereby granted,
C provided that the above copyright notice appear in all copies and that both
C that copyright notice and this permission notice appear in supporting
C documentation, and that the names of University of Miami and/or RSMAS not be
C used in advertising or publicity pertaining to distribution of the software
C without specific, written prior permission.
C
C UNIVERSITY OF MIAMI DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
C INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
C SHALL UNIVERSITY OF MIAMI BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
C DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
C WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
C OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
C
C !References and Credits:
c
C	Written by:
c		Howard Gordon and Menghua Wang
c 		University of Miami
c	 	Physics Department
c		Coral Gables,Fl
c		Miami, Fl
c		33124
c	Contact: HGordon@umiami.ir.miami.edu
c

!END****************************************************************************
*/
int i,j;
int iband,iside,iaoi,idet,idim;
int ilambda,irad,isun,iwind,iorder;
int M_AOI, N_AOI, problem;
float32 RAY_I_11[12];
float32 RAY_Q_11[12];
float32 RAY_U_11[12];
float32 RAY_I_9[12];
float32 RAY_Q_9[12];
float32 RAY_U_9[12];
float32 AM1[12][12];
int nine_to_11_map[] = {0,1,2,3,4,5,5,6,6,7,8};
int ISIG_MIN, ISIG_MAX;
float SIGMA, SIG_MIN, SIG_MAX;
int ISUN_LOW, ISUN_HIGH;
int J1, J2;
float HSUN, HVIEW, P, Q, tempQ, tempU;
float COSD_PHI[3], SIND_PHI[3];
float I_RAY[1][12], Q_RAY[1][12], U_RAY[1][12];
float F00, F10, F01, F11, VALUE;
float RAY_I[1], RAY_Q[1], RAY_U[1];
static int already = 0;
pol_param_t scratch12, scratch13;

/* n_sides, n_AOIs, n_dets, n_bands are global (in "pol_driver.h") *    /
if (already == 0) {
   printf("AM12_0\n\n");
 
       for(iside=0;iside<n_sides;iside++) {
          for(iaoi=0;iaoi<n_AOIs;iaoi++) {
	      printf("Angle of incidence %f\n\nband              ", aoi.one_d_data[iaoi]);
	      for(iband=0;iband<n_bands;iband++) printf("  %d       ", iband+1);
	      printf("\n");
              for(idet=0;idet<n_dets;idet++) {
	         printf("Detector   %2d  ", idet+1);
                 for(iband=0;iband<n_bands;iband++) {
		     printf(" %f ", am12_0.data[iband][idet][iaoi][iside]);
                 }
	         printf("\n");
              }
          }
       }
  printf("AM13_0\n\n");
       for(iside=0;iside<n_sides;iside++) {
          for(iaoi=0;iaoi<n_AOIs;iaoi++) {
	      printf("Angle of incidence %f\n\nband              ", aoi.one_d_data[iaoi]);
	      for(iband=0;iband<n_bands;iband++) printf("  %d       ", iband+1);
	      printf("\n");
              for(idet=0;idet<n_dets;idet++) {
	         printf("Detector   %2d  ", idet+1);
                 for(iband=0;iband<n_bands;iband++) {
		     printf(" %f ", am13_0.data[iband][idet][iaoi][iside]);
                 }
	         printf("\n");
              }
          }
       }
   if (am12_1.data != NULL) {
     printf("AM12_1\n\n");
 
       for(iside=0;iside<n_sides;iside++) {
          for(iaoi=0;iaoi<n_AOIs;iaoi++) {
	      printf("Angle of incidence %f\n\nband              ", aoi.one_d_data[iaoi]);
	      for(iband=0;iband<n_bands;iband++) printf("  %d       ", iband+1);
	      printf("\n");
              for(idet=0;idet<n_dets;idet++) {
	         printf("Detector   %2d  ", idet+1);
                 for(iband=0;iband<n_bands;iband++) {
		     printf(" %f ", am12_1.data[iband][idet][iaoi][iside]);
                 }
	         printf("\n");
              }
          }
       }
       }
   if (am13_1.data != NULL) {
     printf("AM13_1\n\n");
       for(iside=0;iside<n_sides;iside++) {
          for(iaoi=0;iaoi<n_AOIs;iaoi++) {
	      printf("Angle of incidence %f\n\nband              ", aoi.one_d_data[iaoi]);
	      for(iband=0;iband<n_bands;iband++) printf("  %d       ", iband+1);
	      printf("\n");
              for(idet=0;idet<n_dets;idet++) {
	         printf("Detector   %2d  ", idet+1);
                 for(iband=0;iband<n_bands;iband++) {
		     printf(" %f ", am13_1.data[iband][idet][iaoi][iside]);
                 }
	         printf("\n");
              }
          }
       }
       }
}
  already++;
/   *
printf("I\n");
       for(iwind=0;iwind<n_wind_ray;iwind++) {
          printf("**************\nSigma = %f\n**************\n", sigma.one_d_data[iwind]);
          for(isun=0;isun<n_sun_ray;isun++) {
	      printf("\nSun %f\nangle       ", sun.one_d_data[isun]);
	      for(irad=0;irad<n_rad_ray;irad++) printf(" %f ", ang.one_d_data[irad]);
	      printf("\n");
              for(iorder=0;iorder<n_order_ray;iorder++) {
	         printf("Iorder   %2d  ", iorder);
                 for(irad=0;irad<n_rad_ray;irad++) {
		     printf(" %f ", eye.data[irad][iorder][isun][iwind]);
                 }
	         printf("\n");
              }
          }
       }
/ * This is a C translation of the original FORTRAN material.
   Most comments are from the original. */
/*(from rayleigh_rough.f)
C	FIGURE OUT THE VALUE OF SIGMA FROM THE WIND SPEED.
*/
/*
 *  These steps are not really necessary if WIND is held constant.  
 *  With WIND = 7.0F, SIGMA=0.19334F, ISIG_MIN=1, ISIG_MAX=2, 
 *  SIG_MIN=0.1F, and SIG_MAX=0.2F.
 *
 	SIGMA    = sqrtf(0.00534F*WIND);   / * COS AND MUNK 1950 (OMNIDIRECTIONAL) * /

	ISIG_MIN = (int)(10.0F*SIGMA) + 0;
	ISIG_MAX = (int)(10.0F*SIGMA) + 1; 
	SIG_MIN  = (float)(ISIG_MIN)/10.0F;
	SIG_MAX  = (float)(ISIG_MAX)/10.0F;
	if (ISIG_MIN > n_wind_ray) ISIG_MIN = n_wind_ray;
	if (ISIG_MAX > n_wind_ray) ISIG_MAX = n_wind_ray;
 */
 
  SIGMA=0.19334F; 
  ISIG_MIN=1;
  ISIG_MAX=2;
  SIG_MIN=0.1000F;
  SIG_MAX=0.2000F;
  
/*printf("ISIG_MIN = %d, 	ISIG_MAX = %d\n", ISIG_MIN, ISIG_MAX );*/
/*printf("SIG_MIN = %f, 	SIG_MAX = %f\n", SIG_MIN, SIG_MAX );*/

/*(from rayleigh_rough.f)
C	FIGURE OUT THE INDEX TO LOCATE PROPER PAIR OF SUN ANGLES
*/
        /* 18-JUL-01: sanity check for SUN. */
	if (SUN < 0.0){
	   return(100);
	   }

	ISUN_LOW  = (int)( SUN )/2;
	ISUN_HIGH = ISUN_LOW + 1;
	if ((ISUN_LOW > n_sun_ray)||(ISUN_HIGH > n_sun_ray)) {
	    return(99);
	    }

/*(from rayleigh_rough.f)
C     FIND THE VIEWING ANGLES FOR INTERPOLATION
*/

        /* 18-JUL-01: sanity check for PHI. */
	if ((PHI < -360.0) || (PHI > 360.0)){
	   return(100);
	   }

        /* 18-JUL-01: sanity check for THETA: we had been trying to run
           files with bad SensorZenith angles and core-dumping. */
	if (THETA < 0.0) {
	   return(100);
	   }

	if (THETA > ang.one_d_data[n_rad_ray-1]) {
	   return(100);
	   }

        J1 = 0;
	J2 = 1;
	while (!((THETA >= ang.one_d_data[J1])&&(THETA <= ang.one_d_data[J2]))) {
	   J1++;
	   J2++;
	}
/*printf("J1 = %d,J2 = %d\n%f %f %f \n", J1,J2,  ang.one_d_data[J1],THETA, ang.one_d_data[J2]);*/

/*(from rayleigh_rough.f)
C     SET UP THE VARIABLES REQUIRED FOR INTERPOLATION
*/
	HSUN  = sun.one_d_data[ISUN_HIGH] - sun.one_d_data[ISUN_LOW];
	HVIEW = ang.one_d_data[J2] - ang.one_d_data[J1];
	P     = (SUN - sun.one_d_data[ISUN_LOW])/HSUN;
	Q     = (THETA - ang.one_d_data[J1])/HVIEW;

/*printf("hsun, hview, p, q: %f %f %f %f\n", HSUN, HVIEW, P, Q);*/

/*	DO M = 1, 3*/
        for(iorder=0;iorder<n_order_ray;iorder++) {
	    COSD_PHI[iorder] = cosf(DEG2RAD*( PHI * (float)iorder ) );
	    SIND_PHI[iorder] = sinf(DEG2RAD*( PHI * (float)iorder ) );
	}

	for (iwind = ISIG_MIN; iwind <= ISIG_MAX; iwind ++ ) {
	  for (iband = 0; iband < 1; iband++) { /* for now just do one band */

	    I_RAY[iband][iwind] = 0.0F;
	    Q_RAY[iband][iwind] = 0.0F;
	    U_RAY[iband][iwind] = 0.0F;
/*
C	FIRST I
*/
	    for(iorder=0;iorder<n_order_ray;iorder++) {

              F00 = eye.data[J1][iorder][ISUN_LOW][iwind]; 
              F10 = eye.data[J1][iorder][ISUN_HIGH][iwind]; 
              F01 = eye.data[J2][iorder][ISUN_LOW][iwind]; 
              F11 = eye.data[J2][iorder][ISUN_HIGH][iwind]; 

/*
         ! SEE ABROMOWITZ AND STEGUN
         ! HANDBOOK OF MATHEMATICAL
         ! FUNCTIONS. PAGE 882        
         ! FORMULA 25.2.67
*/
             VALUE = (1.0F-P)*(1.0F-Q)*F00 + P*Q*F11 + 
	             P*(1.0F-Q)*F10 + Q*(1.0F-P)*F01;

             I_RAY[iband][iwind] = VALUE * COSD_PHI[iorder] + I_RAY[iband][iwind];
	   }

/*
C	THEN Q 
*/
	   for(iorder=0;iorder<n_order_ray;iorder++) {

              F00 = cue.data[J1][iorder][ISUN_LOW][iwind]; 
              F10 = cue.data[J1][iorder][ISUN_HIGH][iwind]; 
              F01 = cue.data[J2][iorder][ISUN_LOW][iwind]; 
              F11 = cue.data[J2][iorder][ISUN_HIGH][iwind]; 

              VALUE = (1.0F-P)*(1.0F-Q)*F00 + P*Q*F11 + 
	              P*(1.0F-Q)*F10 + Q*(1.0F-P)*F01;

              Q_RAY[iband][iwind] = VALUE * COSD_PHI[iorder] + Q_RAY[iband][iwind];
	   }

/*
C	FINALLY U
*/

	   for(iorder=0;iorder<n_order_ray;iorder++) {

              F00 = you.data[J1][iorder][ISUN_LOW][iwind]; 
              F10 = you.data[J1][iorder][ISUN_HIGH][iwind]; 
              F01 = you.data[J2][iorder][ISUN_LOW][iwind]; 
              F11 = you.data[J2][iorder][ISUN_HIGH][iwind]; 

              VALUE = (1.0F-P)*(1.0F-Q)*F00 + P*Q*F11 + 
	              P*(1.0F-Q)*F10 + Q*(1.0F-P)*F01;

              U_RAY[iband][iwind] = VALUE * SIND_PHI[iorder] + U_RAY[iband][iwind];
	   }

	}  /* for (iband = 0; iband < 1; iband++) */
	}

/*(from rayleigh_rough.f)
C	DO LINEAR INTERPOLATION (IN SIGMA SPACE) TO GET THE RAYLEIGH 
C	STOKES VECTOR
*/
	if (ISIG_MIN == ISIG_MAX) {
	   for (iband = 0; iband < 1; iband++) {
		RAY_I_9[iband] = I_RAY[iband][ISIG_MIN];
	 	RAY_Q_9[iband] = Q_RAY[iband][ISIG_MIN];
		RAY_U_9[iband] = U_RAY[iband][ISIG_MIN];
	   }
	}
	else {
	   for (iband = 0; iband < 1; iband++) {
	      RAY_I_9[iband] = I_RAY[iband][ISIG_MIN]
             + ( (SIGMA - SIG_MIN)/(SIG_MAX - SIG_MIN) ) 
             * ( I_RAY[iband][ISIG_MAX] - I_RAY[iband][ISIG_MIN] );


	      RAY_Q_9[iband] = Q_RAY[iband][ISIG_MIN] 
             + ( (SIGMA - SIG_MIN)/(SIG_MAX - SIG_MIN) ) 
             * ( Q_RAY[iband][ISIG_MAX] - Q_RAY[iband][ISIG_MIN] );


	      RAY_U_9[iband] = U_RAY[iband][ISIG_MIN]
             + ( (SIGMA - SIG_MIN)/(SIG_MAX - SIG_MIN) ) 
             * ( U_RAY[iband][ISIG_MAX] - U_RAY[iband][ISIG_MIN] );
	   }
	}

/* 06-AUG-01:
 * Right here: do the adjustment for scan line orientation 
 ************************************************************/
        scan_line_orientation *= 2.0F;            /* double and */
	scan_line_orientation *= DEG2RAD;         /* convert to radians */
        for (iband = 0; iband < 1; iband++) {
	   tempQ = RAY_Q_9[iband];
	   tempU = RAY_U_9[iband];
	   
	   RAY_Q_9[iband] = cosf(scan_line_orientation)*tempQ -
	                    sinf(scan_line_orientation)*tempU;

	   RAY_U_9[iband] = sinf(scan_line_orientation)*tempQ +
	                    cosf(scan_line_orientation)*tempU;

	}

    
/*(from new_modis_pol_corr_sub.f)
   
C	Find bounding AOI entries
C
C	Extrapolate off end of table (scan angles > 45)
C
C	Note: HRG prefers second 15.5 dataset over first 15.5 dataset

*/
if (ZAOI <= aoi.one_d_data[1]) {
   M_AOI = 2;
   N_AOI = 1;
   P = (ZAOI-aoi.one_d_data[N_AOI])/(aoi.one_d_data[M_AOI]-aoi.one_d_data[N_AOI]);
   }
else if (ZAOI >= aoi.one_d_data[n_AOIs-1]) {
   M_AOI = n_AOIs-1;
   N_AOI = n_AOIs-2;
   P = (ZAOI-aoi.one_d_data[N_AOI])/(aoi.one_d_data[M_AOI]-aoi.one_d_data[N_AOI]);
    }
else {
   problem = 1;
   for(iaoi=2;iaoi<n_AOIs;iaoi++) {
	if (ZAOI < aoi.one_d_data[iaoi]) {
	    M_AOI = iaoi;
	    N_AOI = iaoi-1;
	    P = (ZAOI-aoi.one_d_data[N_AOI])/(aoi.one_d_data[M_AOI]-aoi.one_d_data[N_AOI]);
	    problem = 0;
	    break;
            }
        }
   if (problem) return(2);
     
      }
/*printf("P = %f\n", P);*/

/*(from new_modis_pol_corr_sub.f)
   
C	GET THE RAYLEIGH SCATTERING STOKES VECTOR
C	NOTE, RAY_I_9, ETC IS FOR 9 SPECTRAL BANDS

C	NOTE, THE TEST DATA ARE FOR THE 9 MODIS SPECTRAL BANDS;  HOWEVER,
C	THERE ARE REALLY 11 PHYSICAL BANDS BECAUSE OF THE TWO TDI BANDS
C	AT 667 AND 678 NM.  FOR THE PURPOSE OF THE TEST, WE READ IN THE
C	DATA FOR SPECTRAL BANDS 1-5 INTO PHYSICAL BANDS  1 - 5,
C	         SPECTRAL BANDS   6 INTO PHYSICAL BANDS  7 & 8,
C	         SPECTRAL BANDS   7 INTO PHYSICAL BANDS  8 & 9, AND
C	         SPECTRAL BANDS 8-9 INTO PHYSICAL BANDS 10 & 11

c	CALL
c     &   RAYLEIGH_IQU ( THETA0, THETAV, PHIV,
c     &                  RAY_I_9, RAY_Q_9, RAY_U_9 )

C	NOW SWITCH FROM WAVELENGTH # TO BAND # (BECAUSE OF THE TDI BANDS)
*/

for (i=10;i>=0;i--) {
    RAY_U_11[i] = RAY_U_9[nine_to_11_map[i]];
    RAY_Q_11[i] = RAY_Q_9[nine_to_11_map[i]];
    RAY_I_11[i] = RAY_I_9[nine_to_11_map[i]];
}

/*(from new_modis_pol_corr_sub.f)

C	COMPUTE THE ERRONIOUS RADIANCE CORRECTION FACTOR
*/

/* 23-JAN-03: take data from am12_0 or am12_1, am13_0 or am13_1, based upon
   a) value of N_SUN (0 -> 0, 1 -> 1) and b) availability of am12_1/am13_1:
   they will not be available if file at LUN 209569 (new_modis_pol_corr5b.hdf)
   has not been staged.  */
   
scratch12 = am12_0;
scratch13 = am13_0;
if ((N_SUN == 1)&&(am12_1.data != NULL)&&(am13_1.data != NULL)) {
   scratch12 = am12_1;
   scratch13 = am13_1;
   }
   
for (iaoi=N_AOI;iaoi<=M_AOI;iaoi++) {
   for (iband=0;iband<1;iband++) { /* for now just do one band */
       AM1[iband][iaoi] =
            (  1.0F    +
     	   scratch12.data[iband][N_DET][iaoi][N_SIDE] * (RAY_Q_11[iband] /
        	          RAY_I_11[iband] ) +
     	   scratch13.data[iband][N_DET][iaoi][N_SIDE] * (RAY_U_11[iband] /
     	                  RAY_I_11[iband] ) );
   }
}

/*(from new_modis_pol_corr_sub.f)

C	CORRECT THE ERRONIOUS RADIANCES (E=ERRONIOUS, C=CORRECTED)
*/
for (iband=0;iband<1;iband++) { /* for now just do one band */
    *RAD = (*RAD)/( AM1[iband][N_AOI]*(1.0F - P) + AM1[iband][M_AOI]*(P) );
}			      
			      

/* end, FORTRAN-to-C material */

return(0);
}

/*************************************************************************
No code beyond this point: this is only a dump of the relevant FORTRAN 
material from file new_modis_pol_corr_sub.f:
(/modisbaselinedcode/OCEANS/STORE/PGE09/MOD_PR18/new_modis_pol_corr_sub.f)

C	Find bounding AOI entries
C
C	Extrapolate off end of table (scan angles > 45)
C
C	Note: HRG prefers second 15.5 dataset over first 15.5 dataset

	IF (ZAOI .LE. AOI(2)) then
	    M_AOI = 3			! off start of table
	    N_AOI = 2
	    P = (ZAOI-AOI(N_AOI))/(AOI(M_AOI)-AOI(N_AOI)) ! P < 0.0
	else if (ZAOI .GE. AOI(NAOI)) then
	    M_AOI = NAOI		! off end of table
	    N_AOI = NAOI-1
	    P = (ZAOI-AOI(N_AOI))/(AOI(M_AOI)-AOI(N_AOI)) ! P > 1.0
	else
	    do I_AOI = 3, NAOI
		if (ZAOI .LT. AOI(I_AOI)) then
		    M_AOI = I_AOI
		    N_AOI = I_AOI-1
		    P = (ZAOI-AOI(N_AOI))/(AOI(M_AOI)-AOI(N_AOI))
		    go to 100				! 0.0 <= P <= 1.0
		end if
	    end do
	    returnstatus = -2		! problem with AOI search
	    return
100	    continue
	end if


C	GET THE RAYLEIGH SCATTERING STOKES VECTOR
C	NOTE, RAY_I_9, ETC IS FOR 9 SPECTRAL BANDS

C	NOTE, THE TEST DATA ARE FOR THE 9 MODIS SPECTRAL BANDS;  HOWEVER,
C	THERE ARE REALLY 11 PHYSICAL BANDS BECAUSE OF THE TWO TDI BANDS
C	AT 667 AND 678 NM.  FOR THE PURPOSE OF THE TEST, WE READ IN THE
C	DATA FOR SPECTRAL BANDS 1-5 INTO PHYSICAL BANDS  1 - 5,
C	         SPECTRAL BANDS   6 INTO PHYSICAL BANDS  7 & 8,
C	         SPECTRAL BANDS   7 INTO PHYSICAL BANDS  8 & 9, AND
C	         SPECTRAL BANDS 8-9 INTO PHYSICAL BANDS 10 & 11

c	CALL
c     &   RAYLEIGH_IQU ( THETA0, THETAV, PHIV,
c     &                  RAY_I_9, RAY_Q_9, RAY_U_9 )

C	NOW SWITCH FROM WAVELENGTH # TO BAND # (BECAUSE OF THE TDI BANDS)

	RAY_I_11(11) = RAY_I_9(9)
	RAY_I_11(10) = RAY_I_9(8)
	RAY_I_11(9)  = RAY_I_9(7)
	RAY_I_11(8)  = RAY_I_9(7)
	RAY_I_11(7)  = RAY_I_9(6)
	RAY_I_11(6)  = RAY_I_9(6)
	RAY_I_11(5)  = RAY_I_9(5)
	RAY_I_11(4)  = RAY_I_9(4)
	RAY_I_11(3)  = RAY_I_9(3)
	RAY_I_11(2)  = RAY_I_9(2)
	RAY_I_11(1)  = RAY_I_9(1)

	RAY_Q_11(11) = RAY_Q_9(9)
	RAY_Q_11(10) = RAY_Q_9(8)
	RAY_Q_11(9)  = RAY_Q_9(7)
	RAY_Q_11(8)  = RAY_Q_9(7)
	RAY_Q_11(7)  = RAY_Q_9(6)
	RAY_Q_11(6)  = RAY_Q_9(6)
	RAY_Q_11(5)  = RAY_Q_9(5)
	RAY_Q_11(4)  = RAY_Q_9(4)
	RAY_Q_11(3)  = RAY_Q_9(3)
	RAY_Q_11(2)  = RAY_Q_9(2)
	RAY_Q_11(1)  = RAY_Q_9(1)

	RAY_U_11(11) = RAY_U_9(9)
	RAY_U_11(10) = RAY_U_9(8)
	RAY_U_11(9)  = RAY_U_9(7)
	RAY_U_11(8)  = RAY_U_9(7)
	RAY_U_11(7)  = RAY_U_9(6)
	RAY_U_11(6)  = RAY_U_9(6)
	RAY_U_11(5)  = RAY_U_9(5)
	RAY_U_11(4)  = RAY_U_9(4)
	RAY_U_11(3)  = RAY_U_9(3)
	RAY_U_11(2)  = RAY_U_9(2)
	RAY_U_11(1)  = RAY_U_9(1)


C	COMPUTE THE ERRONIOUS RADIANCE CORRECTION FACTOR

	DO I_AOI = N_AOI, M_AOI
	    DO I_BAND = 1, NBAND
		AM1(I_BAND,I_AOI) =
     &       (  1.    +
     &	   AM12(I_BAND,N_DET,I_AOI,N_SIDE) * (RAY_Q_11(I_BAND) /
     &   	          RAY_I_11(I_BAND) ) +
     &	   AM13(I_BAND,N_DET,I_AOI,N_SIDE) * (RAY_U_11(I_BAND) /
     &	                  RAY_I_11(I_BAND) ) )
	    END DO
	END DO


C	CORRECT THE ERRONIOUS RADIANCES (E=ERRONIOUS, C=CORRECTED)

	DO I_BAND = 1, NBAND
	   C_RAD(I_BAND) = E_RAD(I_BAND)
     &     / ( AM1(I_BAND,N_AOI)*(1.-P) +
     &	       AM1(I_BAND,M_AOI)*(P) )
  	ENDDO

	if (returnstatus .eq. PGS_S_SUCCESS) then
	 returnstatus = 0
	end if


*************************************************************************/

/*************************************************************************
And this is a dump of the relevant FORTRAN material from file 
rayleigh_rough.f:
(/modisbaselinedcode/OCEANS/STORE/PGE09/MOD_PR18/rayleigh_rough.f)

C	FIGURE OUT THE VALUE OF SIGMA FROM THE WIND SPEED.

	SIGMA    = SQRT(0.00534*WIND)   ! COS AND MUNK 1950 (OMNIDIRECTIONAL)

	ISIG_MIN = IFIX(10*SIGMA) + 1  ! SIGMA = 0.1 GIVES ISIG_MIN = 2 
	ISIG_MAX = IFIX(10*SIGMA) + 2  ! SIGMA = 0.1 GIVES ISIG_MAX = 3 
	SIG_MIN  = FLOAT(ISIG_MIN-1)/10.
	SIG_MAX  = FLOAT(ISIG_MAX-1)/10.
	
c	IF (SIGMA .GT. 0.3999) THEN 
c		ISIG_MIN = NWIND
c		ISIG_MAX = NWIND
c	ENDIF
	IF (ISIG_MIN .GT. NWIND) THEN
		ISIG_MIN = NWIND
	ENDIF
	IF (ISIG_MAX .GT. NWIND) THEN
		ISIG_MAX = NWIND
	ENDIF

C	FIGURE OUT THE INDEX TO LOCATE PROPER PAIR OF SUN ANGLES

	ISUN_LOW  = IFIX( SUN )/2 + 1
	ISUN_HIGH = ISUN_LOW + 1
	if (ISUN_LOW.gt.MSUN .or. ISUN_HIGH.gt.MSUN) then
	  IERROR = 203				! rayleigh: theta0 off table
	  return
	end if

C     FIND THE VIEWING ANGLES FOR INTERPOLATION

	if (THETA .gt. RAY_ANG(MRAD)) then
	  IERROR = 204				! rayleigh: theta off table
	  return
	end if

	if (J1 .gt. 0 .and. J2 .gt. 0) then
	  if (THETA .ge. RAY_ANG(J1) .and. THETA .le. RAY_ANG(J2))
     &	    go to 22				! still using same entry
	end if

      	DO J = 1, MRAD
      	J2 = J
    	IF ( THETA .LT. RAY_ANG(J) ) GO TO 21
	END DO
21    	J1 = J2 - 1
22	CONTINUE

C     THETA IS BETWEEN RAY_ANG(J1) AND RAY_ANG(J2)

C     SET UP THE VARIABLES REQUIRED FOR INTERPOLATION

	HSUN  = RAY_SUN(ISUN_HIGH) -RAY_SUN(ISUN_LOW)
	HVIEW = RAY_ANG(J2) - RAY_ANG(J1)
	P     = ( SUN - RAY_SUN(ISUN_LOW) )/HSUN
	Q     = ( THETA - RAY_ANG(J1) )/HVIEW

c	print *, 'hsun, hview, p, q', hsun, hview, p, q 

	DO M = 1, 3
	    COSD_PHI(M) = COS(RAD( PHI * FLOAT(M-1) ) )
	    SIND_PHI(M) = SIN(RAD( PHI * FLOAT(M-1) ) )
	ENDDO

	DO IWIND = ISIG_MIN, ISIG_MAX
	DO IBAND = 1, NLAMBDA

	I_RAY( IBAND, IWIND ) = 0.
	Q_RAY( IBAND, IWIND ) = 0.
	U_RAY( IBAND, IWIND ) = 0.

C	FIRST I

	DO M = 1, 3

         F00 = RAY_FOR_I( J1, M, ISUN_LOW,  IWIND, IBAND) 
         F10 = RAY_FOR_I( J1, M, ISUN_HIGH, IWIND, IBAND) 
         F01 = RAY_FOR_I( J2, M, ISUN_LOW,  IWIND, IBAND) 
         F11 = RAY_FOR_I( J2, M, ISUN_HIGH, IWIND, IBAND) 

         ! SEE ABROMOWITZ AND STEGUN
         ! HANDBOOK OF MATHEMATICAL
         ! FUNCTIONS. PAGE 882        
         ! FORMULA 25.2.67

            VALUE = (1.-P)*(1.-Q)*F00 + P*Q*F11
     &            + P*(1.-Q)*F10 + Q*(1.-P)*F01

            I_RAY( IBAND, IWIND ) = VALUE * COSD_PHI(M)
     &                     + I_RAY( IBAND, IWIND  )
	ENDDO

C	THEN Q 

	DO M = 1, 3

         F00 = RAY_FOR_Q( J1, M, ISUN_LOW,  IWIND, IBAND) 
         F10 = RAY_FOR_Q( J1, M, ISUN_HIGH, IWIND, IBAND) 
         F01 = RAY_FOR_Q( J2, M, ISUN_LOW,  IWIND, IBAND) 
         F11 = RAY_FOR_Q( J2, M, ISUN_HIGH, IWIND, IBAND) 

            VALUE = (1.-P)*(1.-Q)*F00 + P*Q*F11
     &            + P*(1.-Q)*F10 + Q*(1.-P)*F01

            Q_RAY( IBAND, IWIND ) = VALUE * COSD_PHI(M)
     &                     + Q_RAY( IBAND, IWIND  )
	ENDDO

C	FINALLY U

	DO M = 1, 3

         F00 = RAY_FOR_U( J1, M, ISUN_LOW,  IWIND, IBAND) 
         F10 = RAY_FOR_U( J1, M, ISUN_HIGH, IWIND, IBAND) 
         F01 = RAY_FOR_U( J2, M, ISUN_LOW,  IWIND, IBAND) 
         F11 = RAY_FOR_U( J2, M, ISUN_HIGH, IWIND, IBAND) 

            VALUE = (1.-P)*(1.-Q)*F00 + P*Q*F11
     &            + P*(1.-Q)*F10 + Q*(1.-P)*F01

            U_RAY( IBAND, IWIND ) = VALUE * SIND_PHI(M)
     &                     + U_RAY( IBAND, IWIND  )
	ENDDO

	ENDDO 	! END DO IBAND 
	ENDDO 	! END DO IWIND

C	DO LINEAR INTERPOLATION (IN SIGMA SPACE) TO GET THE RAYLEIGH 
C	STOKES VECTOR

	IF (ISIG_MIN .EQ. ISIG_MAX) THEN  ! SIGMA > 0.4
	   DO IBAND = 1, NLAMBDA
		RAY_I( IBAND ) = I_RAY( IBAND, ISIG_MIN) 
	 	RAY_Q( IBAND ) = Q_RAY( IBAND, ISIG_MIN) 
		RAY_U( IBAND ) = U_RAY( IBAND, ISIG_MIN)
	   ENDDO   ! END DO IBAND 
	ELSE 
	   DO IBAND = 1, NLAMBDA
	      RAY_I( IBAND ) = I_RAY( IBAND, ISIG_MIN) 
     &        + ( (SIGMA - SIG_MIN)/(SIG_MAX - SIG_MIN) ) 
     &        * ( I_RAY( IBAND, ISIG_MAX) - I_RAY( IBAND, ISIG_MIN) )
	      RAY_Q( IBAND ) = Q_RAY( IBAND, ISIG_MIN) 
     &        + ( (SIGMA - SIG_MIN)/(SIG_MAX - SIG_MIN) ) 
     &        * ( Q_RAY( IBAND, ISIG_MAX) - Q_RAY( IBAND, ISIG_MIN) )
	      RAY_U( IBAND ) = U_RAY( IBAND, ISIG_MIN) 
     &        + ( (SIGMA - SIG_MIN)/(SIG_MAX - SIG_MIN) ) 
     &        * ( U_RAY( IBAND, ISIG_MAX) - U_RAY( IBAND, ISIG_MIN) )
	   ENDDO   ! END DO IBAND 
	ENDIF


	RETURN

	END

*************************************************************************/
