/****************************************************************************
!C

!File: smf.c

!Description:   Routines for taking care of error/message reporting for MOD_PR09.exe

!Input Parameters: none
    
!Output Parameters: none

!Input/Output Parameters: none

!Returns: nothing

!Revision History:

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

    MODIS Science Team Members:
      Chris Justice
      MODIS Land Science Team           University of Maryland
      justice@kratmos.gsfc.nasa.gov     NASA/GSFC Code 923
      phone: 301-286-7372               Greenbelt, MD 20771

      Eric F Vermote
      Univerisity of MD.
      NASA's GSFC Code 923
      Greenbelt, MD , 20771
      eric@kratmos.gsfc.nasa.gov
      301-286-6232

      Jim Ray
      Science Systems and Applications Inc.
      NASA's Goddard Space Flight Center Code 923
      Greenbelt MD, 20771
      jim@kratmos.gsfc.nasa.gov, (301) 614-6684

!Design Notes

!END*************************************************************************/

/* Use #ifndef and #define to avoid float64 inclusion of the header file.
   If the file has been previously included, than the preprocessor will 
   skip everything up to the #endif appearing at the end of this file */
#ifndef smf_c 
#define smf_c 

/* SMF module */
#include "smf.h"
#include "PGS_MODIS_37092.h"



void gran_error_smf(PGSt_SMF_code mnemonicstring, char *infostringbuf, char *functionstring,
		    PGSt_PC_Logical file_lun, int32 version, int32 orbitnumber, int32 granulenumber)
/*
!C*****************************************************************************
!Description:  Addes the file LUN, version, orbitnumber, and granule number to 
               the SMF message string, then calls modsmf(). [Historically
	       written as an additional wrapper to modsmf()].

!Input Parameters:
mnemonicstring -- string containing the mnemonic to be called
infostring -- message string that receives the additional information 
functionstring -- string containing the function where the error occurred
file_lun -- file lun of the file that caused the error 
version -- version number of the file that caused the error
orbitnumber -- orbitnumber of the file that caused the error
granulenumber --orbitnumber of the file that caused the error.

!Output Parameters: none

!Input/Output Parameters: none

!Returns: none

!Revision History:
        Revision 1.0, P. Fisher
	1.  Slightly changed output message format to use a ',' vs a '.' as the message separator.

	Original version 1997, P. Fisher (fisher@ltpmail.gsfc.nasa.gov).

!Team-unique Header:

   This software developed by the MODIS Land Science Team for 
   the National Aeronautics and Space Administration, Goddard 
   Space Flight Center, under contract NAS5-96062

!References and Credits

		Paul Fisher
		Science Systems and Applications Inc.
		NASA's Goddard Space Flight Center Code 923
		Greenbelt MD, 20771
 		fisher@ltpmail.gsfc.nasa.gov, (301) 286-2980


!Design Notes
!END***************************************************************************
*/
{
  char infostring[PGS_SMF_MAX_MSG_SIZE];
  char seqmsg[PGS_SMF_MAX_MSG_SIZE];
  int i, j=0, k=0;


  if(strlen(infostringbuf)>PGS_SMF_MAX_MSG_SIZE-1){
    for(i=strlen(infostring);i<PGS_SMF_MAX_MSG_SIZE-1;i++)infostring[i]=infostringbuf[k++];
    infostring[PGS_SMF_MAX_MSG_SIZE-1]='\0';
  }
  else strcpy(infostring, infostringbuf);

  if((int32)file_lun==-1 && version==-1)
    sprintf(seqmsg, ", orbitnumber %d, granule %d", orbitnumber, granulenumber);
  else
    sprintf(seqmsg, "  LUN %ld, version %d, orbitnumber %d, granule %d", 
	    (long int) file_lun, version, orbitnumber, granulenumber);

  if(strlen(infostring)+strlen(seqmsg)>PGS_SMF_MAX_MSG_SIZE-1){
    modsmf(MODIS_E_SMF_ERROR, "", "smf.c, gran_error_smf\0");
    for(i=strlen(infostring);i<PGS_SMF_MAX_MSG_SIZE-1;i++)infostring[i]=seqmsg[j++];
    infostring[PGS_SMF_MAX_MSG_SIZE-1]='\0';
  }
  else strcat(infostring, seqmsg);
  modsmf(mnemonicstring, infostring, functionstring);

  return;

}



void modsmf(PGSt_SMF_code mnemonicstring, char *infostring, char *functionstring)
/*
!C****************************************************************************

!Description:

  This module writes errors and messages through the SDP toolkit to
  LogStatus and causes program termination upon encountering a fatal
  error (as determined by the error mnemonic).  As input it acccepts an
  error mnemonic, string containing user information, and a string
  containing the function name and module where the error/message
  occurs.

!Input Parameters:
  PGSt_SMF_code mnemonicstring      SMF message mnemonic
  char *infostring                  String containing info to be written to 
                                    LogStatus
  char *functionstring              String containing function/module name
                                    to be written to LogStatus

!Output Parameters:
  None

!Revision History:
   Revsion 2.0 1997/08/01
   P Fisher/SSAI, NASA's GSFC Code 923
   Added string length checking.

   Revision 1.5 1995/07/31
   P. Fisher/SDST
   Removed AT_SCF functionality to meet ECS requirements.

   Revision 1.4 1996/05/10
   P. Fisher/SDST
   Changed function to always return 1 on failure-exit, per
   MODIS standards and guidelines.

   Revision 1.3 1995/11/30
   P. Fisher/SDST
   Bug fix from previous version

   Revision 1.2 1995/11/28
   R. Wolfe/MODIS Science support
   Added AT_SCF functionality to notify user if toolkit fails

   Revision 1.1  1995/07/12  19:27:46  fisher
   Initial revision
 

!Team-unique Header:
  This software is modified by the MODIS Science Data Support
  Team for the National Aeronautics and Space Administration,
  Goddard Space Flight Center, under contract NAS5-32373.

!References and Credits:
  Written by Paul S. Fisher

    Research and Data Systems Corporation
    SAIC/GSC MODIS Support Office
    Seabrook MD 20706  

  Modifications by :
      Robert E. Wolfe
      MODIS Land Team Support Group     
      Hughes STX
      Greenbelt, MD 20770  

      Paul Fisher
      Science Cyctem and Applications Inc/NASA's GSFC Code 923
      Greenbelt, MD 20771

!Design Notes:

  The determination of time is through native C function(s).

  Upon determination of a fatal error, exit code 1 is used.
  Exit code 1 used to indicate SDP toolkit error.  These exit functions
  may be updated at a later date if required by the processing string.

  If AT_SCF is defined, output to screen upon smf message failure.

  Designed to run with seed file format of:
    MODIS_X_MNEMONIC STRING %sStatic Message.  %s

  Example of function call:

  modsmf(MODIS_X_MNEMONIC_STRING, "user message string", "function, 
                                                     module.c string");

  -- The mnemonic string must be identical to one contained in an included 
       seed file.
  -- The user message string can contain anything that needs to be written
       to the log(s).
  -- The function, module.c string should contain "the name of the function,
       the name of the module".

					      
Externals:
  constants:
       PGS_SMF_MAX_MSGBUF_SIZE                 (PGS_SMF.h)
       PGS_TRUE                                (PGS_SMF.h)
       PGS_FALSE                               (PGS_SMF.h)
       PGS_E_UNIX                              (PGS_SMF.h)
       PGS_E_UNDEFINED_CODE                    (PGS_SMF.h)
       PGS_SMF_E_LOGFILE                       (PGS_SMF.h)
       PGS_S_SUCCESS                           (PGS_SMF.h)
       NULL                                    <stdio.h>

  typedefs:
       PGSt_SMF_status                         (PGS_SMF.h)
       struct tm                               (smfio.h)<time.h>
       time_t                                  (smfio.h)<time.h>

  functions:
       PGS_SMF_TestFatalLevel                  (PGS_SMF_h)
       PGS_SMF_GetMsgByCode                    (PGS_SMF.h)
       PGS_SMF_SetDynamicMsg                   (PGS_SMF.h)
       exit                                    <stlib.h>
       sprintf                                 <stdio.h>
       time                                    (smfio.h)<time.h>
       localtime                               (smfio.h)<time.h>
       asctime                                 (smfio.h)<time.h>



!END
*****************************************************************************/


{
  
  struct tm *local=NULL;                         /* Time functions */
  time_t t;

  PGSt_SMF_status ret=PGS_S_SUCCESS;  /* Holds string of message 
						   from 
						   PGS_SMF_GetMsgByCode */
	
  char message[PGS_SMF_MAX_MSGBUF_SIZE]={0};    /* holds the message string 
					       associated with
					       the error code returned 
					       by GetMsg */

  char buf[PGS_SMF_MAX_MSGBUF_SIZE]={0};        /* Holds a concatenated dynamic
					       message string */
  char constrainedbuf[PGS_SMF_MAX_MSG_SIZE];

  t=time(NULL);                             /* Calculate timestamp */
  local=(struct tm *)localtime(&t);

  /* Get message from seed file based on error mnemonic */
  ret=PGS_SMF_GetMsgByCode(mnemonicstring, message);
  if(ret!=PGS_S_SUCCESS) exit(1);

  /* Concatenate timestamp, and user massage w/ buffer message */
  sprintf(buf, message, asctime(local), infostring);

  if(strlen(buf)>PGS_SMF_MAX_MSG_SIZE-1)modsmf(MODIS_E_SMF_ERROR, "", "smf.c, modsmf\0");

  strncpy(constrainedbuf, buf, PGS_SMF_MAX_MSG_SIZE);
  constrainedbuf[PGS_SMF_MAX_MSG_SIZE-1]='\0';

  /* Write message to LogStatus */
  ret=PGS_SMF_SetDynamicMsg(mnemonicstring, constrainedbuf, functionstring);
  if(ret!=PGS_S_SUCCESS) exit(1);

  /* Test for fatal error */
  if(PGS_SMF_TestFatalLevel(mnemonicstring)==PGS_TRUE)exit(1);  

  else return;

}

#endif  /* part of #ifndef/#define statements at beginning of file */

